import sys
import re


from tklib.tkutils import pint, pfloat, pconv
from tklib.tkobject import tkObject
from tklib.tkxml import get_root, get_section_inf_all, get_section_inf
from tklib.tkxml import file2dict, xml2dict, dict2xml, get_attrib, to_xml


xml_path = 'opt_2onl.xml'
progress_path = 'progress.opt'
results_path = 'results.opt'

argv = sys.argv
nargs = len(argv)
if nargs > 1:
    xml_path = argv[1]
if nargs > 2:
    progress_path = argv[2]
if nargs > 3:
    results_path = argv[3]


def read_data(infile, app = None, cparams = None, is_print = False):
    if is_print:
        print("")
        print(f"Input file:", infile)

    fp = open(infile, 'r')
    if fp is None:
        terminate(f"Error in read_data(): Can not read [{f}]", usage = usage)
    
    labels = []
    title = fp.readline().strip()
    aa = fp.readline().split()
    ndata = pint(aa[0])
    ncolumn = pint(aa[1])
    data_list = []
    for i in range(ncolumn):
        labels.append(fp.readline().strip())
        data_list.append([])

    for i in range(ndata):
        aa = fp.readline().split()
        if len(aa) < ncolumn:
            continue

        for j in range(len(aa)):
            data_list[j].append(pfloat(aa[j]))

    fp.close()

    return labels, data_list

def read_log(infile, app = None, cparams = None, is_print = False):
    fp = open(infile, 'r')
    if fp is None:
        terminate(f"Error in read_data(): Can not read [{f}]", usage = usage)
    
    E  = []
    DOS_exp   = []
    DOS_gauss = []
    DOS_total = []
    while 1:
        line = fp.readline()
        if not line:
            break

        if len(line) > 1 and line[0] == 'd':
            values = line.split()
            E.append(pfloat(values[1]))
            DOS_exp.append(pfloat(values[2]))
            DOS_gauss.append(pfloat(values[3]))
            DOS_total.append(pfloat(values[4]))

    fp.close()
    
    return {"E": E, "DOS_exp": DOS_exp, "DOS_gauss": DOS_gauss, "DOS_total": DOS_total}


def read_dat(infile, app = None, cparams = None, is_print = False):
    fp = open(infile, 'r')
    if fp is None:
        return None, None

    title = fp.readline().strip()
    _aa = fp.readline().split()
    ndata = pint(_aa[0])
    ncol  = pint(_aa[1])

    labels = []
    data_list = []
    for i in range(ncol):
        labels.append(fp.readline().strip())
        data_list.append([])

    for i in range(ndata):
        line = fp.readline()
        if not line:
            break
        
        _aa = line.split()
        for j, s in enumerate(_aa):
            data_list[j].append(pfloat(s))

    fp.close()

    return labels, data_list
    
def read_progress(infile, vars, app = None, cparams = None, is_print = False):
    inf = {}
    
    fp = open(infile, 'r')
    if not fp:
        return None

    while True:
        pos = fp.tell()
        line = fp.readline()
        if not line:
            break

        line = line.strip()
        if 'EVALUATION:' in line or 'GRADIENT:' in line:
            fp.seek(pos)
            break
        
        _aa = line.split(':', 1)
        if len(_aa) <= 1:
            continue

        inf[_aa[0].strip()] = pfloat(_aa[1])

    convergence = []
    _inf = None
    while True:
        line = fp.readline()
        if not line:
            break

        if 'EVALUATION:' in line or 'GRADIENT:' in line:
            if _inf is not None:
                convergence.append(_inf)
            _inf = {}

        line = line.strip()
#        print(f"l=[{line}]")
        _aa = line.split(',')
        for s in _aa:
            m = re.match("^(.*):(.*?)$", s)
            if m is None:
                continue

            _bb = m.groups()
            if len(_bb) <= 1:
                continue

            _inf[_bb[0].strip()] = pconv(_bb[1])

    convergence.append(_inf)

    xiter = range(len(convergence))
    vallist_dict = {}
    for var in vars:
        vallist_dict[var] = [c[var] for c in convergence]
#        print(f"  {var}:")
#        for i, y in zip(xiter, vallist_dict[var]):
#            print(f"    {i}: {y}")
    rms_key = None
    for key in convergence[0].keys():
        if ':rms' in key:
            rms_key = key
            break
    yrms = [c[rms_key] for c in convergence]

    return xiter, vallist_dict, rms_key, yrms, inf, convergence

def read_results(infile, app = None, cparams = None, is_print = False):
    inf = {}
    
    fp = open(infile, 'r')
    if not fp:
        return None

    while True:
        line = fp.readline()
        if not line:
            break
        
        _aa = line.split('=', 1)
        if len(_aa) <= 1:
            _aa = line.split(':', 1)
        if len(_aa) <= 1:
            continue

        inf[_aa[0].strip()] = pfloat(_aa[1])

    return inf

def read_optimizer_xml(infile, app = None, cparams = None, is_print = False):
    xml_root = get_root(xml_path)
    if not xml_root:
        return None

    parameter_inf = get_section_inf_all(xml_root, ['parameter-list', 'parameter'], ret_type = 'dict')
    setting_inf = get_section_inf_all(xml_root, ['settings', 'setting'], ret_type = 'dict')
    target_inf = get_section_inf_all(xml_root, ['target-list', 'target'], ret_type = 'dict')
    
    vars   = list(parameter_inf.keys())
    init_values = [pfloat(parameter_inf[key]['attrib']['nom']) for key in vars]

    eps    = pfloat(setting_inf['epscn']['attrib']['value'])
    maxiter = pint(setting_inf['maxfev']['attrib']['value'])

    reference = target_inf['idvg']['attrib']['reference']
    simulated = target_inf['idvg']['attrib']['simulated']
    
    return vars, init_values, maxiter, eps, reference, simulated, parameter_inf, setting_inf, target_inf

def main():
    print("")
    print(f"infile:", infile)
    print(f"progress_path:", progress_path)
    print(f"results_path:", results_path)

#    dict_xml = file2dict(infile)
#    print("dict:", dict_xml)
#    _xml = dict2xml(dict_xml)
#    print("xml:", _xml)
    
    xml_root = get_root(infile)
#    tag, text, attrib = get_attrib(xml_root)
#    print("tag=", tag)
#    print("text=", text)
#    print("attrib=", attrib)
#    to_xml('save.xml', xml_root, use_minidom = True, addindent = '    ')

    print()
    print("ATLAS optimization XML:")
    parameter_inf = get_section_inf_all(xml_root, ['parameter-list', 'parameter'], ret_type = 'dict')
#        parameter_inf = get_section_inf(xml_root, ['parameter-list', 'parameter'], ret_type = 'dict')
#        parameter_inf = get_section_inf(xml_root.find('parameter-list'), ['parameter'])
#       parameter_inf = get_section_inf(xml_root, ['.//parameter'])
#        parameter_inf = get_section_inf(xml_root, ['.//parameter-list'])
    print("parameter_inf:")
    for key, val in parameter_inf.items():
         print(f"  {key}: {val}")
    setting_inf = get_section_inf_all(xml_root, ['settings', 'setting'], ret_type = 'dict')
#        setting_inf = get_section_inf(xml_root, ['settings', 'setting'], ret_type = 'dict')
    print("setting_inf:")
    for key, val in setting_inf.items():
         print(f"  {key}: {val}")
    target_inf = get_section_inf_all(xml_root, ['target-list', 'target'], ret_type = 'dict')
#        target_inf = get_section_inf(xml_root, ['target-list', 'target'], ret_type = 'dict')
    print("target_inf:")
    for key, val in target_inf.items():
        print(f"  {key}: {val}")


if __name__ == '__main__':
    main()
    