import os
import sys
import shutil
import copy
import glob
import re
import csv
import numpy as np
from numpy import sin, cos, tan, arccos, arcsin, arctan, sqrt, exp, log, pi
from pprint import pprint

from tklib.tkobject import tkObject
from tklib.tkutils import IsDir, IsFile, SplitFilePath, pconv, pint, pfloat, safe_getelement
from tklib.tkutils import terminate, pint, pfloat, getarg, getintarg, getfloatarg
from tklib.tkutils import SplitFilePath, GetList
from tklib.tkre    import DelQuote, DelSpace
import tklib.tkre as tkre
from tklib.tkfile import tkFile
from tklib.tkcrystal.tkcrystalobject import tkCrystalObject
from tklib.tkcrystal.tkspacegroup import tkSpaceGroup
from tklib.tkprogvars import ProgramDir, DBDir, AtomDBDir, VASPDir, VASPPerlDir, VASPPythonDir
from tklib.tkfile import tkFile
from tklib.tkinifile import tkIniFile
from tklib.tksci.tksci import Reduce01, Round
from tklib.tksci.tksci import a0, NA, torad, todeg, Round, Factorize, IsFactor
from tklib.tksci.tksci import degcos, degsin, degtan, acos,asin, atan, degacos, degasin, degatan
from tklib.tksci.tkintegration import rieman_array_func
from tklib.tksci.tkmatrix import make_matrix1, make_matrix2, make_matrix3
from tklib.tkcrystal.tkcrystal import tkCrystal
from tklib.tkvariousdata import tkVariousData

from tklib.collabo.RH.parse_kpoints_opt import KPOINTS_OPT_Parser, write_kptopt_eig_csv, write_kptopt_dos_csv, write_high_symmetry_kpoints_opt


conv_dic = {
          "alpha": 1
        , "beta": 1
        , "gamma": 1
        , "delta": 1
        , "epsilon": 1
        , "zeta": 1
        , "eta": 1
        , "theta": 1
        , "iota": 1
        , "kappa": 1
        , "lambda": 1
        , "mu": 1
        , "nu": 1
        , "xi": 1
        , "pi": 1
        , "rho": 1
        , "sigma": 1
        , "tau": 1
        , "upsilon": 1
        , "phi": 1
        , "chi": 1
        , "psi": 1
        , "omega": 1
        }


class tkVASP(tkCrystalObject):
    def __init__(self, **args):
        self.crystal     = None
        self.incarinf    = None
        self.kpointsinf  = None
        self.outcarinf   = None
        self.eigenvalinf = None

        self.update(**args)

    def __del__(self):
        pass

    def __str__(self):
        return self.ClassPath()

    def __del__(self):
        pass


    def getdir(self, path):
        if IsDir(path):
            return path

        dirname, basename, filebody, ext = SplitFilePath(path)
        return dirname

    def get_CAR_dir(self, path, use_abspath = False):
        if use_abspath:
            path = os.path.abspath(path)

        if os.path.isdir(path):
            return path

        dirname = os.path.dirname(path)

        return dirname
        
        return os.path.join(CAR_dir, 'INCAR')

    def get_VASPPath(self, path, filename):
        CAR_dir = self.get_CAR_dir(path)
        return os.path.join(CAR_dir, filename)

    def get_INCAR(self, CAR_dir):
        return self.get_VASPPath(CAR_dir, 'INCAR')
#        return os.path.join(CAR_dir, 'INCAR')

    def get_POSCAR(self, CAR_dir):
        return self.get_VASPPath(CAR_dir, 'POSCAR')
#        return os.path.join(CAR_dir, 'POSCAR')

    def get_POTCAR(self, CAR_dir):
        return self.get_VASPPath(CAR_dir, 'POTCAR')
#        return os.path.join(CAR_dir, 'POTCAR')

    def get_KPOINTS(self, CAR_dir):
        return self.get_VASPPath(CAR_dir, 'KPOINTS')
#        return os.path.join(CAR_dir, 'KPOINTS')

    def get_KPOINTS_OPT(self, CAR_dir):
        return self.get_VASPPath(CAR_dir, 'KPOINTS_OPT')

    def get_XML(self, CAR_dir):
        return self.get_VASPPath(CAR_dir, 'vasprun.xml')

    def get_CONTCAR(self, CAR_dir):
        return self.get_VASPPath(CAR_dir, 'CONTCAR')
#        return os.path.join(CAR_dir, 'CONTCAR')

    def get_OUTCAR(self, CAR_dir):
        return self.get_VASPPath(CAR_dir, 'OUTCAR')
#        return os.path.join(CAR_dir, 'OUTCAR')

    def get_DOSCAR(self, CAR_dir):
        return self.get_VASPPath(CAR_dir, 'DOSCAR')
#        return os.path.join(CAR_dir, 'DOSCAR')

    def get_EIGENVAL(self, CAR_dir):
        return self.get_VASPPath(CAR_dir, 'EIGENVAL')
#        return os.path.join(CAR_dir, 'EIGENVAL')

    def read_potcar_inf(self, potcarpath):
        inf = {}

        fp = tkFile(potcarpath, 'r')
        if not fp or fp.fp is None:
            return None

        count = 0
        while 1:
            line = fp.ReadLine()
            if not line:
                break

#  PAW_PBE O 08Apr2002                    
            line = line.strip()
            ret = tkre.Search(r'\S+\s+([A-Z][a-z]?)', line)
            if ret:
                inf["PP{}".format(count)] = line
                inf["Atom{}".format(count)] = ret[1]
                line = fp.ReadLine()
                inf["Ne{}".format(count)] = pfloat(line)
                fp.SkipTo("End of Dataset")
                count += 1

        inf["nPPTypes"] = count
        self.potcarinf = inf
         
        return inf

    def read_poscar(self, poscarpath, is_print = False):
        inf = {}
        
        infp = tkFile(poscarpath, 'r')
        if not infp or infp.fp is None:
            return None

        sample_name = infp.ReadLine().strip()
        if is_print:
            print("sample name: [{}]".format(sample_name))

        a0 = pfloat(infp.ReadLine())
        aij = []
        aij.append(infp.ReadLine().split())
        aij.append(infp.ReadLine().split())
        aij.append(infp.ReadLine().split())
        for i in range(3):
            for j in range(3):
                aij[i][j] = pfloat(aij[i][j])

        if is_print:
            print("")
            print("Base lattice parameter: {} A".format(a0))
            print("Lattice vectors:")
            for i in range(3):
                print("  {:16.12f}  {:16.12f}  {:16.12f}".format(*aij[i]))

        line = infp.ReadLine()
        is_print = 1
#        print("line=", line)
        if re.search(r'^[\s\d\r\n]+$', line):
            POTCAR_path   = self.get_POTCAR(poscarpath)
#            print("POTCAR_path=", POTCAR_path)
            inf = self.read_potcar_inf(POTCAR_path)
#            print("inf=", inf)
            npp = inf['nPPTypes']
            atomtypes = []
            for i in range(npp):
                atomtypes.append(inf.get(f'Atom{i}', None))
            natoms = line.split()
        else:
            atomtypes = line.split()
            natoms = infp.ReadLine().split()

        for i in range(len(natoms)):
            natoms[i] = pint(natoms[i])

        if is_print:
            print("")
            print("Atom types and numbers:")
            for i in range(len(atomtypes)):
                print("  {:>4}: {:4d}".format(atomtypes[i], natoms[i]))

        sitetypes = []
        for ia in range(len(atomtypes)):
            for n in range(natoms[ia]):
                sitetypes.append(atomtypes[ia])

#Selective dynamics
        line = infp.ReadLine().strip()
#Direct
        if line.lower() != 'direct':
            line = infp.ReadLine()

        if is_print:
            print("")
            print("Atom positions:")

        pos = []    
        count = 0
        while 1:
            prevpos = infp.Tell()
            line = infp.ReadLine()
            if not line:
                break

            pos0 = line.split()
            if len(pos0) < 3:
                break

            try:
                pos0[0] = float(pos0[0])
            except:
                currpos = infp.Seek(prevpos)
                break

            pos.append([pfloat(pos0[0]), pfloat(pos0[1]), pfloat(pos0[2])])

            if is_print:
                print("  {:4}: ({:16.12f}, {:16.12f}, {:16.12f})"
                    .format(sitetypes[count], pos[count][0], pos[count][1], pos[count][2]))

            count += 1

        if is_print:
            print("")
            print("Build Crystal object:")

        cry = tkCrystal()
        cry.SetSampleName(sample_name)
        cry.SetCrystalName(sample_name)
        aij = a0 * np.array(aij)
#        print("aij_original=\n", aij)
        cry.SetLatticeVectors(aij)
#        print("aij_cry=\n", cry.LatticeVectors())
        latt = cry.LatticeParameters()
        for i in range(len(pos)):
            cry.AddAtomSite(name = sitetypes[i], pos = pos[i])

        cry.ExpandCoordinates()

#       cry.PrintInf()

        inf["SampleName"]  = sample_name
        inf["a0"]          = a0
        inf["aij"]         = np.array(a0) * aij
        inf["nAtomTypes"]  = len(atomtypes)
        inf["AtomTypes"]   = atomtypes
        inf["nAtoms"]      = natoms
        inf["crystal"]     = cry

        self.crystal = cry
        self.poscarinf = inf

        return cry

    def read_incar_inf(self, incarpath):
        fp = tkFile(incarpath, 'r')
        if not fp or fp.fp is None:
            return None

        inf = {'SYSTEM': ''}
        while 1:
            line = fp.ReadLine()
            if not line:
                break
            
            line = tkre.Sub(r'[!#].*$', '', line)
#            print("line=", line)
            aa = line.split(';')
            for i in range(len(aa)):
                ret = fp.Search(r'(\S+)\s*[=:]\s*(\S.*)\s*', aa[i])
                if ret is None:
                    continue
#        print("key=", ret)
                inf[ret[1]] = pconv(ret[2])

        self.incarinf = inf
        
        return inf

    def read_oszicar_inf(self, oszicarpath):
        fp = tkFile(oszicarpath, 'r')
        if not fp or fp.fp is None:
            return None

        inf = {}
        ion_steps = []
        all_steps = []
        ion_step = 1
        while 1:
            line = fp.ReadLine()
            if not line or line == '':
                break
                
#            print("l: ", line, end = '')                    
            if 'DAV:' in line:
                aa = line.split()
                all_steps.append({ 'ion_step': ion_step, 'electron_step': pint(aa[1]), 'E': pfloat(aa[2]), 'dE': pfloat(aa[3])})
#                print("E=", aa[2])
            elif 'F=' in line:
                aa = line.split()
                ion_steps.append({ 'ion_step': aa[0], 'F': pfloat(aa[2]), 'E0': pfloat(aa[4]), 'dE': pfloat(aa[6])})
#                print("F=", aa[2], "  E=", aa[4])
                ion_step += 1
            elif 'ncg' in line:
                pass
            else:
                pass

        fp.Close()

        inf['ion_steps'] = ion_steps
        inf['all_steps'] = all_steps

        self.oszicarinf = inf

        return inf

    def read_outcar_inf(self, outcarpath):
        fp = tkFile(outcarpath, 'r')
        if not fp or fp.fp is None:
            return None

        inf = {}
        
# read input parameters
#        line = fp.SkipTo(r'^\s*SYSTEM\s*=')
#        if line:
#            match = tkre.Search(r'SYSTEM\s*=\s*(\S.*?)\s*$', line)
#            if match:
#                inf['SYSTEM'] = match[1]

        line = fp.SkipTo(r'^\s*INCAR:')
        while 1:
            line = fp.ReadLine()
            if not line:
                break
#            if '----------' in line:
            if 'POTCAR:' in line:
                break

#            print("line=", line)
            ret = tkre.Search(r'(\S+)\s*=\s*(\S+)', line)
            if ret is not None and len(ret) >= 3:
                inf[ret[1]] = pconv(ret[2])
#                print(f"{ret[1]}={ret[2]}")

        line = fp.SkipTo("Startparameter for this run:")
        while 1:
            line = fp.ReadLine()
            if not line:
                break
            if '----------' in line:
                break

            aa = line.split(';')
            for i in range(len(aa)):
                ret = tkre.Search(r'(\S+)\s*=\s*(\S+)', aa[i])
                if ret is None:
                    continue

                if len(ret) >= 3:
                    inf[ret[1]] = pconv(ret[2])
#                    print("ret:", ret[1], ret[2], pconv(ret[2]))

#    print("inf=", inf)
#    print("ISPIN=", inf["ISPIN"])

#        s = fp.SkipTo("E-fermi :")
#        s = tkre.Search(r"E-fermi :\s+([+\-\d\.eEdD]+)", line)
#        EF = pfloat(s[1])
        m = fp.SearchTo(r"E-fermi :\s*([+\-\d\.eDdD]+)", defval = "0.0")
        EF = pconv(m[1])
        inf["EF"]      = EF
        inf["E-fermi"] = EF

        fp.Rewind()
# Read average core potnetials
        Vcore = []
# average (electrostatic) potential at core
        line = fp.SkipTo("potential at core")
#  the test charge radii are     1.2793  0.7215
        fp.ReadLine()
#  (the norm of the test charge is              1.0000)
        fp.ReadLine()
#       1 -52.1673       2 -52.1683       3 -52.1600       4 -52.1667       5 -70.7270
#       6 -70.7333       7 -70.7334       8 -70.7355       9 -70.7342      10 -70.7302
        icount = 0
        while 1:
            line = fp.ReadLine()
            if not line:
                break
            aa = line.split()
            if len(aa) <= 1:
                break

            for i in range(0, len(aa), 2):
                Vcore.append(pfloat(aa[i+1]))

        inf["Vcore"] = Vcore

        fp.Rewind()
# read ion step variation of converged total energy 
        count = 0
        energy = []
        energy_without_entropy = []
        energy_zero_sigma = []
        while 1:
            line = fp.SkipTo("\\s*free  energy   TOTEN  =");
            if not line:
                break
        
            pos = fp.Tell()
            s = tkre.Search(r'free  energy   TOTEN  =\s*(\S+)', line)
#            if pos > 0:
#                fp.Seek(pos, 0)

            key = 'FreeEnergy{}'.format(count)
            energy.append(pfloat(s[1]))

            line = fp.SkipTo("\\s*energy  without entrop");
            if not line:
                break

            s = tkre.Search(r'=\s*(\S+)\s.*=\s*(\S+)', line)
            energy_without_entropy.append(pfloat(s[1]))
            energy_zero_sigma.append(pfloat(s[2]))

            if pos > 0:
                fp.Seek(pos, 0)

#            print("E=", energy[count])
            count += 1
        inf['FreeEnergy'] = energy
        inf['TOTEN'] = energy[len(energy) - 1]
        inf['TOTEN_without_entropy'] = energy_without_entropy[len(energy_without_entropy) - 1]
        inf['TOTEN_zero_sigma']      = energy_zero_sigma[len(energy_zero_sigma) - 1]

# read final charges
        line = fp.SkipTo("\\s*total charge", 0)
        line = fp.SkipTo("# of ion     s       p       d       tot")
        line = fp.ReadLine()
        count = 0
        chargeinf = []
        while 1:
            line = fp.ReadLine()
            if not line:
                break
            line = line.strip()
            if tkre.Match(r'^-----', line):
                break
            
            iIon, s, p, d, tot, = line.split()
#            print("i=", iIon, s, p, d, tot)
            chargeinf.append({
                     's'.format(iIon): pfloat(s)
                    ,'p'.format(iIon): pfloat(p)
                    ,'d'.format(iIon): pfloat(d)
                    ,'tot'.format(iIon): pfloat(tot)
                    })
            count += 1                

# Check EF after charge lines
# BZINTS: Fermi energy:  2.623512;********** electrons
        fp.Rewind()
        while 1:
            m = fp.SearchTo(r"BZINTS: Fermi energy:\s*([+\-\d\.eDdD]+)", defval = None)
            if m is None:
                break
            else:
                EF = pconv(m[1])
                inf["EF"]         = EF
                inf["EF(BZINTS)"] = EF

        inf['FinalCharges'] = chargeinf
        inf['nIon'] = count
#        if 'T' in sainf['LHFCALC']:
        if 'T' in inf.get('LHFCALC', 'F'):
            inf['IsHF'] = 1
        else:
            inf['IsHF'] = 0

        if inf.get('METAGGA', '') != '':
            inf['IsMETAGGA'] = 1
        else:
            inf['IsMETAGGA'] = 0

        fp.Close()
        self.outcarinf = inf

        return inf

    def read_n_crystalstructures(self, outcarpath):
        fp = tkFile(outcarpath, 'r')
        n = 0
        if not fp:
            return 0

        while 1:
            line = fp.SkipTo('aborting loop because EDIFF is reached')
            if not line:
                break
        
            n += 1

        return n

    def read_next_crystalstructure(self, fp, template_cry, IsReduce01 = False):
        cry = copy.deepcopy(template_cry)
        inf = {}

        line = fp.SkipTo('aborting loop because EDIFF is reached')
        if not line:
            return None, None

        line = fp.SkipTo('VOLUME and BASIS-vectors')
        for i in range(4):
            fp.ReadLine()
        a1 = fp.ReadLine().split()
        a2 = fp.ReadLine().split()
        a3 = fp.ReadLine().split()
        a1[0], a1[1], a1[2] = pfloat(a1[0]), pfloat(a1[1]), pfloat(a1[2])
        a2[0], a2[1], a2[2] = pfloat(a2[0]), pfloat(a2[1]), pfloat(a2[2])
        a3[0], a3[1], a3[2] = pfloat(a3[0]), pfloat(a3[1]), pfloat(a3[2])
        cry.SetLatticeVectors([a1[:3], a2[:3], a3[:3]])
        latt = cry.LatticeParameters()

        line = fp.SkipTo('POSITION')
# skip ' ------...'
        line = fp.ReadLine()

        count = 0
        while 1:
            line = fp.ReadLine()
            if not line or '-----' in line:
                break
            x, y, z, fx, fy, fz, = line.split()
#           x = pfloat(x) / latt[0]
#            y = pfloat(y) / latt[1]
#            z = pfloat(z) / latt[2]

# Coordinate
            frx, fry, frz = cry.CartesianToFractional(pfloat(x), pfloat(y), pfloat(z))
            if IsReduce01:
                frx, fry, frz = Reduce01(frx), Reduce01(fry), Reduce01(frz)
#            print("")
#            print("***xyz=({},{},{})".format(x,y,z))
#            print("  frxyz=({},{},{})".format(frx,fry,frz))
#            x2,y2,z2 = cry.FractionalToCartesian(frx, fry, frz)
#            print("  2xyz=({},{},{})".format(x2,y2,z2))
#            exit()
# Force
            fx, fy, fz = cry.CartesianToFractional(pfloat(fx), pfloat(fy), pfloat(fz))
            cry.SetAtomSite(count, pos = [frx, fry, frz], 
                                   force = [pfloat(fx), pfloat(fy), pfloat(fz)])
            count += 1

        cry.ExpandCoordinates()

        line = fp.SkipTo("\\s*free  energy   TOTEN  =");
        ret  = fp.Search(r'free  energy   TOTEN  =\s*(\S+)', line)
        inf["TOTEN"] = pfloat(ret[1])
        
        pos = fp.Tell()

#       line = fp.SkipTo('MD-specific')
        line = fp.SkipTo('EKIN')
        if line:
            ret  = fp.Search(r'EKIN\s*=\s*([+\-\d\.eEdD]+)', line, defval = ['', ''])
            inf["EKIN"] = pfloat(ret[1])
            line = fp.SkipTo('temperature')
            ret  = fp.Search(r'EKIN_LAT\s*=\s*([+\-\d\.eEdD]+)', line, defval = ['', ''])
            inf["EKIN_LAT"] = pfloat(ret[1])
            ret  = fp.Search(r'temperature\s+([+\-\d\.eEdD]+)', line, defval = ['', ''])
            inf["temperature"] = pfloat(ret[1])
            line = fp.SkipTo('ETOTAL')
            ret  = fp.Search(r'ETOTAL\s*=\s*([+\-\d\.eEdD]+)', line, defval = ['', ''])
            inf["ETOTAL"] = pfloat(ret[1])
        else:
            fp.Seek(pos)

        return inf, cry

    def convert_kname(self, name):
        if len(name) <= 1:
            return name

        try:
            a = conv_dic[name.lower()]
            name = name.capitalize()
            name = '$\\{}$'.format(name)
        except:
            pass

        return name

    def read_kpoints_linemode(self, fp, inf, cry = None):
        IsEOF  = 0
        kp     = []
        kp_scf = []
        kp_all = []
        kpdic  = []
        ktot   = 0.0
        while 1:
            while 1:
                line1 = fp.ReadLine()
                if not line1:
                    IsEOF = 1
                    break

                line1 = line1.strip()
                if line1 == '':
                    next
                else:
                    break

            if IsEOF:
                break

            while 1:
                line2 = fp.ReadLine()
                if not line2:
                    IsEOF = 1
                    break

                line2 = line2.strip()
                if line2 == '':
                    next
                else:
                    break

            if IsEOF:
                break

#        print("line1: ", line1)
#        print("line2: ", line2)
            a1 = line1.split()
            n1 = len(a1)
            kx0, ky0,  kz0 = pfloat(a1[0]), pfloat(a1[1]), pfloat(a1[2])
            a2 = line2.split()
            n2 = len(a2)
            kx1, ky1,  kz1 = pfloat(a2[0]), pfloat(a2[1]), pfloat(a2[2])
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx1, ky1, kz1])
            ktot += dk

            match = tkre.Search(r'[!#]\s*(\S+)', line1)
            if match:
                name0 = match[1]
            else:
                name0 = ''
            match = tkre.Search(r'[!#]\s*(\S+)', line2)
            if match:
                name1 = match[1]
            else:
                name1 = ''

            name0c = self.convert_kname(name0)
            name1c = self.convert_kname(name1)
            w = 0.0

            kinf = {"kx0": kx0, "ky0": ky0, "kz0": kz0, "kname0": name0, "kname0_conv": name0c,
                    "kx1": kx1, "ky1": ky1, "kz1": kz1, "kname1": name1, "kname1_conv": name1c,
                    "kw": w, "dk": dk, "ktot": ktot}
            kp.append(kinf)
            kp_all.append(kinf.copy())
            kpdic.append(kinf.copy())

        inf['nk']          = len(kp)
        inf['kpoints']     = kp
        inf['kpoints_scf'] = kp_scf
        inf['kpoints_all'] = kp_all
        inf['kpointsdic']  = kpdic

        return inf

    def read_kpoints_pointmode(self, fp, inf, cry = None):
        kp     = []
        kp_scf = []
        kp_all = []
        while 1:
            line1 = fp.ReadLine()
            if not line1:
                break

#            print("line1: ", line1)
            a1 = line1.split()
            n1 = len(a1)
            if n1 < 4:
                continue

            kx1, ky1,  kz1, w = pfloat(a1[0]), pfloat(a1[1]), pfloat(a1[2]), pfloat(a1[3])

            match = tkre.Search(r'[!#]\s*(\S+)', line1)
            if match:
                name1 = match[1]
            else:
                name1 = ''

            name1c = self.convert_kname(name1)

            dk   = None
            ktot = None
            kinf = {"kx0": kx1, "ky0": ky1, "kz0": kz1, "kname0": name1, "kname0_conv": name1c, 
                    "kx1": kx1, "ky1": ky1, "kz1": kz1, "kname1": name1, "kname1_conv": name1c, 
                    "kw": w, "dk": dk, "ktot": ktot}
            kp_all.append(kinf)
            if w == 0.0:
                kp.append(kinf.copy())
            else:
                kp_scf.append(kinf.copy())

            kx0, ky0, kz0 = kx1, ky1, kz1

        nk = len(kp)
        kpdic  = []
        ktot   = 0.0
        dkdic  = 0.0
        dkvec0 = None
        for ik in range(nk):
            ki = kp[ik]
            kx1, ky1,  kz1, w = ki["kx0"], ki["ky0"], ki["kz0"], ki["kw"]
            name1 = ki["kname0"]

            if ik == 0:
                kx0, ky0, kz0 = kx1, ky1, kz1

            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx1, ky1, kz1])
            ktot  += dk
            dkdic += dk
            ki["dk"]   = dk
            ki["ktot"] = ktot

# add to kpdic
            if ik == 0:
                kpdic.append(ki.copy())
            elif ik == nk - 1:
                kpdic.append(ki.copy())
            else:
                kim = kp[ik-1]
                kip = kp[ik+1]
                kx0, ky0,  kz0, w = kim["kx0"], kim["ky0"], kim["kz0"], kim["kw"]
                kx2, ky2,  kz2, w = kip["kx0"], kip["ky0"], kip["kz0"], kip["kw"]
                r2   = (kx1 - kx0) * (kx1 - kx0) + (ky1 - ky0) * (ky1 - ky0) + (kz1 - kz0) * (kz1 - kz0)
                innp = (kx1 - kx0) * (kx2 - kx1) + (ky1 - ky0) * (ky2 - ky1) + (kz1 - kz0) * (kz2 - kz1)
                if r2 > 1.0e-6 and (name1 != '' or abs(innp) < 1.0e-6):
                    kidic = ki.copy()
                    kidic["dk"] = dkdic
                    kpdic.append(kidic)
                    dkdic = 0.0

            kx0, ky0, kz0 = kx1, ky1, kz1

        ktot   = 0.0
        for ik in range(len(kp_scf)):
            ki = kp_scf[ik]
            kx1, ky1,  kz1, w = ki["kx0"], ki["ky0"], ki["kz0"], ki["kw"]
            if ik == 0:
                kx0, ky0, kz0 = kx1, ky1, kz1
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx1, ky1, kz1])
            ktot += dk
            ki["dk"]   = dk
            ki["ktot"] = ktot
            kx0, ky0, kz0 = kx1, ky1, kz1

        ktot   = 0.0
        for ik in range(len(kp_all)):
            ki = kp_all[ik]
            kx1, ky1,  kz1, w = ki["kx0"], ki["ky0"], ki["kz0"], ki["kw"]
            if ik == 0:
                kx0, ky0, kz0 = kx1, ky1, kz1
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx1, ky1, kz1])
            ktot += dk
            ki["dk"]   = dk
            ki["ktot"] = ktot
            kx0, ky0, kz0 = kx1, ky1, kz1

        inf['nk']          = len(kp)
        inf['kpoints']     = kp
        inf['kpoints_scf'] = kp_scf
        inf['kpoints_all'] = kp_all
        inf['kpointsdic']  = kpdic

        return inf


    def read_kpoints(self, KPOINTS_path, cry = None):
        if cry is None:
            cry = self.crystal

        inf = {}

        fp = tkFile(KPOINTS_path, 'r')

        inf['title']      = fp.ReadLine().strip()
        inf['nk_target']  = pint(fp.ReadLine())
        mode              = fp.ReadLine().strip().upper()
        if mode[0] == 'L':
            inf['mode']       = mode
            inf['coord_mode'] = fp.ReadLine().strip()
            inf = self.read_kpoints_linemode(fp, inf, cry)
        else:
            inf['mode']       = mode
            inf['coord_mode'] = mode
            inf = self.read_kpoints_pointmode(fp, inf, cry)

        fp.Close()

        self.kpointsinf = inf

        return inf

    def read_eigenval_opt(self, EIGENVAL_path, normalize_E = True, KPOINTS_inf = None, ISPIN = None, IsHF = None, IsMETAGGA = None, EF = None, NELECT = None, cry = None):
        if cry is None:
            cry = self.crystal
        if NELECT is None:
            NELECT = self.outcarinf["NELECT"]
        if ISPIN is None:
            ISPIN = self.outcarinf["ISPIN"]
        if IsHF is None:
            IsHF = self.outcarinf["IsHF"]
        if IsMETAGGA is None:
            IsMETAGGA = self.outcarinf["IsMETAGGA"]
        if EF is None:
            EF = self.outcarinf["EF"]

        KPOINTS_OPT_path = self.get_KPOINTS_OPT(EIGENVAL_path)
        XML_path = self.get_XML(EIGENVAL_path)

        kpoints_inf = self.read_kpoints(KPOINTS_OPT_path, cry = cry)
#        print("kpoints_inf=", kpoints_inf)

        Nline  = 20
        parser = KPOINTS_OPT_Parser(XML_path, Nline)
        EF     = parser.get_fermi_level()
#    parser.reciprocal_lattice = self.vasprun.final_structure.lattice.reciprocal_lattice.matrix
#    parser.bands_data = self.get_eigenvalues()
#    parser.kpath = self.get_kpath()
#        bands_data = self.get_eigenvalues()
#                    bands_data.append({
#                        'spin': spin,
#                        'kpoint': kpoint,
#                        'eigenvalue': float(r.text.strip()) - fermi_level
#                    })
        kpoints = parser.kpoints
        weights = parser.weights_list
        kpath   = parser.kpath

        inf = {}

        bands_data      = parser.get_eigenvalues()
        eigenvalues_dict, occupancies_dict = parser.eigenvalues_per_kpoint()

        inf["n"]  = None
        inf["nk"] = len(kpoints)
        inf["nLevels"] = max(len(eigenvalues) for eigenvalues in eigenvalues_dict.values())
#        for i, k in enumerate(kpoints):
#            print("k=", i, k)
#        for i, eig in enumerate(eigenvalues_dict.values()):
#            print("eig=", i, len(eig))
#        exit()
            
#        print("nk=", inf['nk'])
#        print("nLevels=", inf['nLevels'])

        ktot_prev = 0.0        
        Elist     = [[]] * inf["nk"]
#        Elist_scf = []
#        Elist_all = []
        kx0 = None
        ky0 = None
        kz0 = None
        for key, E in eigenvalues_dict.items():
            spin, ik = key
            if '1' in spin:
                spin = 0
            else:
                spin = 1

            occ = occupancies_dict[key]
            ktot = kpath[ik]
            wk   = weights[ik]
            kv   = kpoints[ik]
            dk = ktot - ktot_prev
#            print(f"{spin} {ik}: {ktot} {wk} {kv} {occ}: {E}")

            if len(Elist[ik]) > 2:
                _kv0, _kv1, _kv2, _wk, _dk, _ktot, _Eup, _occup, _Edn, _occdn = Elist
            else:
                _Eup = None
                _occup = None
                _Edn = None
                _occdn = None

            if spin == 0:
                Elist[ik] = [kv[0], kv[1], kv[2], wk, dk, ktot, E,    occ, _Edn, _occdn]
            else:
                Elist[ik] = [kv[0], kv[1], kv[2], wk, dk, ktot, _Eup, _occup, E, occ]

            ktot_prev = ktot

#        print()
#        c = 0
#        for E in Elist: 
#            print("E=", E)
#            if len(E) < 2:
#                c += 1
#        print("nk=", inf["nk"])
#        print("c=", c)
        
        inf['EList']     = Elist
        inf['EList_scf'] = Elist
        inf['EList_all'] = Elist

        self.eigenvalinf = inf

        return inf


    def read_eigenval(self, EIGENVAL_path, normalize_E = True, KPOINTS_inf = None, ISPIN = None, IsHF = None, IsMETAGGA = None, EF = None, NELECT = None, cry = None):
        INCAR_path    = self.get_INCAR(EIGENVAL_path)
        incar = self.read_incar_inf(INCAR_path)
        if incar:
            LKPOINTS_OPT = incar.get('LKPOINTS_OPT', None)
            if LKPOINTS_OPT is None or 'false' in LKPOINTS_OPT.lower():
                LKPOINTS_OPT = False
            else:
                LKPOINTS_OPT = True
        else:
            LKPOINTS_OPT = True
        
        KPOINTS_OPT_path = self.get_KPOINTS_OPT(EIGENVAL_path)
        if LKPOINTS_OPT and os.path.isfile(KPOINTS_OPT_path):
            return self.read_eigenval_opt(EIGENVAL_path, normalize_E, KPOINTS_inf, ISPIN, IsHF, IsMETAGGA, EF, NELECT, cry)

        if cry is None:
            cry = self.crystal
        if NELECT is None:
            NELECT = self.outcarinf["NELECT"]
        if ISPIN is None:
            ISPIN = self.outcarinf["ISPIN"]
        if IsHF is None:
            IsHF = self.outcarinf["IsHF"]
        if EF is None:
            EF = self.outcarinf["EF"]

        try:
            kptinf = KPOINTS_inf['kpoints_all']
        except:
            kptinf = None
#        print("kpinf=", kpinf)

        inf = {}

        fp = tkFile(EIGENVAL_path, 'r')

#    4    4    1    1
        fp.ReadLine()
#  0.1236685E+02  0.3291500E-09  0.3291500E-09  0.5272308E-09  0.5000000E-15
        fp.ReadLine()
#  1.000000000000000E-004
        fp.ReadLine()
#  CAR 
        fp.ReadLine()
# ZnO [PAW_PBE54] (band)                  
        inf["title"] = fp.ReadLine().strip()
#     36    204     24
        aa = fp.ReadLine().split()
        inf["n"]       = pint(aa[0])
        inf["nk"]      = pint(aa[1])
        inf["nLevels"] = pint(aa[2])
        Elist     = []
        Elist_scf = []
        Elist_all = []
        kx0 = None
        ky0 = None
        kz0 = None
        for ik in range(inf["nk"]):
# next lines can be brank, to be skipped
            while 1:
                pos = fp.Tell()
                line = fp.ReadLine()
                if line == '':
                    break

#                print("line1=", line)
                if line.strip() != '':
                    fp.Seek(pos)
                    break

# 
#            fp.ReadLine()

#  0.0000000E+00  0.0000000E+00  0.0000000E+00  0.4901961E-02
            line = fp.ReadLine()
#            print("line2=", line)
            aa = line.split()
            kx = pfloat(aa[0])
            ky = pfloat(aa[1])
            kz = pfloat(aa[2])
            wk = pfloat(aa[3])

#    1      -16.924278   1.000000
            Eups   = []
            Edns   = []
            occups = []
            occdns = []
            for iL in range(inf["nLevels"]):
                line = fp.ReadLine()
#                print("line2=", line)
                aa = line.split()
                aa[1] = pfloat(aa[1])
                if len(aa) > 2:
                    aa[2] = pfloat(aa[2])
                else:
                    if ISPIN == 1:
                        aa.append(0.0)
                if ISPIN == 1:
                    if normalize_E:
                        Eups.append(aa[1] - EF)
                    else:
                        Eups.append(aa[1])
                    occups.append(aa[2])
                else:
                    if normalize_E:
                        Eups.append(aa[1] - EF)
                        Edns.append(aa[2] - EF)
                    else:
                        Eups.append(aa[1])
                        Edns.append(aa[2])
                    occups.append(pfloat(aa[3]))
                    occdns.append(pfloat(aa[4]))

            dk   = None
            ktot = None
            kinf = [kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns]
            Elist_all.append(kinf)
            if IsHF:
                if not kptinf or kptinf[ik]['kw'] == 0.0:
                    Elist.append(kinf)
                else:
                    Elist_scf.append(kinf)
            else:
                Elist.append(kinf)

        fp.Close()

# Calculate k distances
        ktot = 0.0
        for ik in range(len(Elist)):
            kx, ky, kz, wk, _dk, _ktot, Eups, occups, Edns, occdns = Elist[ik]
            if kx0 is None:
                kx0, ky0, kz0 = kx, ky, kz
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx, ky, kz])
            ktot += dk
            kx0 = kx
            ky0 = ky
            kz0 = kz

            Elist[ik] = [kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns]

        ktot = 0.0
        for ik in range(len(Elist_scf)):
            kx, ky, kz, wk, _dk, _ktot, Eups, occups, Edns, occdns = Elist_scf[ik]
            if kx0 is None:
                kx0, ky0, kz0 = kx, ky, kz
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx, ky, kz])
            ktot += dk
            kx0 = kx
            ky0 = ky
            kz0 = kz

            Elist_scf[ik] = [kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns]

        ktot = 0.0
        for ik in range(len(Elist_all)):
            kx, ky, kz, wk, _dk, _ktot, Eups, occups, Edns, occdns = Elist_all[ik]
            if kx0 is None:
                kx0, ky0, kz0 = kx, ky, kz
            dk = cry.GetReciprocalDistanceFromK([kx0, ky0, kz0], [kx, ky, kz])
            ktot += dk
            kx0 = kx
            ky0 = ky
            kz0 = kz

            Elist_all[ik] = [kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns]


        inf['EList']     = Elist
        inf['EList_scf'] = Elist_scf
        inf['EList_all'] = Elist_all
        
        self.eigenvalinf = inf

        return inf

    def find_band_edges_from_dos(self, E, DOS, EF0 = 0.0, DOSth = 1.0e18):
        EV = 1.0e30
        EC = 1.0e30
        i0 = -1
        for i in range(len(E)):
#            print("i=", i, E[i], EF0)
            if E[i] >= EF0:
                i0 = i
                break

        for i in range(i0, 0, -1):
#            print("i2=", i, DOS[i], DOSth)
            if DOS[i] > DOSth:
                EV = E[i+1]
                break

        for i in range(i0, len(E), +1):
#            print("i3=", i, DOS[i], DOSth)
            if DOS[i] > DOSth:
                EC = E[i-1]
                break

        if EC < EV:
            EC = EF0
            EV = EF0

        return EV, EC

    def gbandedges(self, dir = '.', fn_outcar = 'OUTCAR', fn_eig = 'EIGENVAL', ef = None):
        inf = {}

        """
        if ef is None:
            ief = 0
        else:
            ief = 1

        if not ief:
            if os.path.exists(fn_outcar):
                for line in open(fn_outcar, 'r'):
                    if line.find('E-fermi') > -1:
                        ef = float(line.split()[2])
            else:
                print(f'Error in tkVASP.gbandedges: [{fn_outcar}] does not exist!')
                return None
        """

        if not os.path.exists(fn_eig):
            print(f'Error in tkVASP.gbandedges: [{fn_eig}] does not exist!')
            return None

        f_eig = open(fn_eig)
        lines = f_eig.readlines()
        f_eig.close()

        nspin   = int(lines[0].split()[3])
        nkpoint = int(lines[5].split()[1])
        nband   = int(lines[5].split()[2])

        eup   = ef + 99.0
        edown = ef - 99.0
        k = []
        for i in range(nkpoint):
            iline = 7 + i * (nband + 2)
#    print(f"lines[{iline}]: {lines[iline]}", end = '')
            k.append(lines[iline].split()[0:3])
            for j in range(nband):
                for s in range(nspin):
                    eval = float(lines[8 + i * (nband + 2) + j].split()[s + 1])
                    d    = eval - ef 
                    if (d > 0):
                        if (eval < eup):
                            eup = eval
                            kup = i
                            bup = j
                            sup = s
                    if (d < 0):
                        if (eval > edown):
                            edown = eval
                            kdown = i
                            bdown = j
                            sdown = s

        inf["nspin"]   = nspin
        inf["nkpoint"] = nkpoint
        inf["nband"]   = nband
        inf["nkpoint"] = nkpoint
        inf["nlines"]  = len(lines)

        inf["EF"] = ef
        inf["Eg"] = eup - edown

        inf["ECBM"]      = eup
        inf["spin_CBM"]  = sup + 1
        inf["ik_CBM"]    = kup+1
        inf["k_CBM"]     = [float(k) for k in k[kup]]
        inf["iband_CBM"] = bup + 11
        inf["spin_CBM"]  = sup + 1

        inf["EVBM"]      = edown
        inf["ik_VBM"]    = kdown + 1
        inf["k_VBM"]     = [float(k) for k in k[kdown]]
        inf["iband_VBM"] = bdown + 11
        inf["spin_VBM"]  = sdown + 1

#        print("inf=", inf)
        return inf

    def find_band_edges_from_eigenval(self, EF0 = 0.0, normalize_E = True, eigenvalinf = None, ISPIN = None, occ_th = 0.5, occ_th0 = 1.0e-4):
        if ISPIN is None:
            if self.outcarinf:
                ISPIN = self.outcarinf["ISPIN"]
            elif self.incarinf:
                ISPIN = self.incarinf["ISPIN"]
            else:
                ISPIN = 1
        if EF0 is None:
            EF0 = 0.0
        if eigenvalinf is None:
            eigenvalinf = self.eigenvalinf
        EPS = 1.0e-3

        nk      = eigenvalinf["nk"]
        nLevels = eigenvalinf["nLevels"]
#        print("k points in EIGENVAL:")
#        print("nk=", nk)
#        print("nLevels=", nLevels)
        EV    = -1e10
        EC    =  1e10
        EHOMO = -1.0e10
        ELUMO =  1.0e10
        EList = eigenvalinf['EList']
        for i in range(len(EList)):
            el = EList[i]
            kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns = el
            print("  ({:8.4f} {:8.4f} {:8.4f}) w={:8.4g}  {:8.4f} {:12.4f}"
                    .format(kx, ky, kz, wk, dk, ktot))
            for il in range(nLevels):
                if occups[il] > occ_th0 and EHOMO < Eups[il]:
                    EHOMO = Eups[il]
                if occups[il] > occ_th and EV < Eups[il]:
#                if Eups[il] < EF0 + EPS and EV < Eups[il]:
                    EV = Eups[il]

                if occups[il] <= occ_th0 and ELUMO > Eups[il]:
                    ELUMO = Eups[il]
                if occups[il] <= occ_th and EC > Eups[il]:
#                if Eups[il] > EF0 - EPS and EC > Eups[il]:
                    EC = Eups[il]
#                    print(f"  il={il} E={Eups[il]} occ={occups[il]} EF0={EF0}")
#                    print("   CBM")

                if ISPIN == 2:
                    if occdns[il] > occ_th0 and EHOMO < Edns[il]:
                        EHOMO = Edns[il]
                    if occups[il] > occ_th and EV < Edns[il]:
#                    if Edns[il] < EF0 + EPS and EV < Edns[il]:
                        EV = Edns[il]

                    if occdns[il] <= occ_th0 and ELUMO > Edns[il]:
                        ELUMO = Edns[il]
                    if occups[il] <= occ_th and EC > Edns[il]:
#                    if Edns[il] > EF0 - EPS and EC > Edns[il]:
                        EC = Edns[il]

#        print("edges:", EHOMO, ELUMO, EV, EC)

        if EC < EV:
            EC = EF0
            EV = EF0
            EHOMO = EF0
            ELUMO = EF0
#        print("edges:", EHOMO, ELUMO, EV, EC)

        inf = { "EV": EV, "EC": EC, "Eg": EC - EV, "EF0": EF0, "EHOMO": EHOMO, "ELUMO": ELUMO }

        return inf

# lm: 'tot'|'up'|'dn': Total DOS, atom_type + 'tot'|'up'|'dn': Atom-decomposed DOS
#     's' 'p', 'd' + ''|'up'|'dn': l-decomposed DOS   's', 'px', etc + ''|'up'|'dn': lm-decomposed DOS
    def cal_lmdecomposed_dos(self, inf, atom_type, lm, save_lmdos = False, outfile = None, is_print = False):
        npdos = inf["npdos"]
        atom_types = inf["atom_types"]  
        atom_sites = inf["atom_sites"]
        lmlabels  = inf["lmlabels"]
        ks    = inf["ks"]
        ndata = inf["ndata"]
#pdos_atom_list[iatomsite][lm][idata]
        pdos_atom_list = inf["PDOS_atom_list"]

        if is_print:
            print(f"Calculate lm-deomposed DOS for atom [{atom_type}] lm=[{lm}]")

        if npdos == 0:
            if is_print:
                print(f"  No lm-decomposed data (npdos={npdos}. Skip")
            return None

        lmdos = np.zeros(ndata)
        for ia in range(len(atom_sites)):
            atom_name = atom_sites[ia]

            skip = True
            choise_total = ['tot', 'up', 'dn']
            if lm in choise_total:
                skip = False
            choise_atom = [f'{atom_name} tot', f'{atom_name} up', f'{atom_name} dn']
            if lm in choise_atom or atom_name == atom_type:
                skip = False

            if skip:
                continue

            for iE in range(ndata):
                for ilm in range(len(lmlabels)):
                    lmstr = lmlabels[ilm]
#                    print(f"at={atom_type}/{atom_name} ilm={ilm} lm=[{lm}]/[{lmstr}]")

                    skip = True
# total/up/dnを通す
                    if lm == 'tot' or lm == f'{atom_name} tot':
                        skip = False
                    elif (lm == 'up' or lm == f'{atom_name} up') and 'up' in lmstr:
                        skip = False
                    elif (lm == 'dn' or lm == f'{atom_name} dn') and 'dn' in lmstr:
                        skip = False

# lm=s up, px up etcを通す
                    if lm == lmstr:
                        skip = False
# lm=s, p, d, px, py, etcを通す => s up, s dn, sを通す
                    elif lm in lmstr:
                        skip = False

                    if skip:
                        continue

#                    print(" pass")

                    if lm == 'tot up' and 'up' in lmstr:
                        lmdos[iE] += ks * pdos_atom_list[ia][ilm][iE]
                    elif lm == 'tot dn' and 'dn' in lmstr:
                        lmdos[iE] += ks * pdos_atom_list[ia][ilm][iE]
                    else:
                        lmdos[iE] += ks * pdos_atom_list[ia][ilm][iE]
                
        if save_lmdos:
#           print("lmdos=", lmdos)
            data_list = inf["data_list"][0]
            xE      = data_list[0]
            pdos_atom_list[ia][ilm][iE]

            Elabels = ["E (eV)", f"{atom_type} {lm}"]
            ylist   = [xE, lmdos]

            for ia in range(len(atom_sites)):
                atom_name = atom_sites[ia]
                if lm != 'tot' and atom_name != atom_type:
                    continue
        
                for ilm in range(len(lmlabels)):
                    lmstr = lmlabels[ilm]
                    if lm != 'tot' and lm != f'{atom_type} tot' and lm not in lmstr:
                        continue
    
                    Elabels.append(f"{atom_type}#{ia} {lmstr}#{ilm}")
                    ylist.append(pdos_atom_list[ia][ilm])

            if outfile is None:
                if lm == 'tot':
                    outfile = f"pdos_total.xlsx"
                else:
                    outfile = f"pdos_{atom_type}_{lm}.xlsx"
            tkVariousData().to_excel(outfile, Elabels, ylist)

        return lmdos

    def read_doscar(self, DOSCAR_path, cry = None, EF = None, normalize_E = True, unit = '', 
                IsSpinPolarized = None, IsNonCollinear = None, SpinOrbit = None, print_level = 1):
        if EF is None:
            EF = self.outcarinf.get("EF", 0.0)
        if cry is None:
            cry = self.crystal
        ISPIN = 1
        if IsSpinPolarized is None:
            ISPIN = self.outcarinf.get("ISPIN", 1)
            if ISPIN == 1:
                IsSpinPolarized = 0
            else:
                IsSpinPolarized = 1
        if SpinOrbit is None:
            LSORBIT = self.outcarinf.get("LSORBIT", 0)
            if 'T' in LSORBIT:
                SpinOrbit = 1
            else:
                SpinOrbit = 0
        if IsNonCollinear is None:
            LNONCOLLINEAR = self.outcarinf.get("LNONCOLLINEAR", 0)
            if 'T' in LNONCOLLINEAR:
                IsNonCollinear = 1
            else:
                IsNonCollinear = 0

        poscar_path = self.get_POSCAR(DOSCAR_path)
        if print_level:
            print(f"Read crystal structure from [{poscar_path}]")
        cry = self.read_poscar(poscar_path, is_print = False)
        poscar_inf = self.poscarinf
        atom_types = poscar_inf["AtomTypes"]
        sites = cry.ExpandedAtomSiteList()
        atom_sites = []
        for site in sites:
            atom_sites.append(site.AtomNameOnly())
        natom_types = len(atom_types)
        natom_sites = len(atom_sites)
        if print_level:
            print("natom_types=", natom_types)
            print("atom_types =", atom_types)
            print("natom_types=", natom_sites)
            print("atom_sites =", atom_sites)
            print("natom_sites=", natom_sites)

        inf = {}

        if print_level:
            print()
            print(f"Read DOSCAR from [{DOSCAR_path}]")
        fp = tkFile(DOSCAR_path, 'r')
        if not fp or fp.fp is None:
#            terminate("Error in read_average_core_potentials: Can not read [{}]".format(DOSCAR_path))
            return { "nE": 0, "E": [], "TotalDOSup": [], "TotalDOSdn": [], "Neup": [], "Nedn": [] }

        aa = fp.readline().split()
        nions_with_empty = pint(aa[0])
        nions            = pint(aa[1])
        include_PDOS     = pint(aa[2])
        inf["NCDIJ"]     = pint(aa[3])
        if print_level:
            print("nions=", nions)
            print("nions_with_empty=", nions_with_empty)

        if nions != natom_sites:
            print()
            print(f" ERROR: nions={nions} taken from {DOSCAR_path} is different from "
                + f"natom_sites={natom_sites} taken from {poscar_path}")
            input("Press ENTER to terminate>>\n")
            exit()

        aa = fp.readline().split()
        Vcell = pfloat(aa[0])
        a, b, c = pfloat(aa[1]), pfloat(aa[2]), pfloat(aa[3])
        inf["POTIM"] = pfloat(aa[4])

        line = fp.readline()
        TEBEG = pfloat(line)

        fp.readline()

        line = fp.readline()
        sample_name = line.strip()
#        print("s=", inf["SampleName"])

# t/PDOSの読み込み
        if print_level >= 2:
            print("Read total DOS")

        line = fp.ReadLine()
        aa = line.split()
        if len(aa) < 5:
#0123456789012345678901230123456789012345678901234567890123456789012345678901234567890123456789
#     10.00000000    -30.0000000040001     -0.73600000      1.00000000
#            print(f"lines=[{line[0:16]}][{line[16:32]}][{line[32:37]}][{line[37:53]}][{line[53:]}]")
            Emax   = pfloat(line[0:16])
            Emin   = pfloat(line[16:32])
            nE     = pint(line[32:37])
            Efermi = pfloat(line[37:53])
#            print("E=", Emax, Emin, nE, Efermi)
        else:
            Emax   = pfloat(aa[0])
            Emin   = pfloat(aa[1])
            nE     = pint(aa[2])
            Efermi = pfloat(aa[3])

# 指数が3桁の場合、eが削除されるので追加
        def vfloat(s):
            m = re.sub(r"(\d)\-(\d)", "$1e\-$2$", s)
            if m:
                s = m
            return pfloat(s)

        ylist_total = None
        for i in range(nE):
            line = fp.readline()
            if not line:
                break

            aa = line.split()
            if len(aa) == 0:
                break

            nc = len(aa)
            if ylist_total is None:
                ylist_total = np.zeros([nc, nE])

            for ic in range(nc):
                ylist_total[ic][i] = vfloat(aa[ic])
#            print("e=", ylist[0][i], ylist[1][i], ylist[2][i])

# lm-decomposed DOSの読み込み
        if print_level >= 2:
            print("Read lm-decomposed DOS")

        nPDOS = 0
        ylist_lmdos = []
        while 1:
            line = fp.ReadLine()
            if not line:
                break

            if print_level >= 2:
                print(f"  Read lm-decomposed DOS for {nPDOS}-th site ({atom_sites[nPDOS]})")
            
            ylist = None
            for idata in range(nE):
                line = fp.readline()
                if not line:
                    break

                aa = line.split()
                if len(aa) == 0:
                    break

                ncolumn = len(aa)
                if ylist is None:
                    ylist = np.zeros([ncolumn - 1, nE])

                for ic in range(1, ncolumn):
                    ylist[ic - 1][idata] = vfloat(aa[ic])

            ylist_lmdos.append(ylist)
            nPDOS += 1

# SpinOrbit == 1 の場合、さらに mx, my, mzのデータが続く
        fp.Close()

        if len(ylist_lmdos) == 0 or ylist_lmdos[0] is None:
            ylist_lmdos = [[], []]
            npdos = 0
        else:
            npdos = len(ylist_lmdos[0])

        if normalize_E:
            Elist = [e - EF for e in ylist_total[0]]
        else:
            Elist = ylist_total[0]

        Vcell = cry.Volume()
        k = 1.0 / (Vcell * 1.0e-24)
        if unit == '/cm3':
            for i in range(1, len(ylist_total)):    
                ylist_total[i] *= k

            if npdos > 0:
                for isite in range(nions):
                    for ilm in range(npdos):
#                        print("i=", isite, ilm)
                        ylist_lmdos[isite][ilm] *= k
        
        if nc <= 3:
            tDOSuplist       = ylist_total[1]
            tDOSdnlist       = None
            Neuplist         = ylist_total[2]
            Nednlist         = None
        else:
            tDOSuplist       = ylist_total[1]
            tDOSdnlist       = ylist_total[2]
            Neuplist         = ylist_total[3]
            Nednlist         = ylist_total[4]

        PDOS_atom_list = ylist_lmdos

# total / sum DOSの計算
        if tDOSdnlist is None:
            tDOSlist = tDOSuplist
            Nelist   = Neuplist
        else:
            tDOSlist = np.zeros(nE)
            Nelist   = np.zeros(nE)
            for i in range(nE):
                tDOSlist[i] = tDOSuplist[i] + tDOSdnlist[i]
                Nelist[i]   = Neuplist[i] + Nednlist[i]

        if tDOSdnlist is None:
            if npdos == 3:
                lmlabels = ["s", "p", "d"]
            else:
                lmlabels = ["s", "py", "pz", "px", "dxy", "dyz", "dz2", "dxy", "dx2-y2"]
        else:
            if npdos == 6:
                lmlabels = ["s up", "s dn", "p up", "p dn", "d up", "d dn"]
            else:
                lmlabels = ["s up", "s dn", 
                      "py up", "py dn", "pz up", "pz dn", "px up", "px dn",
                      "dxy up", "dxy dn", "dyz up", "dyz dn", "dz2 up", "dz2 dn", "dxz up", "dxz dn", "dx2-y2 up", "dx2-y2 dn"]

# 原子種ごとにPDOS (lmlabels全て) の和を取る
# IsSpinPolarized == 0の場合、2倍する
        if SpinOrbit == 1 and IsSpinPolarized == 0:
            ks = 2.0
        else:
            ks = 1.0

        PDOS_list = {}
        sumDOS_list = np.zeros(nE)
        if npdos > 0:
            for iion in range(nions):
                atom_name = atom_sites[iion]
                PDOS_list[atom_name] = np.zeros(nE)

            for iion in range(nions):
                atom_name = atom_sites[iion]
                for ilm in range(len(lmlabels)):
                    for iE in range(nE):
                        v = ks * PDOS_atom_list[iion][ilm][iE]
                        PDOS_list[atom_name][iE] += v
                        sumDOS_list[iE] += v 

            interstitialDOS_list = tDOSlist - sumDOS_list
        else:
            interstitialDOS_list = np.zeros(nE)

        inf["cry"] = cry

        inf["SampleName"] = sample_name
        inf["IsSpinPolarized"] = IsSpinPolarized
        inf["ISPIN"]           = ISPIN
        inf["SpinOrbit"]       = SpinOrbit
        inf["IsNonCollinear"]  = IsNonCollinear
        inf["ks"]              = ks

        inf["nions_with_empty"] = nions_with_empty
        inf["nions"]            = nions
        inf["atom_types"]  = atom_types
        inf["atom_sites"]  = atom_sites
        inf["natom_types"] = natom_types
        inf["natom_sites"] = natom_sites
        inf["Vcell"]  = Vcell
        inf["a"]      = a
        inf["b"]      = b
        inf["c"]      = c
        inf["EF"]     = EF
        inf["Efermi"] = Efermi
        inf["TEBEG"]  = TEBEG
        inf["include_PODS"] = include_PDOS

        inf["Emax"]   = Emax
        inf["Emin"]   = Emin
        inf["nE"]     = nE

        inf["E"]          = Elist
        inf["TotalDOS"]   = tDOSlist
        inf["TotalDOSup"] = tDOSuplist
        inf["TotalDOSdn"] = tDOSdnlist
        inf["Ne"]         = Nelist
        inf["Neup"]       = Neuplist
        inf["Nedn"]       = Nednlist

        inf["npdos"]             = npdos
        inf["lmlabels"]          = lmlabels
        inf["PDOS_list"]         = PDOS_list
#        inf["PDOS_list_labels"]  = atom_types
        inf["PDOS_atom_list"]    = PDOS_atom_list
        inf["sumDOS_list"] = sumDOS_list

        include_sum = False
        if tDOSdnlist is None:
            if include_sum:
                data_list = [Elist, tDOSlist, sumDOS_list, interstitialDOS_list]
                labels = ["E (eV)", "Total", "sum", "interstitial"]
            else:
                data_list = [Elist, tDOSlist, interstitialDOS_list]
                labels = ["E (eV)", "Total", "interstitial"]
            
            if npdos > 0:
                for atom_name in atom_types:
                    data_list.extend([PDOS_list[atom_name]])
                labels.extend(atom_types)
        else:
            if include_sum:
                data_list = [Elist, tDOSlist, tDOSuplist, tDOSdnlist, sumDOS_list, interstitialDOS_list]
                labels = ["E (eV)", "Total", "Total up", "Total dn", "sum", "interstitial"]
            else:
                data_list = [Elist, tDOSlist, tDOSuplist, tDOSdnlist, interstitialDOS_list]
                labels = ["E (eV)", "Total", "Total up", "Total dn", "interstitial"]

            if npdos > 0:
                for atom_name in atom_types:
                    data_list.extend([PDOS_list[atom_name]])
                labels.extend(atom_types)

        inf["data_list"] = data_list
        inf["labels"]    = labels
        inf["lm_labels"] = lmlabels
#        for id in range(len(data_list)):
#            print("id=", id, labels[id], type(data_list[id]))
#            for iE in range(len(data_list[id])):
#                print("  ", iE, data_list[id][iE])

        return inf

    def read_DOS(file, cry = None, IsSpinPolarized = None, IsNonCollinear = None, EF = None, normalize_E = True, unit = '', print_level = 1):
        inf = {}
        if '.dat' in file:
            E_raw, dos_raw = np.genfromtxt(file, skip_header=1, usecols=(0,1), unpack=True)
            inf["E"]        = E_raw
            inf["nE"] = len(E_raw)
            inf["TotalDOS"] = dos_raw
        else:
            inf = self.read_doscar(file, cry = cry, EF = EF, normalize_E = normalize_E, unit = unit, 
                        IsSpinPolarized = IsSpinPolarized, IsNonCollinear = IsNonCollinear, print_level = print_level)

        return inf

    def read_files(self, file, file_types, EF = None, normalize_E = True, unit = '/cm3', data_for_bandedges = None, 
                    exit_by_error = False, print_level = 0, terminate = terminate): 
        base_path = self.getdir(file)

        INCAR_path    = self.get_INCAR(base_path)
        POSCAR_path   = self.get_POSCAR(base_path)
        POTCAR_path   = self.get_POTCAR(base_path)
        CONTCAR_path  = self.get_CONTCAR(base_path)
        OUTCAR_path   = self.get_OUTCAR(base_path)
        EIGENVAL_path = self.get_VASPPath(base_path, 'EIGENVAL')
        DOSCAR_path   = self.get_VASPPath(base_path, 'DOSCAR')
        XML_path      = self.get_VASPPath(base_path, 'vasprun.xml')
        if print_level >= 1:
            print("")
            print("CAR dir   : ", base_path)
            print("  INCAR   : ", INCAR_path)
            print("  POSCAR  : ", POSCAR_path)
            print("  POTCAR  : ", POTCAR_path)
            print("  CONTCAR : ", CONTCAR_path)
            print("  OUTCAR  : ", OUTCAR_path)
            print("  EIGENVAL: ", EIGENVAL_path)
            print("  DOSCAR  : ", DOSCAR_path)
            print("  XML     : ", XML_path)

        inf = {}
        if 'INCAR' in file_types:
            if print_level >= 1:
                print("*** Read INCAR from [{}]".format(INCAR_path))
            inf["INCAR"] = self.read_incar_inf(INCAR_path)
            if exit_by_error and inf["INCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(INCAR_path))

        if 'POTCAR' in file_types:
            if print_level >= 1:
                print("*** Read POTCAR from [{}]".format(POTCAR_path))
            inf["POTCAR"] = self.read_potcar_inf(POTCAR_path)
            if exit_by_error and inf["POTCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(POTCAR_path))

        if 'POSCAR' in file_types:
            if print_level >= 1:
                print("*** Read crystal structure from [{}]".format(POSCAR_path))
            inf["POSCAR"] = self.read_poscar(POSCAR_path)
            if exit_by_error and inf["POSCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(POSCAR_path))

        if 'CONTCAR' in file_types:
            if print_level >= 1:
                print("*** Read crystal structure from [{}]".format(CONTCAR_path))
            inf["CONTCAR"] = self.read_poscar(CONTCAR_path)
            if exit_by_error and inf["POSCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(CONTCAR_path))

        if 'OUTCAR' in file_types:
            if print_level >= 1:
                print("*** Read OUTCAR from [{}]".format(OUTCAR_path))
            inf["OUTCAR"] = self.read_outcar_inf(OUTCAR_path)
            if exit_by_error and inf["OUTCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(OUTCAR_path))

        if 'DOSCAR' in file_types:
            if print_level >= 1:
                print("*** Read DOSCAR from [{}]".format(DOSCAR_path))
            inf["DOSCAR"] = self.read_doscar(DOSCAR_path, EF = EF, normalize_E = normalize_E, unit = unit, print_level = print_level)
            if exit_by_error and inf["DOSCAR"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(DOSCAR_path))

        if 'EIGENVAL' in file_types:
            if print_level >= 1:
                print("*** Read EIGENVAL from [{}]".format(EIGENVAL_path))
            inf["EIGENVAL"] = self.read_eigenval(EIGENVAL_path, normalize_E = normalize_E)
            if exit_by_error and inf["EIGENVAL"] is None:
                terminate("Error in tkvasp.read_files: Can not read [{}]".format(EIGENVAL_path))

        return inf

