import os
import sys
import math
import numpy as np

from OpenGL.GL import *
from OpenGL.GLUT import *
from OpenGL.GLU import *


def draw_sphere(x, y, z, radius, 
                R = 1.0, G = 0.0, B = 0.0, alpha = 1.0, slices = 10, stacks = 10):
#    glColor3f(R, G, B)
    current_material = glGetFloatv(GL_CURRENT_COLOR)
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, [R, G, B, alpha])

    glPushMatrix()
    glTranslatef(x, y, z)
    glutSolidSphere(radius, slices, stacks)
    glPopMatrix()

    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, current_material)

def draw_cylinder(x0, y0, z0, x1, y1, z1, radius = 0.1, 
                  R = 0.5, G = 0.5, B = 0.5, alpha = 1.0, slices = 10, stacks = 10): 
#    glColor3f(R, G, B)
    current_material = glGetFloatv(GL_CURRENT_COLOR)
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, [R, G, B, alpha])

    # Calculate direction vector
    direction = np.array([x1 - x0, y1 - y0, z1 - z0])
    length = np.linalg.norm(direction)
    if length == 0.0: return

    direction /= length

    # Calculate rotation axis and angle
    up = np.array([0, 0, 1])
    axis = np.cross(up, direction)
    angle = np.arccos(np.dot(up, direction)) * 180 / np.pi

    # Draw cylinder (shaft of the arrow)
    glPushMatrix()
    glTranslatef(x0, y0, z0)
    glRotatef(angle, axis[0], axis[1], axis[2])
    gluCylinder(gluNewQuadric(), radius, radius, length, slices, stacks)
    glPopMatrix()

    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, current_material)

def draw_arrow(x0, y0, z0, x1, y1, z1, radius = 0.1, arrow_h = 0.2, arrow_r = 0.2, 
               R = 1.0, G = 0.0, B = 0.0, alpha = 0.5, slices = 10, stacks = 10):
#    glColor3f(R, G, B)
    current_material = glGetFloatv(GL_CURRENT_COLOR)
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, [R, G, B, alpha])

    # Calculate direction vector
    direction = np.array([x1 - x0, y1 - y0, z1 - z0])
    length = np.linalg.norm(direction)
    direction /= length

    # Calculate rotation axis and angle
    up = np.array([0, 0, 1])
    axis = np.cross(up, direction)
    angle = np.arccos(np.dot(up, direction)) * 180 / np.pi

    # Draw cylinder (shaft of the arrow)
    glPushMatrix()
    glTranslatef(x0, y0, z0)
    glRotatef(angle, axis[0], axis[1], axis[2])
    gluCylinder(gluNewQuadric(), radius, radius, length - arrow_h, slices, stacks)
    glPopMatrix()

    # Draw cone (head of the arrow)
    glPushMatrix()
    glTranslatef(x1, y1, z1)
    glRotatef(angle, axis[0], axis[1], axis[2])
    glTranslatef(0, 0, -arrow_h)
    gluCylinder(gluNewQuadric(), arrow_r, 0, arrow_h, slices, slices)
    glPopMatrix()

    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, current_material)

