import tklib.tkimport as imp
numpy = imp.import_lib("numpy", stop_by_error = False)
mpl   = imp.import_lib("matplotlib", stop_by_error = False)
mplcursors = imp.import_lib("mplcursors", stop_by_error = False)
tk  = imp.import_lib("tkinter", stop_by_error = False)
from tkinter import Tk, Menu
from tkinter import ttk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
imp.messages(stop_by_error = False)


from tklib.tkobject import tkObject
from tklib.tkparams import tkParams


class tkPlot_tkinter(tkObject):
    def __init__(self, plt, parent = None, title = None):
        self.plt = plt
        self.canvas = None

        if parent is None:
            self.parent = tk.Tk()
        else:
            self.parent = parent

        if title:
            parent.title(title)

    def create_window(self, fig):
        self.add_toolbar()
#        self.add_toolbar_button(text = "test")
#        self.update()

    def update(self):
        if self.canvas:
            self.canvas.draw()
        self.parent.update()

    def show(self):
        self.update()
    
    def pause(self, t):
        self.update()
        self.plt.pause(t)

    def draw(self):
#        self.plt.draw()
#        self.canvas.draw()
        self.update()

    def get_current_fig_manager(self):
        return self.plt.get_current_fig_manager()
        
    def tight_layout(self, **kwargs):
        return self.plt.tight_layout(**kwargs)

    def subplots_adjust(self, **kwargs):
        return self.plt.subplots_adjust(**kwargs)

    def axes(self, *args, **kwargs):
        return self.plt.axes(*args, **kwargs)

    def subplots(self, ny, nx, figsize = (8, 6), dpi = 100, tight_layout = True, **kwargs):
        self.fig = Figure(figsize = figsize, dpi = dpi)
        if tight_layout:
            self.fig.set_constrained_layout(True)

        self.plot_axes = []
        i = 1
        for _nx in range(nx):
            for _ny in range(ny):
                ax = self.fig.add_subplot(ny, nx, i, **kwargs)
                self.plot_axes.append(ax)
                i += 1
        return self.fig, self.plot_axes

    def legend(self, ax, line_inf_list = None, draggable = True, **kwargs):
        if line_inf_list is not None:
            ins = sum(line_inf_list)
            labels = [l.get_label() for l in ins]
            _legend = ax.legend(ins, labels, **kwargs)
        else:
            _legend = ax.legend(**kwargs)

        _legend.set_draggable(draggable)

    def add_toolbar(self):
        self.toolbar = tk.Frame(self.parent)
        self.toolbar.pack(side = tk.TOP, fill = tk.X)

    def add_toolbar_button(self, text, command = None):
        if command:
            btn = tk.Button(self.toolbar, text = text, command = command)
        else:
            btn = tk.Button(self.toolbar, text = text)
        btn.pack(side = tk.LEFT)
        return btn

# フィギュアをTkinterキャンバスに埋め込み
    def add_canvas(self, fig = None, parent = None):
        if parent is None:
            parent = self.parent

        if fig is None:
            fig = self.fig

        self.canvas = FigureCanvasTkAgg(fig, master = parent)
        self.canvas.get_tk_widget().pack(side = tk.TOP, fill = tk.BOTH, expand = 1)

        # Matplotlibのツールバーを作成
        self.mpl_toolbar = NavigationToolbar2Tk(self.canvas, self.parent)
        self.mpl_toolbar.update()

        self.canvas.draw()
        
        return self.canvas
        
