import tklib.tkimport as imp
np    = imp.import_lib("numpy", stop_by_error = False)
pd    = imp.import_lib("pandas", stop_by_error = False)
mpl   = imp.import_lib("matplotlib", stop_by_error = False)
mplcursors = imp.import_lib("mplcursors", stop_by_error = False)
tk  = imp.import_lib("tkinter", stop_by_error = False)
from tkinter import Tk, Menu
from tkinter import ttk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure
imp.messages(stop_by_error = False)


import sys
import random
from matplotlib import pyplot as plt
import matplotlib.widgets as wg


from tklib.tkobject import tkObject
from tklib.tkparams import tkParams


class RangeSelector():
    def __init__(self, mode = 'x', axis = None, color = 'red', linestyle = 'dashed', linewidth = 0.5, print_level = 1):
        self.ax = axis
        self.color = color
        self.linestyle = linestyle
        self.linewidth = linewidth
        self.print_level = print_level

        self.x0 = 0.0
        self.x1 = 0.0
        self.y0 = 0.0
        self.y1 = 0.0
        
        self.vline0 = None
        self.vline1 = None
        self.hline0 = None
        self.hline1 = None

        self.set_mode(mode)

        self.button_press   = self.ax.figure.canvas.mpl_connect('button_press_event',   self.on_press)
        self.button_release = self.ax.figure.canvas.mpl_connect('button_release_event', self.on_release)
        self.motion         = self.ax.figure.canvas.mpl_connect('motion_notify_event',  self.on_mouse_move)

    def set_mode(self, mode):
        self.moving = False
        self.mode_x = 'x' in mode
        self.mode_y = 'y' in mode

        if self.vline0 is None:
            self.vline0 = self.ax.axvline(self.x0, color = self.color, linestyle = self.linestyle, linewidth = self.linewidth)
        if self.vline1 is None:
            self.vline1 = self.ax.axvline(self.x1, color = self.color, linestyle = self.linestyle, linewidth = self.linewidth)
        if self.hline0 is None:
            self.hline0 = self.ax.axhline(self.y0, color = self.color, linestyle = self.linestyle, linewidth = self.linewidth)
        if self.hline1 is None:
            self.hline1 = self.ax.axhline(self.y1, color = self.color, linestyle = self.linestyle, linewidth = self.linewidth)
        self.vline0.set_visible(False)
        self.vline1.set_visible(False)
        self.hline0.set_visible(False)
        self.hline1.set_visible(False)

    def on_press(self, event):
        if event.button != 1: # 左クリック以外
            return

        if event.inaxes != self.ax:
            return

        self.moving = True

        self.x0 = event.xdata
        self.y0 = event.ydata
        if self.mode_x:
            self.vline0.set_xdata([self.x0])
            self.vline0.set_visible(True)
        if self.mode_y:
            self.hline0.set_ydata([self.y0])
            self.hline0.set_visible(True)

        self.ax.figure.canvas.draw()

    def on_mouse_move(self, event):
        if not self.moving:
            return
        if event.inaxes != self.ax:
            return

        self.x1 = event.xdata
        self.y1 = event.ydata
        if self.mode_x:
            self.vline1.set_xdata([self.x1])
            self.vline1.set_visible(True)
        if self.mode_y:
            self.hline1.set_ydata([self.y1])
            self.hline1.set_visible(True)
        
        self.ax.figure.canvas.draw()

    def on_release(self, event):
        if not self.moving:
            return

        self.x1 = event.xdata
        self.y1 = event.ydata

        self.finalize()

    def finalize(self):
        if self.mode_x:
            self.vline1.set_xdata([self.x1])
            if self.print_level:
                print(f"Selected x range: {self.x0} - {self.x1}")
        if self.mode_y:
            self.hline1.set_ydata([self.y1])
            if self.print_level:
                print(f"Selected y range: {self.y0} - {self.y1}")

        self.ax.figure.canvas.draw()
        self.moving = False


class tkPlotEvent(tkObject):
    def __init__(self, plt, distance = 'r', **args):
        self.plt = plt
# データ点との距離の測り方: 'r', 'x', 'y'
        self.distance = distance
# マウスクリック時に最近接データの情報をコンソールに表示
        self.hit_itarget = None
        self.hit_axisinf = None
        self.stop_flag   = False

# mplcursors/annotate用変数
        self.annotation = tkParams()

# マウスカーソル用変数
        self.follow_mouse = tkParams()

# データ移動変数
        self.move_points = tkParams()

# stop buttun用コンテナ
        self.stop_button = None

        super(tkObject, self).__init__(**args)
#        super(tkParams, self).__init__(parameter_file, app, args)
#        self.update(**args)

        self.axis_inf = []

    def button_click(self, e):
        self.stop_flag = True

    def finalize_button(self, text = 'finished'):
        self.button.label.set_text(text)

    def add_button(self, button_region = [0.15, 0.95, 0.10, 0.03], plot_region = [0.92, 0.15],
                        text = 'stop', color = '#f8e58c', hovercolor = '#38b48b', callback = None):
        if callback is None:
            callback = self.button_click

        self.button_region = button_region
        self.plot_region   = plot_region
        self.color = color
        self.hovercolor = hovercolor
        self.callback = callback
        self.ax_button = plt.axes(button_region)

        self.button = wg.Button(self.ax_button, text, color = color, hovercolor = hovercolor)
        self.button.on_clicked(callback)

        return self.button
    
    def layout(self, show = False):
#        self.plt.tight_layout()
        self.plt.subplots_adjust(top = self.plot_region[0], bottom = self.plot_region[1])
        
        if show:
            if use_pause is not None:
                self.plt.pause(use_pause)
            else:
                self.plt.show()

    def remove(self, index):
        if type(index) is int:
            self.axis_inf.pop(index)
            return

        for i in range(len(self.axis_inf)):
            if self.axis_inf[i]['label'] == index:
                self.axis_inf.pop(i)
                return

    def add_stop_button(self, button_region = [0.15, 0.95, 0.10, 0.03], plot_region = [0.92, 0.15], 
                label = 'stop', color = '#f8e58c', hovercolor = '#38b48b'):
        def on_click(e):
            self.stop_button.status = 'stop'
            print("Stop button was pressed. Terminating...")

        def set_text(label = None):
            if label is not None:
                self.stop_button.button.label.set_text(label)

        def adjust_plot(plot_region = None):
            if plot_region is None:
                plot_region = self.stop_button.plot_region
            plt.subplots_adjust(top = plot_region[0], bottom = plot_region[1])

        class tkStopButton():
            def __init__(self, plt, button_region, plot_region):
                self.plt = plt

                self.__status = 'running'
                self.button_region = button_region
                self.plot_region = plot_region

                self.on_click    = on_click
                self.set_text    = set_text
                self.adjust_plot = adjust_plot

                self.ax = self.plt.axes(button_region)
                self.button = wg.Button(self.ax, label, color = color, hovercolor = hovercolor)
                self.button.on_clicked(self.on_click)

            @property
            def status(self):
                return self.__status

            @status.setter
            def status(self, f):
                self.__status = f

        self.stop_button = tkStopButton(self.plt, button_region, plot_region)

    def add_data(self, axis_inf = {}):
#{"label": "test predict",     "plot_type": "2D", "axis": axis, "data": test_predict_data,  
# "xlist": [index_test_list, *x_test_list], "xlabels": ["index", *x_labels]})

        plot_type = axis_inf['plot_type']

# プロットの仕方によってデータ構造が違うので、ここで統一しておく
        axis = axis_inf['axis']
        data = axis_inf['data']
        if axis_inf.get('xlabel', None):
            axis.xlabel = axis_inf['xlabel']
        else:
            axis.xlabel = axis.get_xlabel()
        if axis_inf.get('ylabel', None):
            axis.ylabel = axis_inf['ylabel']
        else:
            axis.ylabel = axis.get_ylabel()

        if plot_type == '2D' or plot_type == 'plot':
            for i in range(len(data)):
                data[i].xdata = data[i].get_xdata()
                data[i].ydata = data[i].get_ydata()
        elif plot_type == 'scatter':
            data = [data]
            axis_inf['data'] = data
            data[0].xdata = axis_inf['x']
            data[0].ydata = axis_inf['y']
        else:
            print("")
            print(f"Error in tkplotevent.add_data(): Invalid plot_type [{plot_type}]")
            input(">>")
            exit()

        self.axis_inf.append(axis_inf)
        return axis_inf

# マウスクリック時にクリックされたグラフ軸を特定
# axesが与えられたときはmplcursors、与えられない場合はplot/scatter用
    def find_target_axes(self, event, axes = None):
        if axes:
            naxis = len(axes)
        else:
            naxis = len(self.axis_inf)
        itargets     = []
        hit_axisinfs = []
        for i in range(naxis):
#            print("a=", event.inaxes, self.axis_inf[i]["axis"])
            if axes:
                if event.inaxes == axes[i]:
                    return i, axes[i]
            else:
                if event.inaxes == self.axis_inf[i]["axis"]:
                   itargets.append(i)
                   hit_axisinfs.append(self.axis_inf[i])

        if axes:
            return None, None
        
        if len(itargets) == 0:
            return None, None

        return itargets, hit_axisinfs

    def convert_coord(self, x0, y0, axis_s, axis_t):
        if axis_s is None or axis_t is None or x0 is None or y0 is None:
            return None, None

        xlim0 = axis_s.get_xlim()
        ylim0 = axis_s.get_ylim()
        xlim1 = axis_t.get_xlim()
        ylim1 = axis_t.get_ylim()

        x1 = xlim1[0] + (xlim1[1] - xlim1[0]) / (xlim0[1] - xlim0[0]) * (x0 - xlim0[0])
        y1 = ylim1[0] + (ylim1[1] - ylim1[0]) / (ylim0[1] - ylim0[0]) * (y0 - ylim0[0])

        return x1, y1

    def prepare_popup_menu(self, fig, parent = None):
        self.popup_menu = tkParams()
        self.popup_menu.fig = fig
        if parent is None:
            self.popup_menu.parent = Tk() # Tkinterのルートウィンドウを作成（非表示）
            self.popup_menu.parent.withdraw()   # ウィンドウを表示しない
        else:
            self.popup_menu.parent = parent

        self.popup_menu.menu = Menu(self.popup_menu.parent, tearoff = 0)

        def add_menu(label = "run", command = lambda: print("runが選択されました")):
            self.popup_menu.menu.add_command(label = label, command = command)

        self.popup_menu.add_menu = add_menu

    def register_popup_menu_event(self, on_click = None):
        def on_click_default(event):
            if event.button == 3: # 右クリック（button==3）
                self.popup_menu.menu.tk_popup(event.guiEvent.x_root, event.guiEvent.y_root)

        if on_click is None:
            on_click = on_click_default

        self.popup_menu.fig.canvas.mpl_connect('button_press_event', on_click)

    def prepare_move_points(self):
        self.move_points.gco = None

        def add(ax, x, y, marker = 'o', picker = 15, color = 'red'):
            ax.plot(x, y, marker, markerfacecolor = color, markeredgecolor = color, picker = picker)
        self.move_points.add = add

    def prepare_move_text(self, fig = None):
        self.move_text = tkParams()
        self.move_text.fig = fig
        self.move_text.text_inf = []

        def get_xy(ax, x, y, x_list, y_list, x_target, frac, xlim, ylim):
            if x is None or y is None or x_list is None or y_list is None: return None, None

            if x_list is not None:
                if isinstance(x_list, pd.DataFrame):
                    _x_list = x_list.to_numpy().T
                else:
                    _x_list = x_list
                if isinstance(y_list, pd.DataFrame):
                    _y_list = y_list.to_numpy().T
                else:
                    _y_list = y_list
                if not isinstance(_x_list[0], np.ndarray):
                    _x_list = np.array([_x_list])
                if not isinstance(_y_list[0], np.ndarray):
                    _y_list = np.array([_y_list])
                _x_list = _x_list.flatten()
                _y_list = _y_list.flatten()

                auto_determine = False
                if x_target is None:
                    if frac is None:
                        auto_determine = True
                        frac = random.random()

                    _min = min(_x_list)
                    _max = max(_x_list)
                    if xlim:
                        _min = max([_min, xlim[0]])
                        _max = min([_max, xlim[1]])
                    x_target = _min + (_max - _min) * frac
#                print("x_target=", frac, _min, _max, x_target)

                x = x_target
                y = np.interp(x, x_list, y_list)

                if auto_determine and ylim and (y < ylim[0] or ylim[1] < y):
                    for f in np.arange(0.0, 1.0, 0.025):
                        x = _min + (_max - _min) * f
                        y = np.interp(x, x_list, y_list)
                        if ylim[0] <= y and y <= ylim[1]:
                            break
                    else:
                        print("no data found")
                        x = _min + (_max - _min) * 0.5
                        y = np.interp(x, x_list, y_list)

                '''
                idx_min = 0
                dx_min = 1.0e300
                for i, _x in enumerate(_x_list):
                        absdx = abs(_x - x_target)
                        if absdx < dx_min:
                            idx_min = i
                            dx_min = absdx

                x = _x_list[idx_min]
                y = _y_list[idx_min]
                '''

            return x, y

        def add_annotation(ax, ax_ref, x = None, y = None, x_list = None, y_list = None, 
                        x_offset = 0, x_target = None, frac = None, xlim = None, ylim = None,
                        text = "no name", fontsize = 10, 
                        ha = 'center', va = 'center', color = 'black', fc = "w", ec = "none", alpha = 0.5):
            x, y = get_xy(ax, x, y, x_list, y_list, x_target, frac, xlim, ylim)
            if x is None: return

#            print("annotate:", x, x_offset, x + x_offset, y, text)
            hannotate = ax.annotate(text, xy = (x + x_offset, y), xytext = (0, 0), color = color,
#            hannotate = ax.annotate(text, xy = (x, y), xytext = (x, y),
                    textcoords = 'offset points', fontsize = fontsize,
                    bbox={'facecolor': fc, 'alpha': alpha, 'pad': 0.5, 'edgecolor': ec, 'boxstyle': 'round'})
#                    bbox = dict(boxstyle = "round,pad=0.3", fc = fc, ec = ec))

            self.move_text.text_inf.append([ax, ax_ref, hannotate, False, "annotate"])

        def add_text(ax, ax_ref, x = None, y = None, x_list = None, y_list = None, 
                    x_offset = 0, x_target = None, frac = None, xlim = None, ylim = None,
                    text = "no name", fontsize = 10, ha = 'center', va = 'center', color = 'black', fc = "w", ec = "none", alpha = 0.5):
            x, y = get_xy(ax, x, y, x_list, y_list, x_target, frac, xlim, ylim)

            htext = ax.text(x + x_offset, y, text, fontsize = fontsize, ha = ha, va = va, color = color,
                    bbox={'facecolor': fc, 'alpha': alpha, 'pad': 0.5, 'edgecolor': ec, 'boxstyle': 'round'})

            self.move_text.text_inf.append([ax, ax_ref, htext, False, "text"])

        self.move_text.add_text       = add_text
        self.move_text.add_annotation = add_annotation


    def register_move_text_event(self, fig = None, activate = False):
        if fig is not None:
            self.move_text.fig = fig

        self.move_text.is_active = activate

        def activate_func(flag = True):
            self.move_text.is_active = flag

        self.move_text.activate = activate_func


        def on_click(event):
            for inf in self.move_text.text_inf:
                annot       = inf[2]
                contains, _ = annot.contains(event)
                if contains:
                    inf[3] = True
                    return

        def on_release(event):
            for inf in self.move_text.text_inf:
                inf[3] = False

        def on_move(event):
            if not self.move_text.is_active: return

            if event.button == 1:
                for inf in self.move_text.text_inf:
                    ax          = inf[0]
                    ax_ref      = inf[1]
                    annot       = inf[2]
                    is_dragging = inf[3]
                    _type       = inf[4]

# ref_axisでの値
                    x0 = event.xdata
                    y0 = event.ydata
# axisでの値に戻す
                    x1, y1 = self.convert_coord(x0, y0, ax_ref, ax)
                    if x1 is None:
                        return

                    if is_dragging:
                        if event.xdata is not None and event.ydata is not None:
                            if _type == 'text':
                                annot.set_x(x1)
                                annot.set_y(y1)
                                annot.xy = (x1, y1)
                            elif _type == 'annotate':
                                annot.set_x(0)
                                annot.set_y(0)
                                annot.xy = (x1, y1)
                            self.move_text.fig.canvas.draw()
                        return

        self.move_text.fig.canvas.mpl_connect('button_press_event',   on_click)
        self.move_text.fig.canvas.mpl_connect('button_release_event', on_release)
        self.move_text.fig.canvas.mpl_connect('motion_notify_event',  on_move)

    def register_move_points_event(self, fig):
        self.move_points.fig = fig

        def motion(event):
            if self.move_points.gco is None:
                return
            x = event.xdata
            y = event.ydata
            self.move_points.gco.set_data([x], [y])
            plt.draw()

        def onpick(event):
            self.move_points. gco = event.artist
#            plt.title(self.move_points.gco)

        def release(event):
            self.move_points.gco = None

        fig.canvas.mpl_connect('motion_notify_event', motion)
        fig.canvas.mpl_connect('pick_event', onpick)
        fig.canvas.mpl_connect('button_release_event', release)

    def prepare_annotation(self):
        self.annotation.sel      = None
        self.annotation.axes     = []
        self.annotation.ref_axes = []
        self.annotation.lines    = []
        self.annotation.labels   = []
        self.annotation.x_lists  = []
        self.annotation.y_lists  = []
        self.annotation.inf_list = []
        self.annotation.annotation_formats = []
        self.annotation.inf_formats = []
        self.annotation.x = None
        self.annotation.y = None
        self.annotation.refresh_cursors = None

        self.cursor = None
        self.annotation.id_add_connection = False

        def add_line(label, axis, ref_axis, x_list, y_list, line, inf_list = None, annotation_format = None, inf_format = None):
            self.annotation.axes.append(axis)
            self.annotation.ref_axes.append(ref_axis)
            self.annotation.lines.append(line)
            self.annotation.labels.append(label)
            self.annotation.x_lists.append(x_list)
            self.annotation.y_lists.append(y_list)
            self.annotation.inf_list.append(inf_list)
            self.annotation.annotation_formats.append(annotation_format)
            self.annotation.inf_formats.append(inf_format)

        self.annotation.add_line = add_line

    def register_annotation_event(self, fig, activate = True, on_mouse_move = None, on_click = None, print_level = 0):
        self.annotation.fig = fig
        self.annotation.is_active = activate

        def activate_func(flag = True):
            self.annotation.is_active = flag

        self.annotation.activate = activate_func

        def on_mouse_move_annotate(event):
            if not self.annotation.is_active: return

            cur_sel = self.annotation
            if cur_sel.sel is None:
                return

            eps = 1.0e-2

#twinx/y()を使った場合、event.xdata/ydataは、最後のaxis (ref_axis) における数値 => 変換する必要あり
#selection objectのtarget[0]/[1]は、選択されたグラフ線 (axis) の数値
            iaxis, ref_axis = self.find_target_axes(event, axes = cur_sel.ref_axes)
            axis  = cur_sel.axis
# ref_axisでの値
            x0 = event.xdata
            y0 = event.ydata
# axisでの値
            prev_x = cur_sel.sel.target[0]
            prev_y = cur_sel.sel.target[1]
# axisでの値に戻す
            x1, y1 = self.convert_coord(x0, y0, ref_axis, axis)
            if x1 is None:
                return

            self.annotation.curpos = {"sel": self.annotation.sel, "event": event, "iaxis": iaxis, "axis": axis, "x": x1, "y": y1}

#前の座標からの移動量がepsより大きかったら更新
            r2 = pow((prev_x - x1) / cur_sel.x, 2.0) + pow((prev_y - y1) / cur_sel.y, 2.0)
            if r2 > eps:
                cur_sel.sel.annotation.set_visible(False)
                self.annotation.fig.canvas.draw_idle()

            self.annotation.sel = None

#クリックしたときconsoleに表示するフォーマット
        def inf_format(plotevent, sel, iline, idata, label, inf, x = 0, y = 0, opt = None, format = None):
            if inf is None:
                return

            args = {"label": label, "iline": iline, "idata": idata}
            for key, val in inf.items():
                if type(val) is str:
                    args[key] = val
                else:
                    try:
                        args[key] = val[idata]
                    except:
                        args[key] = val

            text = format.format(**args)
            print(text)

#マウスがhoverしたときannotationに表示するフォーマット
        def annotation_format(plotevent, sel, iline, idata, label, inf, x, y, opt = None, format = None):
            args = {"label": label, "iline": iline, "idata": idata, "x": x, "y": y}
            for key, val in inf.items():
                if type(val) is str:
                    args[key] = val
                else:
                    try:
                        args[key] = val[idata]
                    except:
                        args[key] = val

            if format is not None:
                text = format.format(**args)
            else:
                text = f"{label}".strip() + ": (" + f"{x:10.4g}".strip() + ", "  + f"{y:10.4g}".strip() + ")"

            sel.annotation.set(text = text)
#                sel.annotation.set(text = text, position = (x0, y0))
#            print(text)

        def on_click_default(event):
            if not self.annotation.is_active: return

            if event.button == 1 or event.button == 3: # 左クリックか右クリック
                if event.xdata is None or event.ydata is None: return

#                curpos = self.annotation.curpos
                iline = None
                label = None
                format = None
                sel = self.annotation.sel
                if sel is None:
                    idata = None
#                    return
                else:
                    if sel.artist in self.annotation.lines:
                        iline = self.annotation.lines.index(sel.artist)
                        if iline is None:
                            pass
#                            return

                    idata = int(sel.index + 0.499999)
                    label  = self.annotation.labels[iline]
                    format = self.annotation.inf_formats[iline]

                inf = None
                if iline is not None and len(self.annotation.inf_list) > iline:
                    if event.inaxes != self.annotation.axes[iline] and event.inaxes != self.annotation.ref_axes[iline]:
                        print(f"Warning in tkplotevent.on_click_default(): axis is not matched with line#{iline}")
                        return

                    inf = self.annotation.inf_list[iline]
                    if inf is None:
                        print("no inf")
#                        return

                if format is None:
                    format = self.annotation.inf_formats[0]

                opt = tkParams()
                opt.event = event
                opt.sel = None
                if type(format) is str:
                    inf_format(self, sel, iline, idata, label, inf, x = event.xdata, y = event.ydata, opt = opt, format = format)
                elif format is None and inf is not None:
                    for key, val in inf.items():
                        print(f"{key}: {val[idata]}")
                elif format is not None:
                     format(self, sel, iline, idata, label, inf, x = event.xdata, y = event.ydata, opt = opt)
                else:
                    print(f"Warning intkplotevent.on_click_default(): No data is selected (format is None and inf is None)")
                
        if on_click is None:
            on_click = on_click_default

        if on_mouse_move is None:
            on_mouse_move = on_mouse_move_annotate

        self.annotation.fig.canvas.mpl_connect('button_press_event', on_click)
        fig.canvas.mpl_connect('motion_notify_event', on_mouse_move)

#        @self.cursor.connect("add")
        def on_add(sel, print_level = print_level):
            if self.annotation.is_active:
                sel.annotation.set_visible(True)
            else:
                sel.annotation.set_visible(False)
                return
            
            self.annotation.sel = sel
            cur_sel = self.annotation
#            iaxis, axis = self.find_target_axes(sel, axes = self.annotation.axes)

            iline = cur_sel.lines.index(sel.artist)
            idata = int(sel.index + 0.4999)
            cur_sel.axis = cur_sel.axes[iline]

            x0 = cur_sel.sel.target[0]
            y0 = cur_sel.sel.target[1]

            sel.annotation.arrow_patch.set_alpha(0)

            xlist = self.annotation.x_lists[iline]
            if type(xlist) is str:
                self.annotation.x = xlist
            else:
                try:
                    self.annotation.x = xlist[idata]
                except:
                    self.annotation.x = xlist

            ylist = self.annotation.x_lists[iline]
            if type(ylist) is str:
                self.annotation.y = ylist
            else:
                try:
                    self.annotation.y = ylist[idata]
                except:
                    self.annotation.y = ylist

            if self.annotation.y <= 0.0:
                self.annotation.y = 1.0

            label  = self.annotation.labels[iline]
            format = self.annotation.annotation_formats[iline]
            inf    = self.annotation.inf_list[iline]
            opt = tkParams()
            opt.event = None
            opt.sel = sel
            if format is None:
                return
            if type(format) is str:
                text = annotation_format(self, sel, iline, idata, label, inf, x0, y0, opt = opt, format = format)
            else:
                text = format(self, sel, iline, idata, label, inf, x0, y0, opt = opt)

#        self.cursor = mplcursors.cursor(self.annotation.lines, hover = True)

        def refresh_cursors(lines = None):
            if lines is not False:
                lines = self.annotation.lines

            if self.cursor:
                self.cursor.remove()

            self.cursor = mplcursors.cursor(lines, hover = True)
            self.annotation.id_add_connection = \
                    self.cursor.connect("add", lambda sel: on_add(sel, print_level = print_level))

        if self.annotation.refresh_cursors is None:
            self.annotation.refresh_cursors = refresh_cursors
            refresh_cursors()

#        self.annotation.id_add_connection = \
#                   self.cursor.connect("add", lambda sel: on_add(sel, print_level = print_level))

    def register_follow_mouse_event(self, fig, activate = True, on_mouse_move = None, print_level = 0):
        self.follow_mouse.fig = fig
        self.follow_mouse.gco = None
        self.follow_mouse.is_active = activate

        def activate_func(flag = True):
            self.follow_mouse.is_active = flag

        self.follow_mouse.activate = activate_func

        def on_mouse_move(event):
            if not self.follow_mouse.is_active: 
                return

            x = event.xdata
            y = event.ydata
            for ia, ax in enumerate(self.follow_mouse.axes):
                if event.inaxes == ax:
#            ln_1.set_data([x], [y])
                    if self.follow_mouse.marker[ia] is None:
                        self.follow_mouse.vlines[ia].set_xdata([x])
                        self.follow_mouse.hlines[ia].set_ydata([y])
                    else:
                        self.follow_mouse.vlines[ia].set_data(x, y)

            if self.follow_mouse.gco is not None:
                self.follow_mouse.gco.set_data(x, y)

            plt.draw()

        self.follow_mouse.axes = []
        self.follow_mouse.marker = []
        self.follow_mouse.vlines = []
        self.follow_mouse.hlines = []
        def add(ax, line = 'dashed', width = 0.5, marker = None, size = 3.0, color = 'red'):
            self.follow_mouse.axes.append(ax)
            self.follow_mouse.marker.append(marker)
            if marker is None:
                vline = ax.axvline(0, linestyle = line, linewidth = width, color = color)
                hline = ax.axhline(0, linestyle = line, linewidth = width, color = color)
                self.follow_mouse.vlines.append(vline)
                self.follow_mouse.hlines.append(hline)
            else:
                ln, = ax.plot([], [], marker, markersize = size, markeredgecolor = color, markerfacecolor = color)
                self.follow_mouse.vlines.append(ln)

        self.follow_mouse.add = add

        if on_mouse_move is None:
            on_mouse_move = on_mouse_move

        fig.canvas.mpl_connect('motion_notify_event', on_mouse_move)

    def register_event(self, fig, event = "button_press_event", callback = None):
        self.fig = fig
        if callback is None:
            callback = lambda event: self.onclick(event)

        fig.canvas.mpl_connect(event, callback)
#            fig_contour.canvas.mpl_connect("button_press_event", onclick)

    def register_click(self, fig, event = "button_press_event", callback = None):
        if callback is None:
            callback = lambda event: self.onclick(event)
            
        self.register_event(fig, event = event, callback = callback)

    def register_pick(self, fig, event = "pick_event", callback = None):
        if callback is None:
            callback = lambda event: self.onpick(event)
            
        self.register_event(fig, event = event, callback = callback)

    def register_redraw(self, fig, event = "draw_event", callback = None):
        if callback is None:
            callback = lambda event: self.ondraw(event)
            
        self.register_event(fig, event = event, callback = callback)

    def register_key(self, fig, event = "key_press_event", callback = None):
        if callback is None:
            callback = lambda event: self.onkey(event)
            
        self.register_event(fig, event = event, callback = callback)

    def onkey(self, event):
#        print('you pressed', event.key, event.xdata, event.ydata)
        if event.key == 'q':
            plt.close(event.canvas.figure)
        sys.stdout.flush()

    def onclick(self, event):
#            print("click", event.inaxes)
        itargets, hit_axisinfs = self.find_target_axes(event)
        if itargets is None:
            print("Warning in tkplotevent.onclick(): Can not find axes")
            return

        iinf, idata, idx, r2 = self.find_nearest_data(event.xdata, event.ydata, itargets, hit_axisinfs)
        self.display_data(iinf, idata, idx, self.axis_inf[iinf])

    """
#selected_point = None
    def onclick2(self, event):
        global selected_point  # クリックイベント内で変数を変更するために必要です
        recover = False
    
        if event.inaxes == ax:
            x_click, y_click = event.xdata, event.ydata
#            print("clicked")
            for i in range(len(x)):
                if abs(x[i] - x_click) < 0.1 and abs(y[i] - y_click) < 0.1:
                    print("hit")
                    # クリックしたデータポイントを太い線に変更します
                    line.set_marker('o')
                    line.set_linestyle('-')
                    line.set_markersize(8)
                    line.set_linewidth(1)
                    line.set_color('red')
                    selected_point = line  # 選択されたデータポイントを記憶します
                    fig.canvas.draw()
                    return
                else:
                    recover = True

# 新しい場所をクリックした場合、以前の選択を元に戻す
        if recover and selected_point:
            selected_point.set_marker('o')
            selected_point.set_linestyle('-')
            selected_point.set_markersize(8)
            selected_point.set_linewidth(1)
            selected_point.set_color('blue')
            selected_point = None
            fig.canvas.draw()
    """


    def onpick(self, event):
        if hasattr(event.artist, 'get_text'):
            clicked_label = event.artist.get_text()
        elif hasattr(event.artist, 'get_label'):
            clicked_label = event.artist.get_label()
        else:
            return

        print(f"Data [{clicked_label}] is clicked")

    def ondraw(self, event):
        print("draw")
        """
#       legend.set_picker(True)
        for text in legend.get_texts():
            text.set_picker(True)

#        fig.canvas.draw()
#       plt.pause(1.0e-5)
#        fig.canvas.mpl_connect('pick_event', on_legend_label_click)
        for text in legend.get_texts():
            text.set_picker(10)
        for line in legend.get_lines():
            line.set_picker(10)
            line.set_pickradius(5)  # クリックを検出する半径を設定
        """

# マウスクリック時に最近接データを検索
    def find_nearest_data(self, x, y, itarget_list, hit_axisinf_list):
        minr2     = 1.0e300
        iinf_hit  = None
        idata_hit = None
        ihit      = None
        for iaxis in range(len(hit_axisinf_list)):
            hit_axisinf = hit_axisinf_list[iaxis]
            axis        = hit_axisinf['axis']
            axis_scale  = hit_axisinf.get('axis_scale', None)
            data        = hit_axisinf['data']
            plot_type   = hit_axisinf['plot_type']

            xlim0   = axis.get_xlim()
            ylim0   = axis.get_ylim()
            xrange0 = xlim0[1] - xlim0[0]
            yrange0 = ylim0[1] - ylim0[0]
            if axis_scale is not None:
                xlim    = axis_scale.get_xlim()
                ylim    = axis_scale.get_ylim()
            else:
                xlim    = axis.get_xlim()
                ylim    = axis.get_ylim()

            if type(data) is not list:
                data = [data]

            xrange = xlim[1] - xlim[0]
            yrange = ylim[1] - ylim[0]
            if type(data) is not list:
                data = [data]

# x,yは検出axisの座標で与えられる。検出axisとaxis_scaleが異なる場合、変換する必要がある
#            print("axis=", iaxis, axis, axis_scale, axis is not axis_scale)
            if axis_scale and axis is not axis_scale:
#                print("x, y converted from ", xi, yi, " to ", end = '')
                x0 = xlim[0] + xrange / xrange0 * (x - xlim0[0])
                y0 = ylim[0] + yrange / yrange0 * (y - ylim0[0])
#                print(x0, y0)
            else:
                x0 = x
                y0 = y

            for idata in range(len(data)):
                xdata = data[idata].xdata
                ydata = data[idata].ydata
#                xdata = data[idata].get_xdata()
#                ydata = data[idata].get_ydata()
                for i in range(len(xdata)):
                    xi = xdata[i]
                    yi = ydata[i]
#                    print(f"i={i:4} xi={xi:10.3g} yi={yi:10.3g} x0={x0:10.3g} y0={y0:10.3g}")

                    if xi is None or yi is None:
                        continue

                    if self.distance == 'r':
                        dx = (x0 - xi) / xrange
                        dy = (y0 - yi) / yrange
                        r2 = dx * dx + dy * dy
                    elif self.distance == 'x':
                        dx = (x0 - xi) / xrange
                        r2 = dx * dx
                    elif self.distance == 'y':
                        dy = (y0 - yi) / yrange
                        r2 = dy * dy
                
                    if minr2 > r2:
                        minr2 = r2
#                        print("\n**iaxis is updated:", iaxis, itarget_list[iaxis])
                        iinf_hit  = itarget_list[iaxis]
                        idata_hit = idata
                        ihit  = i
#                    print("  => ", minr2, ihit)

        if iinf_hit is None:
            print("")
            print(f"Error: Invalid type [{plot_type}] in find_nearest_data()")
            return None, None, None, None

        return iinf_hit, idata_hit, ihit, minr2


# 最近接データの情報をコンソールに表示
    def display_data(self, iinf, idata, idx, hit_axisinf = None):
        if hit_axisinf is None:
            hit_axisinf = self.axisinf[iinf]

        plot_type = hit_axisinf['plot_type']
        label  = hit_axisinf['label']
        axis   = hit_axisinf['axis']
        xlabel = axis.xlabel
        ylabel = axis.ylabel
        data = hit_axisinf['data']
        if type(data) is list:
            data = data[0]

        print("")
        if plot_type == '2D' or plot_type == 'plot' or plot_type == 'scatter':
            xdata   = data.xdata
            ydata   = data.ydata
            xlist   = hit_axisinf.get('xlist', None)
            xlabels = hit_axisinf.get('xlabels', None)
#            print("xlist=", xlist, xlabels)
            if xlist is None or xlabels is None:
                print(f"clicked for the data [{label}] in the {xlabel} - {ylabel} plot (inf #{iinf}  idata #{idata}  data index #{idx})")
                print(f"  #{idx:4}: {xlabel:>12}: {xdata[idx]:<12.4g}  {ylabel:>12}: {ydata[idx]:<12.4g}")
            else:
                print(f"clicked for the data [{label}] in the {xlabel} - {ylabel} plot (inf #{iinf}  data index #{idx})")
                print(f"  #{idx:4}: {xlabel:>5}: {xdata[idx]:5}  {ylabel:>12}: {ydata[idx]:<12.4g}")
                nx = len(xlabels)
                for i in range(nx):
                    print(f"    {i}: ", end = '')
                    if type(xlist[i][idx]) is int:
                        print(f"{xlabels[i]:>10}: {xlist[i][idx]:<10d}  ")
                    elif type(xlist[i][idx]) is float or isinstance(xlist[i][idx], np.float64):
                        print(f"{xlabels[i]:>10}: {xlist[i][idx]:<10.4g}  ")
                    else:
                        print(f"{xlabels[i]:>10}: {xlist[i][idx]:<10}  ")

