import os

try:
    import tklib.tkimport as imp
except Exception as e:
    print()
    print("######################################################################")
    print("###########  ERROR ERROR ERROR ERROR ERROR ERROR #####################")
    print("######################################################################")
    print(f"# Failed to import [tklib.tkimport] module ({e}).")
    print(f"#  Add [tkProg]{os.sep}tklib{os.sep}python to PYTHONPATH variable")
    print(f"#  Current PYTHONPATH:", sys.path)
    print("######################################################################")
    input("Press ENTER to terminate>>")
    exit()
    
mplcursors = imp.import_lib("mplcursors", stop_by_error = False)
np         = imp.import_lib("numpy",      stop_by_error = False)
mpl        = imp.import_lib("matplotlib", stop_by_error = False)
imp.messages(stop_by_error = True)

import sys
#import numpy as np
from numpy import sin, cos, tan, pi, exp, log, sqrt, log10
from matplotlib import pyplot as plt
from matplotlib import gridspec


from tklib.tkutils import terminate, pint, pfloat, getarg, getintarg, getfloatarg, analyze_varstr, save_csv
from tklib.tkutils import is_exist, is_file, is_dir
from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile

from tklib.tkvariousdata import tkVariousData
from tklib.tksci.tkmatrix import make_matrix1
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me, Ea_Arrhenius, kayserToHz
from tklib.tkcrystal.tkalamode import read_input, read_taudata, read_log, read_kpoints, read_properties, cal_contrib, convert_k_list, convert_k
from tklib.tkgraphic.tkplotevent import tkPlotEvent, RangeSelector
from tklib.tkgui.tksimple_gui import get_window_from_plt, CustomDialog_with_config


"""
Plot ALMODE output
"""

usage_str = '''
f"(i) usage: python {sys.argv[0]} plot fig_type plot_type target kpoint_log tau_dat kmin_plot kmesh_plot output_path(.xlsx)"
f"           fig_type : 'simple', 'sym', 'fat', 'contour', '3d'"
f"           plot_type: 'linear', 'log'"
f"           target   : 'tau', 'v', 'l', 'kxx', 'kxz', etc"
f"       ex: python {sys.argv[0]} {cparams.mode} {cparams.figtype} {cparams.plottype} {cparams.target} {cparams.kplog_file} {cparams.taudat_file} {cparams.kmin_plot} {cparams.kmesh_plot} {cparams.outfile}"
f"(ii) usage: python {sys.argv[0]} make_excel input_path kpoint_log tau_dat output_path(.xlsx)"
f"       ex: python {sys.argv[0]} {cparams.mode} {cparams.infile} {cparams.kplog_file} {cparams.taudat_file} {cparams.outfile}"
f"(iii) usage: python {sys.argv[0]} plot_contrib target input_path kpoint_log tau_dat output_path(.xlsx)"
f"       ex: python {sys.argv[0]} {cparams.mode} {cparams.target} {cparams.infile} {cparams.kplog_file} {cparams.taudat_file} {cparams.outfile}"
'''[1:-1]



def initialize():
#================================
# Global variables
#================================
    app                    = tkApplication(usage_str  = usage_str)
    app.config.plugin_dir  = app.replace_path(None, template = ["{dirname}", "plugin/plot_ALAMODE"])
    appconfig_path, config = app.read_app_config(print_level = 1)
    argv, narg             = app.get_argv()

    cparams             = app.get_params()
    cparams.debug       = 0
    cparams.print_level = 0

    cparams.colors = ['black', 'red', 'blue', 'darkgreen', 'darkorange', 'navy', 
              'slategray', 'hotpink', 'olive', 'chocolate', 'magenta', 
              'green']

    cparams.figsize         = [6, 8]
    cparams.legend_fontsize = 3
    cparams.fontsize        = 16
    cparams.alpha           = 0.5

# mode: 
    cparams.mode    = 'plot'

    cparams.infile      = 'rta.in'
    cparams.kplog_file  = 'k774.log'
    cparams.taudat_file = 'tau_300K-k774.dat'
    cparams.outfile     = 'db-out.xlsx'
    
    cparams.kmin_plot = 0.0
    cparams.kmesh_plot = ''

# figtype: 'simple', 'sym', 'fat', 'contour', '3d'
    cparams.figtype = 'sym'
# plottype: 'linear', 'log'
    cparams.plottype = 'log'
# target: 'tau', 'l', 'v', 'kxx'
    cparams.target  = 'tau'

    cparams.fatplot_maxsize = 5.0
    cparams.nfmesh   = 301
    cparams.nymesh   = 101
    cparams.wDOS     = 0.02
    cparams.wy_ratio = 0.002
    cparams.ymin_contour = 1.0e-5

    return app, cparams

#=============================
# Treat argments
#=============================
def usage(app):
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def update_vars(app, cparams, apply_default = True):
    argv = app.argv
#    if len(argv) <= 1:
#        app.terminate(usage = usage)

    cparams.mode = getarg(1, cparams.mode)
    if cparams.mode == 'plot':
        cparams.figtype     = getarg(2, cparams.figtype)
        cparams.plottype    = getarg(3, cparams.plottype)
        cparams.target      = getarg(4, cparams.target)
        cparams.kplog_file  = getarg(5, cparams.kplog_file)
        cparams.taudat_file = getarg(6, cparams.taudat_file)
        cparams.kmin_plot   = getfloatarg(7, cparams.kmin_plot)
        cparams.kmesh_plot  = getarg(8, cparams.kmesh_plot)
        cparams.outfile     = getarg(9, cparams.outfile)
    elif cparams.mode == 'plot_contrib':
        cparams.target      = getarg(2, cparams.target)
        cparams.infile      = getarg(3, cparams.infile)
        cparams.kplog_file  = getarg(4, cparams.kplog_file)
        cparams.taudat_file = getarg(5, cparams.taudat_file)
        cparams.outfile     = getarg(6, cparams.outfile)
    elif cparams.mode == 'make_excel':
        cparams.infile      = getarg(2, cparams.infile)
        cparams.kplog_file  = getarg(3, cparams.kplog_file)
        cparams.taudat_file = getarg(4, cparams.taudat_file)
        cparams.outfile     = getarg(5, cparams.outfile)
    else:
        app.terminate(f"Error in update_vars: Invalid mode [{cparams.mode}]", usage = usage)
        
#    header, ext              = os.path.splitext(app.argv[0])
#    filebody                 = os.path.basename(header)
#    cparams.parameter_file   = filebody + '.in'

    header, ext              = os.path.splitext(app.argv[0])
    filebody                 = os.path.basename(header)
    cparams.parameter_file   = filebody + '.in'
    cparams.parameter_bkfile = filebody + '.in.bak'

def make_excel(app, cparams):
    print("")
    print(f"mode: {cparams.mode}")
    print(f"input file     : {cparams.infile}")
    print(f"kpoint log file: {cparams.kplog_file}")
    print(f"tau data file  : {cparams.taudat_file}")

    print("")
    print(f"Read input file from {cparams.infile}")
    inf = read_input(cparams.infile)
    if inf is None:
        print(f"Warning: Can not read [{cparams.infile}]. Skip")
    else:
        for key, val in inf.items():
            print(f"  {key}: {val}")

    sn_list, ik_list, is_list, kx_list, ky_list, kz_list, w_list, fk_list, fTHz_list, _tau_list, v_list, MFP_list, \
                 kxx_list, kxy_list, kxz_list, kyx_list, kyy_list, kyz_list, kzx_list, kzy_list, kzz_list, inf_log \
                 = read_properties(cparams.kplog_file, cparams.taudat_file)

    praij = inf_log['praij']
    kxc_list, kyc_list, kzc_list = convert_k_list(praij, kx_list, ky_list, kz_list)

    labels    = ['sn', 'ik', 'is', 'kx', 'ky', 'kz', 'kx,c', 'ky,c', 'kz,c', 
                 'w', 'f(cm-1)', 'f(THz)', 'tau(ps)', 'v(m/s)', 'MFP(nm)', 
                 'kxx(W/mK)', 'kxy', 'kxz', 'kyx', 'kyy', 'kyz', 'kzx', 'kzy', 'kzz']
    data_list = [sn_list, ik_list, is_list, kx_list, ky_list, kz_list, kxc_list, kyc_list, kzc_list, 
                 w_list, fk_list, fTHz_list, _tau_list, v_list, MFP_list,
                 kxx_list, kxy_list, kxz_list, kyx_list, kyy_list, kyz_list, kzx_list, kzy_list, kzz_list]

    print("")
    print(f"Save to [{cparams.outfile}]")
    tkVariousData().to_excel(cparams.outfile, labels, data_list)

    app.terminate(pause = True)

def plot_contrib(app, cparams):
    print("")
    print(f"mode  : {cparams.mode}")
    print(f"target: {cparams.target}")
    print(f"input file     : {cparams.infile}")
    print(f"kpoint log file: {cparams.kplog_file}")
    print(f"tau data file  : {cparams.taudat_file}")

    print("")
    print(f"Read input file from {cparams.infile}")
    inf = read_input(cparams.infile)
    if inf is None:
        print(f"Warning: Can not read [{cparams.infile}]. Skip")
    else:
        for key, val in inf.items():
            print(f"  {key}: {val}")

    sn_list, ik_list, is_list, kx_list, ky_list, kz_list, w_list, fk_list, fTHz_list, _tau_list, v_list, MFP_list, \
             kxx_list, kxy_list, kxz_list, kyx_list, kyy_list, kyz_list, kzx_list, kzy_list, kzz_list, inf_log \
                 = read_properties(cparams.kplog_file, cparams.taudat_file)

#    labels = ['sn', 'ik', 'is', 'kx', 'ky', 'kz', 'w', 'f(cm-1)', 'f(THz)', 'tau(ps)', 'v(m/s)', 'MFP(nm)', 
#              'kxx(uW/mK)', 'kxy(uW/mK)', 'kxz(uW/mK)', 'kyx(uW/mK)', 'kyy(uW/mK)', 'kyz(uW/mK)', 'kzx(uW/mK)', 'kzy(uW/mK)', 'kzz(uW/mK)']
#    data_list = [sn_list, ik_list, is_list, kx_list, ky_list, kz_list, w_list, fk_list, fTHz_list, _tau_list, v_list, MFP_list,
#                 kxx_list, kxy_list, kxz_list, kyx_list, kyy_list, kyz_list, kzx_list, kzy_list, kzz_list]
#    print("")
#    print(f"Save to [{cparams.outfile}]")
#    tkVariousData().to_excel(cparams.outfile, labels, data_list)

    kappa_labels = ['kxx(uW/mK)', 'kxy(uW/mK)', 'kxz(uW/mK)', 'kyx(uW/mK)', 'kyy(uW/mK)', 'kyz(uW/mK)', 'kzx(uW/mK)', 'kzy(uW/mK)', 'kzz(uW/mK)']
    kappa_list = [kxx_list, kxy_list, kxz_list, kyx_list, kyy_list, kyz_list, kzx_list, kzy_list, kzz_list]
    index = None
    for idx, l in enumerate(kappa_labels):
        if cparams.target in l:
            index = idx
            break
    if index is None:
        app.terminate(f"Error in plot_contrib(): Invalid target [{cparams.target}]", pause = True)

    prop_label = kappa_labels[index]
    prop_list  = kappa_list[index]

    print(f"index: {index}")
    print(f"labels: {prop_label}")

    xp_list, xlogp_list, nhit_list, wprop_list, contrib_list \
        = cal_contrib(prop_list, prop_label, w_list, nprop_mesh = 501)

    print("")
    print(f"Save contribution to [{cparams.outfile}]")
    tkVariousData().to_excel(cparams.outfile, 
                [prop_label, "weight", f"w * {prop_label}", "contribution"], 
                [xp_list, nhit_list, wprop_list, contrib_list])

#plot
    print()
    print("Plot")
    fig, ax = plt.subplots(2, 1, sharex = 'col', figsize = (10, 8))

    plot_event = tkPlotEvent(plt)
    root = get_window_from_plt(plt)
    plot_event = tkPlotEvent(plt)
    plot_event.prepare_annotation()
    plot_event.prepare_popup_menu(fig, parent = root)
    plot_event.prepare_move_text(fig)

    selector0 = RangeSelector('', ax[0], color = 'green', print_level = 0)
    selector1 = RangeSelector('', ax[1], color = 'blue', print_level = 0)

# popup_menuをpluginから読み込む
    popup_menu = plot_event.popup_menu.menu

    vars = cparams.copy()
    vars.caller = "contribution"
    vars.plot_event = plot_event
    vars.selector0 = selector0
    vars.selector1 = selector1

    if len(app.config.plugin_dir) >= 1 and app.config.plugin_dir[0] not in "/\\":
        plugin_dir = app.replace_path(None, template = ["{dirname}", app.config.plugin_dir])
    print()
    print(f"Read plugins from [{plugin_dir}]")
    module_names, modules = app.load_modules(plugin_dir, "*.py", target = "popup_menu", sort = True, is_print = True)
    for m in modules:
        if hasattr(m, "add_popup_menu"):
            m.add_popup_menu(popup_menu, app = app, vars = vars, parent = root)


    ax2 = ax[0].twinx()
    ax[0].tick_params(labelsize = cparams.fontsize)
    ax2.tick_params(labelsize = cparams.fontsize)
    line, = ax[0].plot(xp_list, nhit_list, label = 'weight', color = 'black', linewidth = 0.5)
    plot_event.annotation.add_line('weight', ax[0], ax2, xp_list, nhit_list, line)
    label = f'w*{prop_label}'
    line, = ax2.plot(xp_list, wprop_list, label = label, color = 'blue', linewidth = 0.5)
    plot_event.annotation.add_line(label, ax2, ax2, xp_list, wprop_list, line)
#    ax[0].set_xlabel(f'{prop_label}', fontsize = cparams.fontsize)
#    ax[0].set_xscale('log')
    ax[0].set_ylabel('weight', fontsize = cparams.fontsize)

    ax2.set_ylabel(label, color = 'blue', fontsize = cparams.fontsize)
    ax2.tick_params(axis='y', colors='blue')
#    ax[0].legend(fontsize = cparams.legend_fontsize)

    h1, l1 = ax[0].get_legend_handles_labels()
    h2, l2 = ax2.get_legend_handles_labels()
    _legend = ax[0].legend(h1 + h2, l1 + l2)
    _legend.set_draggable(True)

    ax[1].tick_params(labelsize = cparams.fontsize)
    line, = ax[1].plot(xp_list, contrib_list, label = 'contribution')
    plot_event.annotation.add_line('contribution', ax[1], ax[1], xp_list, contrib_list, line)
    ax[1].set_xlabel(f'{prop_label}', fontsize = cparams.fontsize)
    ax[1].set_ylabel('contribution', fontsize = cparams.fontsize)
    ax[1].set_ylim([1.0e-3, 1.1])
    ax[1].set_xscale('log')
    ax[1].set_yscale('log')
#   _legend = ax[1].legend(fontsize = cparams.legend_fontsize)
#    _legend.set_draggable(True)

    plot_event.register_annotation_event(fig, activate = True, print_level = 0)
    plot_event.register_move_text_event(fig, activate = False)
    plot_event.register_popup_menu_event()

    plt.tight_layout()
    plt.pause(1.0e-5)

    app.terminate(pause = True)

def plot(app, cparams):
    print("")
    print(f"mode: {cparams.mode}")
    print(f"kpoint log file: {cparams.kplog_file}")
    print(f"tau data file: {cparams.taudat_file}")
    print(f"kappa min to plot: {cparams.kmin_plot}")
    print(f"k mesh to plot   : {cparams.kmesh_plot}")
    print(f"Plot type        : {cparams.plottype}")
    print(f"Target           : {cparams.target}")

    print("")
    print(f"Read k points from {cparams.kplog_file}")
    inf_log, kp_list, kp_list_ik = read_log(cparams.kplog_file)
    if kp_list is None:
        app.terminate(f"Error: Can not read [{cparams.kplog_file}].", pause = True)
    print("  # of k points: ", len(kp_list))
    print("  K points:")
    for i in range(len(kp_list)):
        l = kp_list[i]
        ik = l["ik"]
        kx = l["kx"]
        ky = l["ky"]
        kz = l["kz"]
        w  = l["w"]
        print(f"    {i:2d}: ik={ik:2d}: k=({kx:12.6f}, {ky:12.6f}, {kz:12.6f}) w={w:12.6f}")
    
    kmesh = cparams.kmesh_plot
    iklist = None
    if kmesh != '':
        print()
        print("k mesh selection:")

        dkx = 1.0e300
        dky = 1.0e300
        dkz = 1.0e300
        for i in range(len(kp_list)):
            l = kp_list[i]
            ik = l["ik"]
            kx = l["kx"]
            ky = l["ky"]
            kz = l["kz"]
            if i == 0:
                kx0 = kx
                ky0 = ky
                kz0 = kz
            else:
                dk = abs(kx - kx0)
                if 1.0e-3 < dk < dkx:
                    dkx = dk
                dk = abs(ky - ky0)
                if 1.0e-3 < dk < dky:
                    dky = dk
                dk = abs(kx - kz0)
                if 1.0e-3 < dk < dkz:
                    dkz = dk

        _aa = kmesh.split(',')
        if len(_aa) == 1:
            nx = int(_aa[0])
            ny = nx
            nz = nx
        elif len(_aa) == 2:
            nx = int(_aa[0])
            ny = nx
            nz = int(_aa[1])
        elif len(_aa) >= 3:
            nx = int(_aa[0])
            ny = int(_aa[1])
            nz = int(_aa[2])
        nx -= 1
        ny -= 1
        nz -= 1
        kxs = [0.5 / nx * i for i in range(-nx, 0, 1)] + [0.0] + [0.5 / nx * i for i in range(1, nx + 1)]
        kys = [0.5 / ny * i for i in range(-ny, 0, 1)] + [0.0] + [0.5 / ny * i for i in range(1, ny + 1)]
        kzs = [0.5 / nz * i for i in range(-nz, 0, 1)] + [0.0] + [0.5 / nz * i for i in range(1, nz + 1)]
        print("nx=", nx+1, "kxs=", kxs)
        print("ny=", ny+1, "kys=", kys)
        print("nz=", nz+1, "kzs=", kzs)
        
        iklist = []
        for i in range(len(kp_list)):
            l = kp_list[i]
            ik = l["ik"]
            kx = l["kx"]
            ky = l["ky"]
            kz = l["kz"]

            included = False
            for kx0 in kxs:
                if abs(kx - kx0) < dkx:
                    included = True
                    break
            if not included:
#                print(" not included")
                continue

            included = False
            for ky0 in kys:
                if abs(ky - ky0) < dky:
                    included = True
                    break
            if not included:
#                print(" not included")
                continue

            included = False
            for kz0 in kzs:
                if abs(kz - kz0) < dkz:
                    included = True
                    break
            if not included:
#                print(" not included")
                continue

            if included:
                print(f"  {i:04d}: ik={ik:03d}: k=({kx:12.6f}, {ky:12.6f}, {kz:12.6f}):", end = '')
                print(" included")
                iklist.append(ik)
            else:
#                print("   not included")
                pass

        print("# of ik included:", len(iklist))
        print("  iklist=", iklist)
        print("  kxs=", kxs)
        print("  kys=", kys)
        print("  kzs=", kzs)
        print(f"  dk=({dkx}, {dky}, {dkz})")
            
    print("")
    print(f"Read tau data from {cparams.taudat_file}")
    tau_list = read_taudata(cparams.taudat_file, 
                    target = cparams.target, iklist = iklist)

    print("tau data:")
    for i in range(len(tau_list)):
        l = tau_list[i]
        kstr = '('
        for k in l["kappa"]:
            kstr += f"{k:8.3g},"
        kstr = kstr.rstrip(',') + ")"
        print(f'  ik={l["ik"]:03d} is={l["is"]:03d}: f={l["f"]:10.4g} tau={l["tau"]:10.4g} v={l["v"]:10.4g} MFP={l["l"]:10.4g} kappa={kstr}',  )

#    print("")
#    print(f"Save to [{cparams.outfile}]")
#    tkVariousData().to_excel(cparams.outfile, labels, data_list)

#====================================================
# Plot condition
#====================================================
    print("")
    print("Plot")
    print(f"Figure type: {cparams.figtype}")
    if (cparams.figtype == 'contour' or cparams.figtype == '3d') and cparams.plottype == 'log':
        print(f"  Plot type [{cparams.plottype}] is not allowed for figure type [{cparams.figtype}]")
        cparams.plottype = 'linear'
        print(f"  Plot type is changed to [{cparams.plottype}]")
        if 'k' not in cparams.target:
            print(f"  Target [{cparams.target}] is not allowed for figure type [{cparams.figtype}]")
            cparams.target = 'kxx'
            print(f"  Target is changed to [{cparams.target}]")
    print(f"Plot type  : {cparams.plottype}")
    print(f"Target: {cparams.target}")
    
    ncolors = len(cparams.colors)

#=========================
# Prepare for target
#=========================
    xx   = []
    yy   = []
    itau = []
    icolor = []
    nk = len(kp_list)
    print("nk=", nk)
    for ik in range(nk):
        xx.append([])
        yy.append([])
        itau.append([])
        icolor.append(ik % ncolors)

    if cparams.target == 'tau':
        key = 'tau'
        ylabel = '$\\tau$ (ps)'
        ylim0  = None
    elif cparams.target == 'v':
        key = 'v'
        ylabel = '$v_g$ (m/s)'
        ylim0  = 1.0e-2
    elif cparams.target == 'l':
        key = 'l'
        ylabel = 'MFP (nm)'
        ylim0  = 1.0e-3
    elif 'k' in cparams.target:
        kindex_dict = {'kxx': 0, 'kxy': 1, 'kxz': 2,
                       'kyx': 3, 'kyy': 4, 'kyz': 5, 
                       'kzx': 6, 'kzy': 7, 'kzz': 8
                      }
        key = 'kappa'
        index = kindex_dict[cparams.target]
        ylabel = '$\kappa_{' + cparams.target.replace('k', '') + '}$ ($\mu$W/m/K)'
        ylim0  = max([1.0e-6, cparams.kmin_plot])

    xmax = 0.0
    ymax = 0.0
    for i in range(len(tau_list)):
        l  = tau_list[i]
        ik = l["ik"] - 1
        x  = l["f"]
        if 'k' in cparams.target:
            v = l[key][index]
            if i == 0:
                print(f"** [{index}] is selected for the top figure")
        else:
            v = l[key]
            if i == 0:
                print(f"** {cparams.target} is selected for the top figure")

        if abs(v) > 0.0:
            xx[ik].append(x)
            yy[ik].append(v)
            itau[ik].append(i)
            if xmax < x:
                xmax = x
            if ymax < v:
                ymax = v

#=========================
# Build kappa contour data
#=========================
    fstep = xmax / (cparams.nfmesh - 1)
    if cparams.figtype == 'contour' or cparams.figtype == '3d':
        fstep   = xmax / (cparams.nfmesh - 1)
        ystep   = ymax / (cparams.nymesh - 1)
        wy      = ymax * cparams.wy_ratio
        ifrange = int(4.0 * cparams.wDOS / fstep + 1.0001)
        iyrange = int(4.0 * wy / ystep + 1.0001)
        af      = ifrange / 0.832554611
        ay      = iyrange / 0.832554611

        wtot = 0.0
        for i in range(-ifrange, ifrange + 1):
            for j in range(-iyrange, iyrange + 1):
                wtot += exp( -(i / af)**2 - (j / ay)**2 )

        print("")
        print("Contour parameters")
        print(f"  nfmesh   : {cparams.nfmesh}")
        print(f"  nymesh   : {cparams.nymesh}")
        print(f"  wf       : {cparams.wDOS}")
        print(f"  wy(ratio): {cparams.wDOS}")
        print(f"  wy       : {wy}")
        print(f"  ifrange  : {ifrange}")
        print(f"  iyrange  : {iyrange}")
        print(f"  wtot     : {wtot}")

        f_contour   = [i * fstep for i in range(cparams.nfmesh)]
        y_contour   = [i * ystep for i in range(cparams.nymesh)]
        z_contour   = np.zeros([cparams.nfmesh, cparams.nymesh])
        y_kappa     = np.zeros(cparams.nfmesh)
#        print("dim=", len(f_contour), len(y_contour), len(z_contour), len(z_contour[0]))
        kappamax = 0.0
        for i in range(len(tau_list)):
            l    = tau_list[i]

            if 'k' in cparams.target:
                v = l[key][index]
                kappalabel = '$\kappa_{' + cparams.target.replace('k', '') + '}$'
                if i == 0:
                    print(f"** kappa[{index}] is selected for the contour map")
            else:
                v = l['kappa'][0]
                kappalabel = '$\kappa_{xx}$'
                if i == 0:
                    print(f"** kxx is selected for the contour map")

            f    = l["f"]
            idxf = int(f / fstep)
            idxy = int(v / ystep)
            ik   = l["ik"] - 1
            w    = kp_list[ik]["w"]
#            print("i=", i, f, v, idxf, idxy)

            if kappamax < v:
                kappamax = v

            ifs = max([idxf - ifrange, 0])
            ife = min([cparams.nfmesh-1, idxf + ifrange])
            iys = max([idxy - iyrange, 0])
            iye = min([cparams.nymesh-1, idxy + iyrange])
#            print("i=", i, f, idxf, idxy, ifs, ife, iys, iye, w, v)
            for jf in range(ifs, ife + 1):
                for jy in range(iys, iye + 1):
                    dz = w * v * exp( -((jf - f / fstep) / af)**2 - ((jy - v / ystep) / ay)**2 ) / wtot
                    z_contour[jf][jy] += dz
#                    if dz < 0.0:
#                        print("negative dz: i=", i, f, idxf, idxy, ifs, ife, iys, iye, w, v, wtot)
#                    elif z_contour[jf][jy] < 0.0:
#                        print("negative z: i=", i, f, idxf, idxy, ifs, ife, iys, iye, w, vwtot)

        for jf in range(cparams.nfmesh):
            for jy in range(cparams.nymesh):
                if z_contour[jf][jy] < kappamax * cparams.ymin_contour:
                    z_contour[jf][jy] = 0.0
                else:
                    z_contour[jf][jy] = log(z_contour[jf][jy])
        
#=========================
# Build phonon DOS and kappa spectrum
#=========================
    ifrange = int(4.0 * cparams.wDOS / fstep + 1.0001)
    af = ifrange / 0.832554611
    wtot = 0.0
    for j in range(-ifrange, ifrange + 1):
        wtot += exp( -(j / af)**2 )
    print("DOS parameters")
    print(f"  nfmesh : {cparams.nfmesh}")
    print(f"  wDOS   : {cparams.wDOS}")
    print(f"  ifrange: {ifrange}")
    print(f"  wtot   : {wtot}")
    
    f_DOS   = [i * fstep for i in range(cparams.nfmesh)]
    y_DOS   = np.zeros(cparams.nfmesh)
    y_kappa = np.zeros(cparams.nfmesh)
    kappamax = 0.0
    for i in range(len(tau_list)):
        l    = tau_list[i]
        f    = l["f"]
        idxf = int(f / fstep)
        ik   = l["ik"] - 1
        w    = kp_list[ik]["w"]

        if 'k' in cparams.target:
            v = l[key][index]
            kappalabel = '$\kappa_{' + cparams.target.replace('k', '') + '}$'
            if i == 0:
                print(f"** kappa[{index}] is selected for the middle figure")
        else:
            v = l['kappa'][0]
            kappalabel = '$\kappa_{xx}$'
            if i == 0:
                print(f"** kxx is selected for the middle figure")

        if kappamax < v:
            kappamax = v

        ifs = max([idxf - ifrange, 0])
        ife = min([cparams.nfmesh-1, idxf + ifrange])
#        print("i=", i, f, idxf, ifs, ife, w, v)
        for j in range(ifs, ife + 1):
            coeff = w * exp( -((j - f / fstep) / af)**2 ) / wtot
            y_DOS[j]   += coeff
            y_kappa[j] += v * coeff

#=========================
# Plot
#=========================
    fig = plt.figure(figsize = cparams.figsize)
#    ax1 = fig.add_subplot(1, 1, 1)
    spec = gridspec.GridSpec(ncols = 1, nrows = 3, height_ratios=[4, 1, 1])
    if cparams.figtype == '3d':
        ax1 = fig.add_subplot(spec[0], projection = '3d')
    else:
        ax1 = fig.add_subplot(spec[0])

    ax2 = fig.add_subplot(spec[1], sharex = ax1)
    ax3 = fig.add_subplot(spec[2], sharex = ax1)

    ax1.tick_params(labelbottom = False)
    ax2.tick_params(labelbottom = False)
    ax1.tick_params(labelsize = cparams.fontsize)
    ax2.tick_params(labelsize = cparams.fontsize)
    ax3.tick_params(labelsize = cparams.fontsize)

# plot_event初期化
    root = get_window_from_plt(plt)
    plot_event = tkPlotEvent(plt)
    plot_event.prepare_annotation()
    plot_event.prepare_move_text(fig)
    plot_event.prepare_popup_menu(fig, parent = root)
    popup_menu = plot_event.popup_menu.menu

#    selector0 = RangeSelector('', ax1, color = 'green', print_level = 0)
#    selector1 = RangeSelector('', ax2, color = 'blue', print_level = 0)

    vars = cparams.copy()
    vars.caller = "plot"
    vars.plot_event = plot_event
#    vars.selector0 = selector0
#    vars.selector1 = selector1

    if len(app.config.plugin_dir) >= 1 and app.config.plugin_dir[0] not in "/\\":
        plugin_dir = app.replace_path(None, template = ["{dirname}", app.config.plugin_dir])
    print()
    print(f"Read plugins from [{plugin_dir}]")
    module_names, modules = app.load_modules(plugin_dir, "*.py", target = "popup_menu", sort = True, is_print = True)
    for m in modules:
        if hasattr(m, "add_popup_menu"):
            m.add_popup_menu(popup_menu, app = app, vars = vars, parent = root)

    def set_vars(vars, ihit, x, y, iline, idata, label, x_label, y_label):
        vars.target   = cparams.target
        vars.ihit     = ihit
        vars.tau_dict = tau_list[vars.ihit]
        kappa_list = vars.tau_dict["kappa"]
        for i, key in enumerate(["kxx", "kxy", "kxz", "kyx", "kyy", "kyz", "kzx", "kzy", "kzz"]):
            vars.tau_dict[key] = kappa_list[i]

        vars.ik       = vars.tau_dict["ik"]
        vars.kp_dict  = kp_list[vars.ik - 1]
        vars._is      = vars.tau_dict["is"]

        vars.label = label
        vars.iline = iline
        vars.idata = idata
        vars.x_label = x_label
        vars.y_label = y_label
    
        return vars

    def find_nearest_data(x, y):
        minr = 1.0e300
        ihit = None
        for i in range(len(tau_list)):
            l = tau_list[i]
            f = l["f"]
            if 'k' in cparams.target:
                v = l[key][index]
            else:
                v = l[key]

            if cparams.plottype == 'log':
                if v > 0.0:
                    r = abs(x - f) / xmax + abs(log(y) - log(v))
                else:
                    r = 1.0e300
            else:
                r = abs(x - f) / xmax + abs(y - v) / ymax
#            print("r=", x, y, f, v, r)
            if minr > r:
                minr = r
                ihit = i
        return ihit

    def inf_format_func(plotevent, sel, iline, idata, label, inf, x = 0, y = 0, opt = None):
        nonlocal vars

        if cparams.figtype == '3d':
            return

        if inf:
            x_label = inf["x_label"]
            y_label = inf["y_label"]
        else:
            x_label = "x"
            y_label = "y"

        ihit  = find_nearest_data(x, y)

        set_vars(vars, ihit, x, y, iline, idata, label, x_label, y_label)

#        print("inf_format: x,y=", x, y)
        tl    = tau_list[ihit]

        ik    = tl["ik"]
        _is   = tl["is"]
        fkayser = tl["f(cm-1)"]
        f     = tl["f"]
        tau   = tl["tau"]
        v     = tl["v"]
        MFP   = tl["l"]
        kappa = tl["kappa"]
        kl    = kp_list[ik - 1]
        kx    = kl["kx"]
        ky    = kl["ky"]
        kz    = kl["kz"]
        w     = kl["w"]
        c     = w / wtot
        praij = inf_log['praij']
        kxc, kyc, kzc = convert_k(praij, kx, ky, kz)
        print("")
        print(f"is={_is} ik={ik} q,prim=({kx:6.3f}, {ky:6.3f}, {kz:6.3f}) q,cart=({kxc:6.3f}, {kyc:6.3f}, {kzc:6.3f}) w={w:6.3f} wtot={wtot:6.3f} coeff=w/wtot={c:10.3g}")
        print(f"    f={fkayser:8.3f} cm-1/{f:8.3f} THz  tau={tau:10.3g} ps  v={v:10.3g} m/s  MFP={MFP:10.3g} nm")
        print(f"    kappa=|{kappa[0]:10.3g} {kappa[1]:10.3g} {kappa[2]:10.3g}| weighted=|{c*kappa[0]:10.3g} {c*kappa[1]:10.3g} {c*kappa[2]:10.3g}|")
        print(f"          |{kappa[3]:10.3g} {kappa[4]:10.3g} {kappa[5]:10.3g}|          |{c*kappa[3]:10.3g} {c*kappa[4]:10.3g} {c*kappa[5]:10.3g}|")
        print(f"          |{kappa[6]:10.3g} {kappa[7]:10.3g} {kappa[8]:10.3g}|          |{c*kappa[6]:10.3g} {c*kappa[7]:10.3g} {c*kappa[8]:10.3g}|")

    def annotation_format_func(plotevent, sel, iline, idata, label, inf, x = 0, y = 0, opt = None):
        nonlocal vars
#        print("annotation_format: x,y=", x, y)

        ihit  = find_nearest_data(x, y)
        x_label = inf["x_label"]
        y_label = inf["y_label"]
        set_vars(vars, ihit, x, y, iline, idata, label, x_label, y_label)

        _x = inf["x"][idata]
        _y = inf["y"][idata]
        ik = inf["ik"]
        kx = inf["kx"]
        ky = inf["ky"]
        kz = inf["kz"]

        text = f"{label}: line#{iline} data#{idata} ik={ik}:\n" \
             + f" k=({kx:6.3g},{ky:6.3g},{kz:6.3g})\n" \
             + f" {x_label}={_x:10.4g}  {y_label}={_y:10.4g}"

        sel.annotation.set(text = text)

    if cparams.figtype == 'sym':
        for ik in range(nk):
            l = kp_list[ik]
            kx = l["kx"]
            ky = l["ky"]
            kz = l["kz"]

            color = cparams.colors[icolor[ik]]
            if kx == 0.0 and ky == 0.0 and kz == 0.0:
                marker = 'D'
                markersize = 7.0
            elif kx == ky and ky == kz:
                marker = '^'
                markersize = 3.0
            elif kx == ky or ky == kz or kz == kx:
                marker = 'v'
                markersize = 5.0
            else:
                marker = 'o'
                markersize = 3.0

            label = f"({kx:6.3f},{ky:6.3f},{kz:6.3f})"
            line, = ax1.plot(xx[ik], yy[ik], label = label, linestyle = '', 
                        marker = marker, markersize = markersize, alpha = cparams.alpha,
                        markerfacecolor = color, markeredgecolor = color)

            plot_event.annotation.add_line(label, ax1, ax1, xx[ik], yy[ik], line,
                    inf_list = {"x_label": "f (THz)", "y_label": cparams.target,
                                "x": xx[ik], "y": yy[ik],
                                "ik": ik, "kx": kx, "ky": ky, "kz": kz, "kp_list": kp_list},
                    annotation_format = annotation_format_func, inf_format = inf_format_func)

    elif cparams.figtype == 'fat' or cparams.figtype == 'contour':
        for i in range(len(tau_list)):
            tl   = tau_list[i]
            f    = tl["f"]
            if 'k' in cparams.target:
                kv = tl[key][index]
                v  = kv
            else:
                kv = tl['kappa'][0]
                v  = tl[key]
                
            if v == 0.0:
                continue

            marker = 'o'
            markersize = max([cparams.fatplot_maxsize * log(kv) / log(kappamax), 0.2])
#            print("i=", i, f, v, kv, kappamax, markersize)
            color = 'black'
            label = f"kv={kv:6.3f}  v={kv:6.3f}"
            line, = ax1.plot(f, v, linestyle = '', alpha = cparams.alpha,
                        marker = marker, markersize = markersize, markerfacecolor = color, markeredgecolor = color)

        if cparams.figtype == 'contour':
                ax1.contourf(f_contour, y_contour, z_contour.T, levels = 20, cmap = 'Oranges')

    elif cparams.figtype == '3d':
        X, Y = np.meshgrid(f_contour, y_contour)
        ax1.plot_surface(X, Y, z_contour.T, cmap = 'Oranges')
        ax1.set_xlabel("$f$ (THz)", fontsize = cparams.fontsize)
        ax1.set_zlabel(kappalabel, fontsize = cparams.fontsize)
        ax1.tick_params(labelbottom = True)

    else:
        color = cparams.colors[0]
        marker = 'o'
        markersize = 3.0

        for ik in range(nk):
            label = f"({kx:6.3f},{ky:6.3f},{kz:6.3f})"
            line, = ax1.plot(xx[ik], yy[ik], label = label, linestyle = '', 
                        marker = marker, markersize = markersize, alpha = cparams.alpha, 
                        markerfacecolor = color, markeredgecolor = color)

            plot_event.annotation.add_line(label, ax1, ax1, xx[ik], yy[ik], line)

    if cparams.plottype != 'linear':
        ax1.set_yscale(cparams.plottype)
    if ylim0 is not None:
        ax1.set_ylim([ylim0, ax1.get_ylim()[1]])
#    _legend = ax1.legend(fontsize = cparams.legend_fontsize, prop={"family":"MS Gothic"}, loc = 'upper left', bbox_to_anchor=(1, 1))
#    _legend.set_draggable(True)

    ax2.plot(f_DOS, y_kappa)
    ax3.plot(f_DOS, y_DOS)

    ax1.set_ylabel(ylabel, fontsize = cparams.fontsize)
    ax2.set_ylabel(kappalabel, fontsize = cparams.fontsize)
    ax3.set_xlabel("$f$ (THz)", fontsize = cparams.fontsize)
    ax3.set_ylabel("DOS", fontsize = cparams.fontsize)

    plt.tight_layout()

    plot_event.register_annotation_event(fig, activate = True, print_level = 0)
    plot_event.register_move_text_event(fig, activate = False)
    plot_event.register_popup_menu_event()

    use_pause = True
    
    if use_pause:
        plt.pause(0.1)
        print("Press ENTER to exit")
        input()
    else:
        print("")
        print("Close the graph window to terminate")
        plt.show()

    
def main(app, cparams):
    app, cparams = initialize()

#--infile引数をチェックし、読み込みファイル名を最初に確定する
    cparams.infile = app.check_arg('--infile', None, vartype = 'int')
    cparams.logfile = app.replace_path(cparams.infile)

    print(f"Open logfile [{cparams.logfile}]")
    app.redirect(targets = ["stdout", cparams.logfile], mode = 'w')
#    app.redirect(targets = ["stdout"], mode = 'a', 
#            redirect_traceback = True, output_traceback = 'stdout', display_type_traceback = 'colored')
#    app.redirect_exception()

#起動時引数で与えられたパラメータを設定
    update_vars(app, cparams,apply_default = True)

    print("")
    print( "==========================================================================")
    print( "==========================================================================")
    print(f"     Plot ALMODE output files")
    print( "==========================================================================")
    print( "==========================================================================")
    print(f"mode  : {cparams.mode}")
    print(f"infile: {cparams.infile}")

#パラメータファイルの読み込み
    print("")
    print("Read [{}]".format(cparams.parameter_file))
    app.inf = cparams.read_parameters(cparams.parameter_file)

#起動時引数で与えられたパラメータを再設定
#    app.force_given_args(cparams)
    update_vars(app, cparams,apply_default = True)
#    cparams.print_parameters()

    if cparams.mode == 'plot':
        plot(app, cparams)
    elif cparams.mode == 'plot_contrib':
        plot_contrib(app, cparams)
    elif cparams.mode == 'make_excel':
        make_excel(app, cparams)
    else:
        app.terminate(f"Error in main: Invalide mode [{cparams.mode}]", usage = usage)

    print("")
    print("Save parameters to [{}]".format(cparams.parameter_file))
    cparams.save_parameters(cparams.parameter_file, "Preferences")


    app.terminate("", usage = usage)


if __name__ == "__main__":
    app, cparams = initialize()
    main(app, cparams)

