import os
import sys
import tkinter as tk


from tklib.tkutils import print_data, pint, pfloat, index2val, pconv_by_type
from tklib.tkutils import terminate, pint, pfloat, getarg, getintarg, getfloatarg, analyze_varstr, save_csv
from tklib.tkutils import is_exist, is_file, is_dir
from tklib.tkparams import tkParams
from tklib.tkfile import tkFile
from tklib.tkapplication import tkApplication
from tklib.tkcrystal.tkalamode import read_input, read_taudata, read_log, read_kpoints, read_properties, cal_contrib, convert_k_list, convert_k

from tklib.tkgui.tksimple_gui import get_window_from_plt, CustomDialog_with_config

if __name__ != "__main__":
    import plot_ALAMODE


def save_anphon_input(outfile, inf):
    aij = inf["aij"]
    SCALE = inf["SCALE"]
    inf["CELL"] = f"  {SCALE:f}\n" \
         + f"  {aij[0][0]:g} {aij[0][1]:g} {aij[0][2]:g}" \
         + f"  {aij[1][0]:g} {aij[1][1]:g} {aij[1][2]:g}" \
         + f"  {aij[2][0]:g} {aij[2][1]:g} {aij[2][2]:g}" 
    inf["ANIME"] = f'{inf["kx"]} {inf["ky"]} {inf["kz"]}'
    
    text = '''&general
 PREFIX = {PREFIX}
 MODE = {MODE}
 NKD = {NKD}; KD = {KD}
 TOLERANCE = {TOLERANCE}
 FCSXML = {FCSXML}
 EMIN = {EMIN}; EMAX = {EMAX}
 DELTA_E = {DELTA_E}
 ISMEAR = {ISMEAR}
 NONANALYTIC = {NONANALYTIC}; BORNINFO = {BORNINFO}
/

&cell
{CELL}
/

&kpoint
2
1 1 1
/

&analysis
 ANIME_FRAMES = {ANIME_FRAMES} ; ANIME_CELLSIZE = {ANIME_CELLSIZE} ; ANIME_FORMAT = {ANIME_FORMAT}
 ANIME = {ANIME}
 PRINTEVEC = {PRINTEVEC}
/
'''.format(**inf)

    print()
    print(f"save anphon input to [{outfile}]")
    fout = tkFile(outfile, "w")
    if fout is None or fout.fp is None: return False
    fout.write(text)
    fout.close()
    
    return True


def run_anphon(app, outfile, logfile, replace_dict):
    ret = save_anphon_input(outfile, replace_dict)
    if not ret: return None

    cmd = f"anphon {outfile} 2>&1 | tee {logfile}"
    print()
    print(f"Run [{cmd}]")
#    ret = os.system(cmd)
    return ret

def execute_command(app = None, vars = None, parent = None, **kwargs):
#    print("runが選択されました")
    if not hasattr(vars, "target") or not hasattr(vars, "ihit"):
        print("** Warning in run.py:execute_command(): No data was selected.")
        return

    cfg = app.cfg
    print()
    print("infile:", cfg.infile)
    print("outfile:", cfg.outfile)
#    print("template dir:", app.cfg.template_dir)
    print("xyz dir:", cfg.xyz_dir)

    print()
    print(f"read log from [{cfg.infile}]")
    inf_log, kp_list, kp_list_ik = read_log(cfg.infile)
    if inf_log is None: return None

#    print("kp_list=", kp_list)
#    print("kp_list_ik=", kp_list_ik)
#    print("inf=", inf_log)

    print()
    tl = vars.tau_dict
#    print(f"tau_dict=", vars.tau_dict)
    kp = vars.kp_dict
    ik = tl['ik']
    kx = kp['kx']
    ky = kp['ky']
    kz = kp['kz']
    _is = tl['is']
    f = tl['f']
    print(f"  ik={ik}  k=({kx:6.3f}, {ky:6.3f}, {kz:6.3f})")
    print(f"  is={_is}")
    print(f"  f={f:10.4g} THz")
    print(f"  {vars.target}={tl[vars.target]:10.4g}")

    aij = inf_log["aij"]
    paij = inf_log["paij"]
    haij = inf_log.get("haij", None)

    print("supercell aij=", aij)
    print("primitive cell aij=", paij)
    print("harmonic supercell aij=", haij)

    print(f"target={vars.target}")
    print(f"ihit={vars.ihit}")

    vars.tollerance = 1.0e-3
    vars.cell_size = "5 5 5"
    vars.Emin = 0.0
    vars.Emax = 2000.0
    vars.dE = .10
    vars.kx = kx
    vars.ky = ky
    vars.kz = kz

    dlg_cfg = tkParams()
    dlg_cfg.widgets = [
            [{"type": "label", "label": "tolerance:", "width": None}, {"type": "entry","varname": "tollerance", "width": 15}],
            [{"type": "label", "label": "cell size:", "width": None}, {"type": "entry","varname": "cell_size", "width": 15}],
            [
              {"type": "label", "label": "E range:", "width": None}, {"type": "entry","varname": "Emin", "width": 15},
              {"type": "label", "label": "-", "width": None}, {"type": "entry","varname": "Emax", "width": 15},
              {"type": "label", "label": ", dE:", "width": None}, {"type": "entry","varname": "dE", "width": 15},
            ],
            [
              {"type": "label", "label": "k: (", "width": None}, {"type": "entry","varname": "kx", "width": 10},
              {"type": "label", "label": ", ", "width": None}, {"type": "entry","varname": "ky", "width": 10},
              {"type": "label", "label": ", ", "width": None}, {"type": "entry","varname": "kz", "width": 10},
              {"type": "label", "label": ")", "width": None}
            ],
            ]

    CustomDialog_with_config(parent, title = "Setup to calculate phonon mode/animation", config = dlg_cfg, vars = vars)

    if dlg_cfg.ret:
       print(f"k= {vars.kx}, {vars.ky}, {vars.kz}")
    else:
       print("Cancelled")
       return

    inf = {}
    inf["PREFIX"]   = f"ik{ik}_{_is}"
    inf["FCSXML"]   = inf_log.get("FCSXML", "")
    inf["BORNINFO"] = "born"

    inf["NKD"] = inf_log["NKD"]
    inf["KD"]  = inf_log["KD"]
    if haij is None: haij = aij
    inf["aij"] = aij
    inf["SCALE"] = 1.0

    inf["MODE"] = "phonons"
    inf["ISMEAR"] = -1
    inf["NONANALYTIC"] = 3
    inf["ANIME_FORMAT"]   = "xyz"
    inf["PRINTEVEC"]      = 0

    inf["kx"] = vars.kx
    inf["ky"] = vars.ky
    inf["kz"] = vars.kz
    inf["ANIME_CELLSIZE"] = vars.cell_size
    inf["TOLERANCE"] = vars.cell_size
    inf["EMIN"] = vars.Emin
    inf["EMAX"] = vars.Emax
    inf["DELTA_E"] = vars.dE
    inf["ANIME_FRAMES"]   = 30

    return run_anphon(app, cfg.outfile, inf["PREFIX"] + ".log", inf)

def add_popup_menu(popup_menu, app = None, vars = None, parent = None, **kwargs):
    if vars is not None and vars.get("caller", "") == 'plot':
        popup_menu.add_command(label="run",  command = lambda: execute_command(app, vars, parent, **kwargs))


def main():
    app = tkApplication()
    app.cfg = tkParams()
    root = tk.Tk()

    app.cfg.infile = "k25.log"
    app.cfg.outfile = "anime.in"
#    app.cfg.template_dir  = app.replace_path(None, template = ["{dirname}", "template"])
    app.cfg.xyz_dir  = app.replace_path(None, template = ["{dirname}", "xyz"])

    vars = tkParams()
    vars.target = "kxx"
    vars.ihit = 3
    vars.tau_dict = {"label": "test", "ik": 2, "is": 3, "f": 1.5, "tau": 2.5, "kxx": 0.1}
    vars.kp_dict  = {"kx": 0.0, "ky": 0.0, "kz": 0.04, "tau": 2.5}

    execute_command(app = app, vars = vars, parent = root)
#    run_anphon(app, ik = 3, _is = 1)



if __name__ == "__main__":
    main()

