from mp_api.client import MPRester
from pymatgen.io.cif import CifWriter
from pymatgen.io.vasp.inputs import Poscar
import os
import pandas as pd
from pathlib import Path

os.chdir(os.path.dirname(os.path.abspath(__file__)))
cwd = Path.cwd()
mainpath = cwd / 'materials'
mainpath.mkdir(exist_ok=True)
########################################################################################################
YOUR_API_KEY = ""
must_include = ["material_id", "formula_pretty", "chemsys"] #Do not change
add_property = ["band_gap", "builder_meta","nsites", "elements", "composition", \
    "composition_reduced", "symmetry", "deprecated", "deprecation_reasons", "is_gap_direct", \
    "is_metal", "is_stable", "formation_energy_per_atom", "is_magnetic", \
    "ordering", "cbm", "vbm", "efermi", "equilibrium_reaction_energy_per_atom", "volume", "density"]
property = must_include + add_property

#https://api.materialsproject.org/docs
#https://api.materialsproject.org/redoc
#https://workshop.materialsproject.org/lessons/04_materials_api/MAPI%20Lesson%20%28filled%29/
with MPRester(YOUR_API_KEY) as mpr:
    docs = mpr.summary.search(exclude_elements=["Hg"], band_gap=(0.5, 1.0), fields=property)
########################################################################################################    
    n_mat = len(docs)
    n_field = len(property)
    data = [[0 for k in range(n_field)] for l in range(n_mat)]

    for i in range(n_mat):
        for j in range(n_field):
            com_1 = 'docs[' + str(i) + '].' + property[j]
            data[i][j] = eval(com_1)
        
        if "structure" in property:
            structure=docs[i].structure
            formula=docs[i].material_id+'_'+docs[i].formula_pretty
            cif=CifWriter(structure)
            subpath_cif = mainpath / f"{formula}.cif"
            subpath_cif.write_text(str(cif))
            poscar=Poscar(structure)
            subpath_poscar = mainpath / f"{formula}.POSCAR"
            subpath_poscar.write_text(str(poscar))


    df = pd.DataFrame(
        data=data,
        columns=property
    )

    df.to_excel('matdata.xlsx', index=False)