import os
import comtypes.client
import fitz  # PyMuPDF

# 入力ファイルと出力ディレクトリの設定
infile = '20240815-Coverletter.docx'
pdf_file = 'output.pdf'
out_dir = 'images_doc'

def convert_word_to_pdf(word_file, pdf_file):
    application = comtypes.client.CreateObject('Word.Application')
    doc = application.Documents.Open(word_file)
    doc.SaveAs(pdf_file, FileFormat=17)  # 17 はPDF形式を示す定数
    doc.Close()
    application.Quit()

def convert_pdf_to_images(pdf_file, out_dir):
    # 画像を保存するディレクトリが存在しない場合は作成
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    # PDFを開く
    doc = fitz.open(pdf_file)
    for page_num in range(len(doc)):
        page = doc.load_page(page_num)
        pix = page.get_pixmap()
        output_path = os.path.join(out_dir, f'page_{page_num + 1}.png')
        pix.save(output_path)


if __name__ == '__main__':
    current_dir = os.getcwd()
    word_path = os.path.join(current_dir, infile)
    pdf_path = os.path.join(current_dir, pdf_file)

    # Word文書をPDFに変換
    convert_word_to_pdf(word_path, pdf_path)

    # PDFを画像に変換
    convert_pdf_to_images(pdf_path, out_dir)
