import pandas as pd
import openpyxl
from pymatgen.core.composition import Composition
from pymatgen.core.periodic_table import Element


from tklib.tktemplate import read_replacement_db, replace_df_columns
from tklib.tkdb.tkdb import df2excel, df2sqlite3, excel2df


db_path = 'Data_S1_ja3c13574_si_002.xlsx'
cif_dir = 'Data_S2_after_optimization'
replacement_db_path = "replacement_db.xlsx"

out_xlsx    = 'converted.xlsx'
out_sqlite3 = 'converted.sqlite3'


def main():
    replace_dict = read_replacement_db(replacement_db_path, print_level = 2)

    df = excel2df(db_path)
    print(df)

#df からindexとsystemを取得し、f"{cif_dir}/{system}}/{index}.cifを、dfの column名 'cif_path'に追加する
    for sn, row in df.iterrows():
        index = row['index']
        system = row['system']
        formula = row['chemical formula']
        composition = Composition(formula)
        c_dict = composition.as_dict()
        #dfの'cif_path'列のdtypeをstrに変更する
        if 'cif_path' not in df.columns:
            df['cif_path'] = None
        df.at[sn, 'cif_path'] = f"{cif_dir}/{system}/{index}.cif"

#構成元素情報を |原子名@M=質量数を四捨五入した整数@Z=原子番号| の形式で結合する
        atom_inf = '|'
        for i, atom_name in enumerate(c_dict.keys()):
            element = Element(atom_name)
            inf = element.data
#            print(inf)
            M = inf['Atomic mass']
            Z = inf['Atomic no']
            atom_inf += f"{atom_name}@M={int(M + 0.49999)}@Z={Z}@|"
        df['atoms'] = atom_inf

#元素ごとのfieldを追加する場合はコメントを外す
        '''
        for i, key in enumerate(c_dict.keys()):
            el_key = f'element{i+1}'
            if el_key not in df.columns:
                df[el_key] = None
#原子数field名を作る
            nel_key = f'nelement{i+1}'
            if nel_key not in df.columns:
                df[nel_key] = None
#原子名fieldを追加
            df.at[sn, el_key] = key
#原子数fieldを追加
            df.at[sn, nel_key] = c_dict[key]
        '''

    df = replace_df_columns(df, replace_dict, print_level = 1)
    print(df)

    df2excel(df, out_xlsx, print_level = 1)
    df2sqlite3(df, out_sqlite3, print_level = 1)

    print()


def test_composition(formula):
#pymatgenを利用し、化学式 "Na2LiO3(OH)3"を例に、構成元素名と構成元素数の辞書変数compositionを作成する
    comp = Composition(formula)
    print(f"{comp=}")
    print(f"{comp.as_dict()=}")
    print(f"{comp.get_el_amt_dict()=}")
    print(comp.get_integer_formula_and_factor())
    print(comp.reduced_formula)
    print(comp.reduced_composition)
    print(comp.weight)
    print(comp.num_atoms)
    print(comp.elements)
    print(comp.formula)
    print(comp.alphabetical_formula)
#    print(comp.oxide_type)
    print(comp.num_atoms)
    print(comp.chemical_system)
    print(comp.fractional_composition)


#__name__を確認し、直接呼ばれた場合のみ、main()を呼び出す
if __name__ == '__main__':
#    test_composition("Na2LiO3(OH)3")
    main()  
