import tkinter as tk

class Tooltip:
    def __init__(self, widget, text):
        self.widget = widget
        self.text = text
        
        # チップウィンドウを作成
        self.tipwindow = tk.Toplevel(widget)
        self.tipwindow.wm_overrideredirect(True)
        self.tipwindow.wm_geometry('150x30+1+1')
        
        # ラベルを作成
        self.label = tk.Label(self.tipwindow, text=self.text, justify=tk.LEFT, background='#ffffe0', relief=tk.SOLID, borderwidth=1)
        self.label.pack(ipadx=1)
        
        # トリガーとなるイベントにコールバック関数を設定
        self.widget.bind('<Enter>', self.enter)
        self.widget.bind('<Leave>', self.leave)
    
    def enter(self, event):
        # チップウィンドウの位置を設定
        x = y = 0
        x, y, cx, cy = self.widget.bbox('insert')
        x += self.widget.winfo_rootx() + 25
        y += self.widget.winfo_rooty() + 20
        self.tipwindow.wm_geometry('+{}+{}'.format(x, y))
        self.showtip()
    
    def leave(self, event):
        self.hidetip()
    
    def showtip(self):
        self.tipwindow.deiconify()
    
    def hidetip(self):
        self.tipwindow.withdraw()

# メインウィンドウを作成
root = tk.Tk()

# ラベルを作成
label = tk.Label(root, text='ホバーしてください。')
label.pack(pady=50)

# フローティングチップスを作成
tip = Tooltip(label, 'ここにチップスの内容が表示されます。')

root.mainloop()