import os
import platform
from numpy import log, abs, sqrt, exp
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg


from tklib.tkutils import pint, pfloat, val2str, val2str
from tklib.tkparams import tkParams
from tklib.tkvariousdata import tkVariousData
from tklib.tksci.tkFit_object import save_fit_config, read_fit_config_from_file
from tklib.tksci.tkFit_mxy_flex import tkFit_mxy
from tklib.tksci.tkFit_lib import conv_input, save_data, add_history

import mobility_polyOS240507 as cmu



ProgramName = "mu_Fit_flex"
config = tkParams()


def initialize_minimize_func():
    os = platform.system()
    config.script = 'mobility_polyOS240312.py'

def get_cmd(xk, fit):
    config.mode = 'fit'
    config.infile = fit.infile
    config.datafile = fit.datafile
    config.fplot = 0
    head = f"python {config.script} {config.mode} {config.infile} {config.datafile}"
    s = ''
    for x in xk:
        s += f" {x}"
    return head + s

def initialize(app):
    argv, narg   = app.get_argv()
    app.cfg  = tkParams()
    cfg      = app.cfg
    cfg.debug       = 0
    cfg.print_level = 0

# File default  parameters
    infile  = "240305出光試料Hallver3.xlsx"
#    infile  = "240305出光試料Hallver3-抜粋.xlsx"
    trainfile = "fit_db.xlsx"
    outfile  = "out.xlsx"
    fitfile  = "mu-fit.xlsx"
    datafile = "mu.xlsx"

# Parameter default  parameters
    Tmin = 0.0
    Tmax = 1000.0
    Nmin = 0.0
    Nmax = 1.0e23

# Optimization default  parameters
    method   = 'nelder-mead'
    jac      = '3-points'
    nmaxiter = 1000
    nmaxcall = 10000
    tol      = 1.0e-5

# Graph default  parameters
    cfg.figsize = [8, 6]
    fontsize = 16
    legend_fontsize = 12
    graphupdateinterval = 10

    initialize_minimize_func()

    arg_config_file = app.replace_path(None, template = ["{dirname}", "{filebody}_arg_config.xlsx"])
    fit_config_file = app.replace_path(None, template = ["{dirname}", "{filebody}_fit_config.xlsx"])

    app.read_arg_config_from_file(arg_config_file)
    read_fit_config_from_file(app, fit_config_file)
#    for inf in app.args_idx_inf:
#        print("inf=", inf)
#    for key, inf in app.args_opt_inf.items():
#        print("key,inf=", key, inf)

    '''
# Positional arguments
    app.add_argument(opt = None, type = "str", var_name = 'file1', defval = '')
    app.add_argument(opt = None, type = "str", var_name = 'file2', defval = '')
    app.add_argument(opt = None, type = "str", var_name = 'file3', defval = '')
    app.add_argument(opt = None, type = "str", var_name = 'file4', defval = '')
    app.add_argument(opt = None, type = "str", var_name = 'file5', defval = '')

# General
    app.add_argument(opt = "--mode", type = "str", var_name = 'mode',  opt_str = "--mode=[init|inf|sim|scan|mapping|sampling|fit|mlr|plot|error]", desc = 'task mode',
                     defval = 'fit')
    app.add_argument(opt = "--use_tkplt", type = "int", defval = 0)

# File configuration
    app.add_argument(opt = '--infile',    type = "str", opt_str = "--infile=path",  desc = 'measured IV Excel/parameter .in file', defval = infile)
    app.add_argument(opt = '--trainfile', type = "str", desc = 'Training history file (.xlsx) for MLR', defval = trainfile)
    app.add_argument(opt = "--outfile",   type = "str", opt_str = "--outfile=path", desc = 'output Excel file',           defval = outfile)
    app.add_argument(opt = "--fitfile",   type = "str", opt_str = "--fitfile=path", desc = 'Fitting result (.xlsx) file', defval = fitfile)
    app.add_argument(opt = "--datafile",  type = "str", opt_str = "input/calculated data .xlsx file",  desc = 'Data file', defval = datafile)

# Data condition
    app.add_argument(opt = "--ix_sort", type = "int|str", desc = 'index iof x_data to sort data_list', defval = '')
    app.add_argument(opt = "--sample", type = "str", defval = "")
    app.add_argument(opt = "--Tmin", type = "float", defval = Tmin)
    app.add_argument(opt = "--Tmax", type = "float", defval = Tmax)
    app.add_argument(opt = "--Nmin", type = "float", defval = Nmin)
    app.add_argument(opt = "--Nmax", type = "float", defval = Nmax)

# For scan
    app.add_argument(opt = "--target_var", type = "str|int",   defval = 0)
    app.add_argument(opt = "--x0",         type = "float|str", defval = "")
    app.add_argument(opt = "--x1",         type = "float|str", defval = "")
    app.add_argument(opt = "--nx",         type = "int",       defval = 5)

# For sampling
    app.add_argument(opt = "--nsample",    type = "int",       defval = 100)

# For MLR
    app.add_argument(opt = "--test_data_ratio", type = "float",   desc = 'Ratio of test data', defval = 0.2)
    app.add_argument(opt = "--fmin_max_mlr",    type = "float",   desc = 'Max value of fmin for MLR',         defval = 60.0)
    app.add_argument(opt = "--alpha",           type = "float",   desc = 'L2 reguralization parameter',       defval = 0.0)
    app.add_argument(opt = "--l1_ratio",        type = "float",   desc = 'L1 reguralization parameter ratio', defval = 0.5)
    app.add_argument(opt = "--use_cv",          type = "int",     desc = 'flag to use CV', defval = 0)
    app.add_argument(opt = "--cv",              type = "int|str", desc = '# of cross validation', defval = "")

# For PHYSBO
    app.add_argument(opt = "--max_num_probes",  type = "int",     defval = 1)
    app.add_argument(opt = "--interval",        type = "int",     defval = 50)
    app.add_argument(opt = "--num_rand_basis",  type = "int",     defval = 400)
    app.add_argument(opt = "--score_mode",      type = "int",     defval = 'EI')
    app.add_argument(opt = "--seed",            type = "int|str", defval = '')

# Optimization configuration
    app.add_argument(opt = "--fhistory",  type = "int", opt_str = "--fhisotry=[0|1]",  desc = 'flag to save hisotry file', defval = 1)
    app.add_argument(opt = "--ffitfiles", type = "int", opt_str = "--ffitfiles=[0|1]", desc = 'flag to save fit files',    defval = 0)
    app.add_argument(opt = "--fplot",     type = "int", opt_str = "--fplot=[0|1]",     desc = 'flag to plot graph',        defval = 1)

    app.varname    = ["N_II_deg0", "kN_II_deg",  "N_NI0",  "kN_NI", "Nii_ndeg0", "kNii_ndeg",       "NT0",    "kNT", "er",  "Eop",   "Nc",   "a2", "delta_EF0", "kdelta_E0"]
    app.unit       = [     "cm-3",          "",   "cm-3",       "",      "cm-3",          "",  "cm-2eV-1",  "cm/eV", "e0",   "eV", "cm-3",     "",        "eV",          ""]
    app.pk_scale = [      "log",          "",    "log",       "",       "log",          "",       "log",       "",   "",     "",  "log",     "",          "",          ""]
#    app.pk_scale = [         "",          "",       "",       "",          "",          "",          "",       "",   "",    "",     "",     "",          "",          ""]
    app.optid      = [          1,           0,        1,        0,           0,           0,           1,        0,    1,      0,      1,      1,           1,           0]
    app.x0         = [     6.5e19,         0.0,   6.5e19,      0.0,      4.3e19,         0.0,      5.5e13,      0.0, 14.0, 0.0446,  2.0e18,   1.6,       0.025,         0.0]
    app.dx         = [     2.0e19,         0.1,   2.0e19,      0.1,      2.0e19,         0.1,      2.0e13,      0.1,  2.0,   0.01,  1.0e18,   0.5,        0.01,        0.01]
    app.kmin       = [     1.0e10,      -1.0e5,   1.0e10,   -1.0e5,      1.0e10,      -1.0e5,      1.0e10,   -1.0e5,  4.0,   0.01,  1.0e17,  0.01,        -0.5,        -0.1]
    app.kmax       = [     1.0e22,       1.0e5,   1.0e22,    1.0e5,      1.0e22,       1.0e5,      1.0e15,    1.0e5, 20.0,   0.10,  1.0e20,   5.0,         0.1,         0.1]
    app.kpenalty   = [        1.0,         1.0,      1.0,      1.0,         1.0,         1.0,         1.0,      1.0,  1.0,    1.0,     1.0, 1.0e6,         1.0,         1.0]

    for i, varname in enumerate(app.varname):
        app.add_argument(opt = f"--{varname}", type = "float", defval = app.x0[i])

    app.add_argument(opt = "--y_scale",     type = "str",   opt_str = "--y_scale=[|log]",         desc = 'Y scale to fit',  defval = "")
    app.add_argument(opt = "--method",      type = "str",   opt_str = "--method=[nelder-mead]",         desc = 'optimization algorism',                      defval = method)
    app.add_argument(opt = "--jac",         type = "str",   opt_str = "--jac=[3-points|2-points|func]", desc = 'first differential',                         defval = jac)
    app.add_argument(opt = "--nmaxiter",    type = "int",   opt_str = "--nmaxiter=int(>1)",             desc = 'maximum interation number for optimization', defval = nmaxiter)
    app.add_argument(opt = "--nmaxcall",    type = "int",   opt_str = "--nmaxcall=int(>1)",             desc = 'maximum number for function calls',          defval = nmaxcall)
    app.add_argument(opt = "--tol",         type = "float", desc = 'eps for optimization',          defval = tol)
    app.add_argument(opt = "--fmax_record", type = "float", desc = 'Upper limit of fmin to record', defval = 1.0e2)

# For error estimation
    app.add_argument(opt = "--kvariance",  type = "float", desc = 'Multiplier of variance for error estimation',  defval = 6.0)
    app.add_argument(opt = "--ncal_error", type = "int", desc = '# of mesh for error probability function',  defval = 11)

# Parameter/plot configuration
    app.add_argument(opt = "--ix_plot", type = "int|str", desc = 'index of x_data to plot as x-axis',  defval = '')
    app.add_argument(opt = "--x_scale", type = "str",     desc = 'x scale to plot [blank|log]',        defval = '')
    app.add_argument(opt = "--print_interval", type = "int", opt_str = "--print_interval=val", desc = 'print interval', defval = 5)
    app.add_argument(opt = "--plot_interval",  type = "int", opt_str = "--plot_interval=val",  desc = 'plot interval',  defval = 5)

# Graph configuration
    app.add_argument(opt = "--fontsize",            type = "int", defval = fontsize)
    app.add_argument(opt = "--legend_fontsize",     type = "int", defval = legend_fontsize)
    app.add_argument(opt = "--graphupdateinterval", type = "int", defval = graphupdateinterval)

    arg_config_outfile = "arg_config.xlsx"
    fit_config_outfile = "fit_config.xlsx"
    save_fit_config(app, fit_config_outfile)
    app.save_arg_config(arg_config_outfile)
    '''

    return cfg

def save_parameters(path, fit, cfg):
    flag_keys       = ["fplot", "fhistory", "ffitfiles"]
    condiction_keys = ["mode", "method", "jac", "nmaxiter", "nmaxcall", "tol", "y_scale"]
    files_keys      = ["infile", "calfile", "outfile", "datafile", "historyfile", "fitfile",
                       "templatepath", 
                       "file1", "file2", "file3", "file4", "file5"]
    data_keys       = ["sample", "Tmin", "Tmax", "Nmin", "Nmax"]
    scan_keys       = ["target_var", "x0", "x1", "nx"]
    exclude_keys    = flag_keys + condiction_keys + fit.varname + files_keys + data_keys + scan_keys \
                      + ["logfile", "parameterfile", "stopfile"]
#                      + ["infile", "calfile", "logfile", "outfile", "datafile", "parameterfile", "stopfile", "historyfile", "fitfile"]

    print(f"Save parameters to [{path}]")
    cfg.save_parameters(path, section = "Preferences", exclude_keys = exclude_keys)
    cfg.save_parameters(path, section = "Files",      keys = files_keys)
    cfg.save_parameters(path, section = "Data",       keys = data_keys)
    cfg.save_parameters(path, section = "Flags",      keys = flag_keys)
    cfg.save_parameters(path, section = "Condition",  keys = condiction_keys)
    cfg.save_parameters(path, section = "Scan",       keys = scan_keys)
    print(f"Save fitting parameters to [{fit.fitfile_path}]")
    fit.save_parameters(fit.fitfile_path, section = "Parameters", keys = fit.varname, vars = cfg)
#    fit.save_fit_config(fit, fit.fitfile_path)

def read_parameters(path, cfg):
#パラメータファイルの読み込み。Parametersセクション以外のみ
    sectios     = ["Preferences", "Flags", "Files", "Scan", "Condition"]
    ignore_keys = ["logfile", "outfile", "parameterfile", "parameterbkfile"]
    print(f"Read parameters from [{path}]")
    for section in sectios:
        cfg.read_parameters(cfg.parameterfile, section = section, ignore_keys = ignore_keys)

def init_fit(app, cfg, print_level = True):
#    print(f"Read parameters from [{cfg.parameterfile}]")
#パラメータファイルの読み込み。Parametersセクション以外のみ
    read_parameters(cfg.parameterfile, cfg)
    app.force_given_args(cfg)

#    cfg.print_parameters(heading = "\ncparams:")
    print("mode                 : {}".format(cfg.mode))
    print("fplot                : {}".format(cfg.fplot))
    print("fhistory             : {}".format(cfg.fhistory))
    print("ffitfiles            : {}".format(cfg.ffitfiles))
    print("stop path            : {}".format(cfg.stopfile))
    print("infile               : {}".format(cfg.infile))
    print("outfile              : {}".format(cfg.outfile))
    print("fitting result file  : {}".format(cfg.fitfile))
    print(f"data file (transfer) : {cfg.datafile}")
    print("parameter file       : {}".format(cfg.parameterfile))
#    print("parameter backup file: {}".format(cfg.parameterbkfile))
    print("history file         : {}".format(cfg.historyfile))
    print()
    print(f"Fitting configuration")
    print(f"  method  : {cfg.method}")
    print(f"  jac     : {cfg.jac}")
    print(f"  tol     : {cfg.tol}")
    print(f"  nmaxiter: {cfg.nmaxiter}")
    print(f"  nmaxcall: {cfg.nmaxcall}")

    fit = tkFit_mxy(method = cfg.method, tol = cfg.tol, nmaxiter = cfg.nmaxiter, nmaxcall = cfg.nmaxcall,
                fplot = cfg.fplot, print_interval = cfg.print_interval, plot_interval = cfg.plot_interval)
    fit.fitfile_path = app.replace_path(None, template = ["{dirname}", cfg.fitfile_template])
    print(f"Fitting parameter file path: {fit.fitfile_path}")
    fit.configure(print_level = 0, cfg = cfg)
    fit.copy_attributes(cfg, ["fmax_record", "infile", "outfile", "datafile", "historyfile", "fitfile", "y_scale"])

# fitのフィッティング変数・条件の定義
    fit.copy_attributes(app, ["varname", "unit", "pk_scale", "optid", "dx", "kmin", "kmax", "kpenalty"])
    x0 = fit.build_parameter_list(fit.varname, source_obj = cfg)
# 起動時引数のパラメターを設定
    fit.pk = x0.copy()
# iniファイルのパラメターを設定
    fit.read_parameters(fit.fitfile_path, section = "Parameters", keys = fit.varname)
#    fit.read_parameters(cfg.parameterfile, section = "Parameters", keys = fit.varname)
#起動時引数で与えられたパラメータを再設定
    app.force_given_args(cfg)
#    fit.print_variables()

    fit.initial_simplex = None
    if cfg.method == 'nelder-mead' or cfg.method == 'simplex':
        if fit.dx:
            fit.initial_simplex = fit.build_initial_simplex()
            for i, splx in enumerate(fit.initial_simplex):
                print(f"  {i}: ", end = '')
                for v in splx:
                    print(f" {v:12.4g}", end = '')
                print()
    
# fitから呼び出される関数の定義
    if cfg.jac == 'func':
        fit.jac = lambda xk: diff1(xk, fit)
    else:
        fit.jac = cfg.jac
    fit.configure(cal_ylist     = lambda pk, run = True: cal_ylist(pk, fit.x_list, fit, run),
                  minimize_func = lambda pk: minimize_func(pk, fit.x_list, fit.y_list, fit.w_list, fit, print_level = print_level),
                  callback      = lambda pk: callback(pk, fit),
                  cal_fmin      = lambda xk_all, x_list, y_list, w_list, run, print_level: \
                                cal_fmin(xk_all, x_list, y_list, w_list, fit = fit, run = run, print_level = print_level),
                 )

# fitのグラフ関連変数の定義
    fit.configure(fplot = fit.fplot, iter = 0, xiter = [], yfmin = [], ycal_list = None, yfin_list = None)
    fit.print_variables(heading = "Fitting parameters:")

    return fit

def read_data(infile, sample, Tmin, Tmax, Nmin, Nmax, ix_sort = '', filetype = ''):
    print(f"Read data from [{infile}]")
    print(f"  target samples: {sample}")
    print(f"  T range: {Tmin} - {Tmax} K")
    print(f"  N range: {Nmin} - {Nmax} cm-3")

    if not os.path.isfile(infile):
        return None, None, None, None, None

    if filetype == 'calc':
        df = pd.read_excel(infile) #, header = True)
        labels = df.columns.tolist()
        data_list = df.values.T
#        print(df)
#        print("labels=", labels, infile)
        if len(labels) >= 4:
            return ['T (K)', 'N (cm-3)', 'mu(obs) (cm2/Vs)', 'mu(sim)'], data_list[0:2], data_list[2:4]

        return ['T (K)', 'N (cm-3)', 'mu(obs) (cm2/Vs)'], data_list[0:2], data_list[2:]
    else:
        data_dict = cmu.read_data(infile, 'IGO+X')

        N_all = []
        T_all = []
        mu_obs_all = []
        for key, d in data_dict.items():
            if sample != "" and sample != key:
                continue

            for i in range(len(d['T'])):
                T = d['T'][i]
                N = d['N'][i]
                mu_obs = d['mu'][i]

                if T < Tmin or Tmax < T:
                    continue
                if N < Nmin or Nmax < N:
                    continue

                T_all.append(T)
                N_all.append(N)
                mu_obs_all.append(mu_obs)
#zipを使って、T_all,N_all,mu_obs_allを並び替える
        if type(ix_sort) is int:
            T_all, N_all, mu_obs_all = zip(*sorted(zip(T_all, N_all, mu_obs_all), key=lambda x: x[ix_sort]))

        return ['T (K)', 'N (cm-3)'], ['mu(obs) (cm2/Vs)'], [T_all, N_all], [mu_obs_all], None

def cal_ylist(xk, x_list, fit = None, run = True, print_level = 1):
    mu_cal_all = cmu.cal_mu_list(x_list[0], x_list[1], *xk, print_level = print_level)

    return [mu_cal_all]

#外部プログラムを呼び出すウ場合の例
    cmd = get_cmd(xk, fit)
    if run:
        print(f"Run [{cmd}]")
        os.system(cmd)
    else:
        print(f"Don't run [{cmd}]")

    labels, x_list, y_list = read_data(fit.datafile, None, fit.cfg.sample, fit.cfg.Tmin, fit.cfg.Tmax, fit.cfg.Tmin, fit.cfg.Nmax, 'calc')
    return [y_list[1]]

def callback(pk, fit, run = True):
    cfg = fit.cfg
    exist_stopfile = os.path.isfile(cfg.stopfile)
    fplot  = getattr(fit, 'fplot', True) # and fit.fplot
    xk_all = fit.recover_parameters(pk, set_member = False)
    n = len(xk_all)

    if fit.stop_flag:
        return False

    print("exist_stopfile:", cfg.stopfile, exist_stopfile)
    if exist_stopfile:
        print()
        print(f"Message: Found stop file [{cfg.stopfile}] for early stop.")
        print()
        fit.stop_flag = True
        return False

    if fplot:
        w = fit.plt.get_current_fig_manager().window
        if w != fit.window:
            print()
            print(f"Message: Graph window is closed for early stop.")
            print()
            fit.stop_flag = True
            return False

    if fit.iter % fit.print_interval == 0:
        print(f"iter: {fit.iter}")
        for i in range(n):
            print(f"  {fit.varname[i]:10}: {xk_all[i]:10.4g} {fit.unit[i]}")
        f = fit.minimize_func(pk)
        print(f"    f={f:12.6g}")

        if fit.get('fmin_list', None) is not None:
            fit.iter_list.append(fit.iter + 1)
            fit.fmin_list.append(f)

    if fplot and fit.iter % fit.plot_interval == 0:
        ycal_list = fit.cal_ylist(xk_all, run = run)

        x_list   = fit.__dict__.get('x_list', None)
        x_labels = fit.__dict__.get('x_labels', None)
        nx = len(x_list)
        if nx == 1 or type(fit.ix_plot) is int:
            if type(fit.ix_plot) is int and fit.ix_plot < nx:
                ix_plot = fit.ix_plot
            else:
                ix_plot = 0
            for i in range(len(fit.y_list)):
                axis = fit.data_axes[i]
                fit.fit_data_list[i][0].set_data(x_list[ix_plot], ycal_list[i])
        else:
            for i in range(len(fit.y_list)):
                axis = fit.data_axes[i]
                fit.fit_data_list[i][0].set_data(range(len(ycal_list[i])), ycal_list[i])

        if fit.get('fmin_list', None) is not None:
            fit.error_data[0].set_data(fit.iter_list, fit.fmin_list)
            fit.error_axis.set_xlim([min(fit.iter_list), max(fit.iter_list)])
            fit.error_axis.set_ylim([min(fit.fmin_list), max(fit.fmin_list)])

#        fit.plt.tight_layout()
        fit.plt.subplots_adjust(top = fit.plot_region[0], bottom = fit.plot_region[1])
        fit.plt.pause(0.0001)

    fit.iter += 1

    fit.retrieve_parameter_list(xk_all, fit.varname, target = cfg)
    fit.print_variables()
    save_parameters(cfg.parameterfile, fit, cfg)

    return True

def cal_penalty(xk_all, fit, print_level = 1):
    kp   = fit.kpenalty
    if kp is None:
        return 0.0

    kmin = fit.kmin
    kmax = fit.kmax
    nkp = len(kp)
    p_tot = 0.0
    for i in range(nkp):
        if xk_all[i] < kmin[i]:
            d = xk_all[i] - kmin[i]
            p = kp[i] * d * d
            p_tot += p
            print(f"**Warning: [{fit.varname[i]}={xk_all[i]}] is smaller than [{kmin[i]}]. Add penalty [{p:10.3g}] to fmin")
            xk_all[i] = kmin[i]
        elif xk_all[i] > kmax[i]:
            d = xk_all[i] - kmax[i]
            p = kp[i] * d * d
            p_tot += p
            print(f"**Warning: [{fit.varname[i]}={xk_all[i]}] is larger than [{kmax[i]}]. Add penalty [{p:10.3g}] to fmin")
            xk_all[i] = kmax[i]

    return p_tot

def cal_difference(ymeas, ysim, y_scale = '', eps = None):
    if y_scale == 'log':
        if eps is not None:
            if ymeas <= 0.0:
                ymeas = eps
            if ysim <= 0.0:
                ysim = eps

        return log(ymeas) - log(ysim)

    return ymeas - ysim

def cal_fmin(xk_all, x_list, y_list, w_list, fit, run = True, print_level = 1):
#    yc_list = ycal_list(x_list, xk_all, fit, run = run, print_level = print_level)
    yc_list = cal_ylist(xk_all, x_list, fit, run = run, print_level = print_level)

    nx = len(x_list[0])
    fmin = 0.0
    eps = 1.0e-20
    yerr_list = []
    for i in range(nx):
        ymeas = fit.y_list[0][i]
        ysim  = yc_list[0][i]
        d = cal_difference(ymeas, ysim, fit.y_scale, eps)
        yerr_list.append(d)
        fmin += d * d

    fit.yerr2_sum = fmin
    fmin = sqrt(fmin / nx)

    fit.yc_list = yc_list
    fit.yerr_list = yerr_list

    return fmin

def cal_total_fmin(xk_all, x_list, y_list, w_list, fit, run, print_level = 1):
    p_tot = cal_penalty(xk_all, fit, print_level = print_level)
    fmin  = cal_fmin(xk_all, x_list, y_list, w_list, fit, run = run, print_level = print_level)

    return fmin + p_tot

def save_history(xk_all, fmin, fit):
    if fit.cfg.fhistory:
        print(f"Add fmin and parameters to [{fit.historyfile}]")
        iter = fit.add_history(None, fit.varname, xk_all, fmin, save_time = True)
#        iter = fit.add_history(fit.historyfile, fit.varname, xk_all, fmin, save_time = True)
        if iter is not None:
            print(f"#{iter}: fmin=", f"{fmin:12.4g}:".strip(), " xk=", end = '')
            for v in xk_all:
                print(f" {v:12.6g}".strip() + " ", end = '')
            print()
        else:
            iter = None
            print(f"  Warning: Failed to add to {fit.historyfile}]")
            print(f"           Check write permission / disk space etc")

    if fit.cfg.ffitfiles and iter is not None:
        save_path = f'fit{iter:04}.xlsx'
        print(f"Save last input and calculation data  to [{save_path}]")
        if hasattr(fit, 'yini_list'):
            ret = save_data(save_path, ['x', 'input', 'initial', 'final'], [fit.x_list[0], fit.y_list[0], fit.yini_list[0], yc_list[0]])
        else:
            ret = save_data(save_path, ['x', 'input', 'final'], [fit.x_list[0], fit.y_list[0], yc_list[0]])
        if not ret:
            print(f"  Warning: Failed to save to [{save_path}]")
            print(f"           Check write permission / disk space etc")

def minimize_func(xk, x_list = None, y_list = None, w_list = None, fit = None, run = True, print_level = 1):
    if fit.icall > fit.nmaxcall:
        if print_level:
            print(f"\nMessage: # of minimize_func calls [{fit.icall}] exceesds nmaxcall [{fit.nmaxcall}].")
            print(f"           Will stop optimization\n")
        fit.stop_flag = True

    if fit.stop_flag:
        return 1.0e300

    if x_list is None:
        x_list = fit.x_list
    if y_list is None:
        y_list = fit.y_list

    nx = len(x_list[0])
    xk_all = fit.recover_parameters(xk, set_member = False)
    fmin = cal_total_fmin(xk_all, x_list, y_list, w_list, fit, run, print_level = print_level)
    if print_level:
        print()
        print(f"  fmin={fmin}")

    if fmin < fit.fmax_record:
        save_history(xk_all, fmin, fit)
    else:
        print(f"Warning: fmin={fmin:g} is larger than fmax_record={fit.fmax_record:g}. Skip recording")

    fit.icall += 1
    fit.fmin = fmin

    return fmin
