import sys
from math import log, sqrt, pow, exp 
import math
import numpy as np
from numpy import log, log10, sqrt, exp
import matplotlib.pyplot as plt
import openpyxl
import pandas as pd
import mplcursors


'''
#############################################################################
Modification from original 2024/3/12 model:
mobility_polyOS240312.py: Added linear N dependence: NT(N) = NT0 + N * kNT etc.
mobility_polyOS240424.py: Changed linear N dependence to log(N) dependence: NT(N) = NT0 + log(N) * kNT etc.
mobility_polyOS240425.py: Added fitting parameters: Nc, a2, delta_EF0, kdelta_EF, where delta_EF = delta_EF0 + kdelta_EF * log(N) for non-degenerated region                        
mobility_polyOS240426.py: Changed delta_EF = delta_EF0 - (kBT/2) * log(N/Nc0)
                          Changed N_II_deg and N_ii_ndeg to use FD-like sigmoid function
mobility_polyOS240507.py: Changed delta_EF = delta_EF0 + kBT * log(N/Nc0)
                          (rejected) Changed sigmoid function to use kBT instead of a2
#############################################################################
'''


Ts_N = [200.0, 250.0, 300.0]
Ns_N = [1.0e15, 2.0e15, 4.0e15, 6.0e15, 8.0e15, 1.0e16, 2.0e16, 4.0e16, 6.0e16, 8.0e16, 
        1.0e17, 2.0e17, 4.0e17, 6.0e17, 8.0e17, 1.0e18, 2.0e18, 4.0e18, 6.0e18, 8.0e18, 9.0e18, 9.5e18,
        1.0e19, 1.5e19, 2.0e19, 3.0e19, 4.0e19, 6.0e19, 8.0e19]

Ts_T = np.arange(20.0, 360.0, 20.0) #[50.0, 100.0, 150.0, 200.0, 250.0, 300.0]
Ns_T = [1.0e16, 1.0e17, 1.0e18, 1.0e19, 1.0e20]

Ts_3D = [50.0, 60.0, 75.0, 90.0, 100.0, 110, 125.0, 140.0, 150.0, 160.0, 175.0, 180.0, 200.0, 225.0, 250.0, 275.0, 300.0, 325.0, 350.0]
Ns_3D = [1.0e15, 2.0e15, 4.0e15, 6.0e15, 8.0e15, 1.0e16, 2.0e16, 4.0e16, 6.0e16, 8.0e16, 
        1.0e17, 2.0e17, 4.0e17, 6.0e17, 8.0e17, 1.0e18, 2.0e18, 4.0e18, 6.0e18, 8.0e18, 9.0e18, 9.5e18,
        1.0e19, 1.5e19, 2.0e19, 3.0e19, 4.0e19, 6.0e19, 8.0e19]

mode = 'N'  # N, T, fit, plot
plot_mode = 'TN'  # TN, NT, TNs, 3d, 3ds

infile   = '240305出光試料Hallver3.xlsx'
datafile = 'mu.xlsx'


# physical constants
h = 6.626e-34  # Planck constant in J s
kB = 1.38e-23  # Boltzmann constant in J/K
e = 1.6e-19  # elementary charge in C
pi = 3.14159265358979323846
e0 = 8.85419e-14
kBe = 8.62e-05      # eV/K
hbar = h / 2.0 / pi / e

me_c2 = 5.11e+05    # me / c / c [eV/c2]
c = 3.00e+10        # c[cm/s]	

# parameters
d = 30              # nm
d *= 0.0000001      # cm

er = 4.999999994138197 #14.0
eopt = 4.05
Eop = 0.0446        # eV
#wop = 1.0 / (hbar*2*pi*c/Eop)

# Effective density of state for conducation band
Nc0 = 2E+18          # [cm-3]	
#Nc_me1 = 2*(2*pi*meff_min*me/(c)^2*kTe/(hbar*2*pi)^2)^(3/2)

# Critical density of Mott transition
Nc = 1.043684749844953e+19  #2E+18          # [cm-3]	
a2 = 0.032384473207834816   #2

N_II_deg0 = 8.59305051866572e+18  #6.5e+19 # (Deg) [cm-3] !!!!!
kN_II_deg = 1332.0490255949967    #0.0     # (Deg) [cm-3] !!!!!

N_NI0 = 8.144847726077806e+19      #4.3e+19     # [cm-3] !!!!!
kN_NI = 89282.25203504661         #0.0

NT0 = 340955769793930.4  #5.5e+13        # [cm-2eV-1]
kNT = 0.0       #0.0

delta_EF0 = -0.06018168976734409  #0.015
#kdelta_EF = -0.00138752  #0.0

meff_min = 0.15
Z_deg = 2
Z_ndeg = 2
#Nii_ndeg0 = 6.61663e19   #1e+19    # (NonDeg)[cm-2eV-1] !!!!!
#kNii_ndeg = -50458.1     #0.0

alpha0 = 0.0   # α[eV-1]
a_bohr = 4.0 * pi * er * e0 * hbar**2 / (meff_min * me_c2 / c / c) / e       # [cm]


fplot = 1


# 2. Parse the command line arguments    
def update_vars():
    global mode, plot_mode, infile, datafile
    global N_II_deg0, kN_II_deg, N_NI0, kN_NI, Nii_ndeg0, kNii_ndeg, NT0, kNT, er, Eop
    global fplot
    global T, N
    
    argv = sys.argv
    nargs = len(argv)
    if nargs > 1:
        mode = argv[1]

    if mode == 'fit':
        if nargs > 2:
            infile = argv[2]
        if nargs > 3:
            datafile = argv[3]
        if nargs > 4:
            N_II_deg0 = float(argv[4])
        if nargs > 5:
            kN_II_deg = float(argv[5])
        if nargs > 6:
            N_NI0 = float(argv[6])
        if nargs > 7:
            kN_NI = float(argv[7])
        if nargs > 8:
            Nii_ndeg0 = float(argv[8])
        if nargs > 9:
            kNii_ndeg = float(argv[9])
        if nargs > 10:
            NT0 = float(argv[10])
        if nargs > 11:
            kNT = float(argv[11])
        if nargs > 12:
            er = float(argv[12])
        if nargs > 13:
            Eop = float(argv[13])
        if nargs > 14:
            fplot = int(argv[14])
    elif mode == 'plot':
        print("a")
        if nargs > 2:
            plot_mode = argv[2]
    else:
        if nargs > 2:
            T = float(argv[2])  # temperature in K
        if nargs > 3:
            N = float(argv[3])  # carrier concentration in m^-3


def cal_mu(T, N, print_level = 0):
    """
    Function to simulate the mobility.
    It calls several functions to calculate different types of mobility and prints them.
    """

    if print_level >= 1:
        print()
        print(f'T: {T}, N: {N}')

#====== changed: start
    kTe = kB * T / e    # kTe, eV

    logN  = log(N)
    logNc = log(Nc)
    logNNc = log(N / Nc0)
#    ke = (logN - logNc) / kTe
    ke = (logN - logNc) / a2
    if ke > 300.0:
        f = 0.0
    elif ke < -300.0:
        f = 1.0
    else:
        expNNc = exp(ke)
        f =  1.0 / (1.0 + expNNc)
#    f =  1.0 / (1.0 + pow(N / Nc, 1.0 / a2))
    N_II = N_II_deg0 + kN_II_deg * logN
    N_II_deg = N_II * (1.0 - f)
    Nii_ndeg = N_II * f
    N_NI     = N_NI0     + kN_NI     * logN
    NT       = NT0       + kNT       * logN
    delta_EF = delta_EF0 + kTe * logNNc
#    delta_EF = delta_EF0 - kTe / 2.0 * logNNc
#    delta_EF = delta_EF0 + kdelta_EF * logN #pow(N, 2.0/3.0)

#C   = -hbar * hbar / (2.0 * a_bohr * a_bohr * meff_min * me_c2 * c * c * kTe)       
    C = hbar**2 / (2.0 * a_bohr**2 * meff_min * me_c2 / (c**2) * kTe)
#====== changed: end

    if print_level >= 1:
        print("a_bohr=", a_bohr)
        print("N_II=", N_II, f" = {N_II_deg0} + {kN_II_deg} * {logN}")
        print("N_II_deg=", N_II_deg) #, f" = {Nii_ndeg0} + {kNii_ndeg} * {logN}")
        print("Nii_ndeg=", Nii_ndeg) #, f" = {Nii_ndeg0} + {kNii_ndeg} * {logN}")
        print("N_NI=", N_NI, f" = {N_NI0} + {kN_NI} * {logN}")
        print("NT=", NT, f" = {NT0} + {kNT} * {logN}")
        print("delta_EF=", delta_EF, f" = {delta_EF0} + (kBT/2) * log({N}/{Nc0})")

    log_n = log(N)
#    =$D$20*(1+2*$D$26*$D$5^2/($D$20*$D$6/($D$7)^2)*(3*PI()^2*(I3))^(2/3))^(1/2)
    meff = meff_min * sqrt(1 + 2.0 * alpha0 * hbar**2 / (meff_min * me_c2 / c**2 * pow((3.0 * pi**2 * N), 2.0/3.0)))
#    print("meff=", meff)
#  original cell: =1/(2*$D$29*$D$27^6*PI())*((EXP($D$28*$D$31^2)*($D$28*$D$31^2-1)/$D$28^2)+1/$D$28^2)
    
# qs[cm-1]  K3: =($D$2/($D$13*$D$3*$F$10)*(1/$I3+1/($D$16*(8)^0.5))^(-1))^0.5
    K_val = (e/(er*e0*kTe)*(1/N+1/(Nc*(8)**0.5))**(-1))**0.5
    if print_level >= 1:
        print("K_val=", K_val)

#qs×a0[-] 
    L_val = K_val*a_bohr
    if print_level >= 1:
        print("L_val=", L_val)
    
#D=Q0/(πa0^2) [cm2]
    M_val = 356 * exp(-6.2*(a_bohr*K_val)**0.675 )+0.05
    if print_level >= 1:
        print("M_val=", M_val)   
   
#k'_1 [-]
    N_val = (M_val/4)**(-1/3)
    if print_level >= 1:
        print("N_val=", N_val)
    
#C x k'_1^2
    O_val = C*N_val**2
    if print_level >= 1:
        print("O_val=", O_val)
    
#'1/a0^4∫k'^3/Qexp(-Ck'^2)dk'
    P_val = - 1/(2*M_val*a_bohr**6*pi*C**2)*((O_val+1)*exp(-1*O_val)-1)+1/(8*pi*a_bohr**6*C**(7/2))*((O_val**(5/2)+5/2*O_val**(3/2)+15/2*O_val**(1/2))*exp(-1*O_val)+15/8*sqrt(pi)*(1-math.erf(sqrt(O_val))))
    if print_level >= 1:
        print("P_val(original)=", P_val)
    P_val = -1/(2*M_val*a_bohr**6*pi*C**2) \
            *((O_val+1)*exp(-1*O_val)-1) \
            +1/(8*pi*a_bohr**6*C**(7/2)) \
            *((O_val**(5/2)+5/2*O_val**(3/2)+15/2*O_val**(1/2))*exp(-1*O_val) \
            +15/8*sqrt(pi)*(1-math.erf(sqrt(O_val))))
    if print_level >= 1:
        print("P_val(original2)=", P_val)
    P1 = 2*M_val*a_bohr**6*pi*C**2    
    P2 = (O_val+1)*exp(-1*O_val)-1
    P3 = 8*pi*a_bohr**6*C**(7/2)
    P4 = O_val**(5/2)+5/2*O_val**(3/2)+15/2*O_val**(1/2)
    P5 = 15/8*sqrt(pi)*(1-math.erf(sqrt(O_val)))
    P_val = -1 / P1 * P2 + 1 / P3 * (P4 * exp(-1*O_val) + P5)
    if print_level >= 1:
        print("P_val(original3)=", P_val)

    mu_NI = P_val/(3*hbar*N_NI)*(2/pi)**0.5*(2*a_bohr**2*C)**(5/2)    
    mu_NI_inv = 1.0 / mu_NI
    if print_level >= 1:
        print()
        print("mu_NI=", mu_NI)
        print("mu_NI_inv=", mu_NI_inv)

#  original cell: =1/$D$2*(3*PI()^2)^(1/3)*$D$13*$D$3*($D$5*2*PI())^2*$I3^(1/3)/($J3*$D$6/($D$7^2))
    AA_val = 1.0 / e * (3 * pi**2)**(1.0/3.0) * er * e0 * (hbar * 2.0 * pi)**2 * N**(1.0/3.0) / (meff * me_c2 / (c**2))
    if print_level >= 1:
        print("AA_val=", AA_val)

    AB_val = 1.0 - meff
#  original cell: =(1+4*AB3/AA3*(1-AB3/8))*LN(1+AA3)-AA3/(1+AA3)
#                   -2*AB3*(1-5*AB3/16)
    AC_val = (1.0 + 4.0 * AB_val / AA_val * (1.0 - AB_val / 8.0)) * log(1.0 + AA_val) - AA_val / (1.0 + AA_val) \
           - 2.0 * AB_val * (1.0 - 5.0 * AB_val / 16.0)          
    if print_level >= 1:
        print("AB_val=", AB_val)
        print("AC_val=", AC_val)

#  original cell: =1/$D$2^2*3*($D$13*$D$3)^2*($D$5*2*PI())^3
#                *$I3/(($J3*$D$6/$D$7^2)^2*$AC3*$D$17^2*($D$18+$I3/$D$17))
# AD
    if print_level >= 1:
        print()
        print("e=", e);
        print("er=", er);
        print("e0=", e0);
        print("hbar=", hbar);
        print("pi=", pi);
        print("c=", c);
        print("meff=", meff);
        print("me_c2=", me_c2);
        print("AC_val=", AC_val);
        print("Z_deg=", Z_deg);
        print("N_II_deg=", N_II_deg);
        print("N=", N);
    mu_II_deg = 1.0 / e**2 * 3.0 * (er * e0)**2 * (hbar * 2 * pi)**3 \
                * N / ((meff * me_c2 / c**2)**2 * AC_val * Z_deg**2 * (N_II_deg + N / Z_deg))
    if print_level >= 1:
        print("mu_II_deg(original)=", mu_II_deg)
    mu_II_deg1 = 1.0 / (e * e) * 3.0 * pow(er * e0, 2.0) * pow(hbar * 2 * pi, 3.0) 
    mu_II_deg2 = N / (pow(meff * me_c2 / c / c, 2.0) * AC_val * Z_deg * Z_deg * (N_II_deg + N / Z_deg))
    mu_II_deg = mu_II_deg1 * mu_II_deg2
    mu_II_deg_inv = 1.0 / mu_II_deg
    if print_level >= 1:
        print("mu_II_deg1=", mu_II_deg1)
        print("mu_II_deg2=", mu_II_deg2)
        print("mu_II_deg=", mu_II_deg)
        print("mu_II_deg_inv=", mu_II_deg_inv)

#====== changed: start
#  original cell: =$AE3*(1/(1+($D$16/$I3)^($D$35)))
    AF_val = mu_II_deg_inv
#    AF_val = mu_II_deg_inv * (1.0 / (1.0 + (Nc / N)**a2))
#====== changed: end

    AG_val = 1.0 / AF_val
# AH
    A_T0_NII_deg = AF_val

# AJ
    A_T0_NNI_NII_deg = mu_NI_inv + A_T0_NII_deg
    if print_level >= 1:
        print("AF_val=", AF_val)
        print("AG_val=", AG_val)
        print("mu_NI_inv=", mu_NI_inv)
        print("A_T0_NII_deg=", A_T0_NII_deg)
        print("A_T0_NNI_NII_deg=", A_T0_NNI_NII_deg)

#  original cell: =(1/$D$2^2)*128*(2*PI())^0.5*($D$13*$D$3)^2*$F$10^(3/2)/(($J3*$D$6/$D$7^2)^0.5*$D$23^2*$D$24)*$AC3
# AL
#    mu_II_ndeg = (1.0 / e**2) * 128.0 * (2.0 * pi)**0.5 * (er * e0)**2 * kTe**(3.0/2.0) / ((meff * me_c2 / c**2)**0.5 * Z_ndeg**2 * Nii_ndeg) * AC_val
#    mu_II_ndeg_inv = 1.0 / mu_II_ndeg
    A_mu_II_ndeg = ((meff * me_c2 / c**2)**0.5 * Z_ndeg**2 * Nii_ndeg)
    if A_mu_II_ndeg == 0.0:
        mu_II_ndeg = 1.0e300
    else:
        mu_II_ndeg = (1.0 / e**2) * 128.0 * (2.0 * pi)**0.5 * (er * e0)**2 * kTe**(3.0/2.0) / A_mu_II_ndeg * AC_val
    if mu_II_ndeg == 0.0:
        mu_II_ndeg_inv = 1.0e300
    else:
        mu_II_ndeg_inv = 1.0 / mu_II_ndeg

    if print_level >= 1:
        print("mu_II_ndeg=", mu_II_ndeg)
        print("mu_II_ndeg_inv=", mu_II_ndeg_inv)

#====== changed: start
#  original cell: =$AM3*(1/(1+($D$16/$I3)^(-$D$35)))
    AN_val = mu_II_ndeg_inv
#    AN_val = mu_II_ndeg_inv * (1.0 / (1.0 + (Nc / N)**(-a2)))
#====== changed: end
    if print_level >= 1:
        print("AN_val=", AN_val)

    if AN_val == 0.0:
        AO_val = 1.0e300
    else:
        AO_val = 1.0 / AN_val

# AP
    A_32 = AN_val * T**(3.0 / 2.0)
    if print_level >= 1:
        print("A_32=", A_32)

#  original cell: =$D$2*$J3*$D$6/$D$7^2/(4*PI()*$D$3*$D$5^2)*(2*$J3*$D$6/$D$7^2*$D$15)^(1/2)*(1/$D$14-1/$D$13)
# AR
    A_op = e * meff * me_c2 / c**2 / (4.0 * pi * e0 * hbar**2) * (2.0 * meff * me_c2 / c**2 * Eop)**0.5 * (1.0 / eopt - 1.0 / er)
    if print_level >= 1:
        print("A_op=", A_op)

# AS
    mu_opt_inv = A_op / (exp(Eop / kTe) - 1.0)
    mu_opt = 1.0 / mu_opt_inv
    if print_level >= 1:
        print("mu_opt_inv=", mu_opt_inv)
        print("mu_opt=", mu_opt)

    mu_ingrain_inv = mu_NI_inv + AF_val + AN_val + mu_opt_inv
    mu_ingrain = 1.0 / mu_ingrain_inv
    if print_level >= 1:
        print("mu_ingrain_inv=", mu_ingrain_inv)
        print("mu_ingrain=", mu_ingrain)

#   original cell: =1/$D$2*(8*$D$13*$D$3*I3)/$D$22^2
# AY
    alpha_GB = 1.0 / e * (8.0 * er * e0 * N) / NT**2
    if print_level >= 1:
        print("alpha_GB=", alpha_GB)

#  original cell: =IF($D$21*(1+AY3/(2*$D$21)-(AY3/$D$21+AY3^2/(4*$D$21^2)^(1/2)))>0,
#                     $D$21*(1+AY3/(2*$D$21)-(AY3/$D$21+AY3^2/(4*$D$21^2)^(1/2))),0)
# AZ
    phi_b = delta_EF * (1.0 + alpha_GB / (2.0 * delta_EF) - (alpha_GB / delta_EF + alpha_GB**2 / (4.0 * delta_EF**2)**0.5))
    if phi_b < 0.0:
        phi_b = 0.0
    if print_level >= 1:
        print("phi_b=", phi_b)

    mu_tot = mu_ingrain * exp(-phi_b / kTe)
    if mu_tot == 0.0:
        mu_tot_inv = 1.0e300
    else:
        mu_tot_inv = 1.0 / mu_tot
    
    if print_level >= 1:
        print("mu_tot=", mu_tot)
        print("mu_tot_inv=", mu_tot_inv)

    mu_GB_inv =  mu_tot_inv - mu_ingrain_inv
    if mu_GB_inv < 1.0e-10:
        mu_GB_inv = 1.0e-10
    mu_GB = 1.0 / mu_GB_inv
    if print_level >= 1:
        print("mu_GB_inv=", mu_GB_inv)
        print("mu_GB=", mu_GB)

# BH
    S_mu_NI =  mu_NI_inv / mu_tot_inv
# BI
    S_mu_NII_deg = AF_val / mu_tot_inv
# BJ
    S_mu_NII_ndeg = AN_val / mu_tot_inv
# BK
    S_mu_opt = mu_opt_inv / mu_tot_inv
#   original cell: =$BD3/$BC3
# BL
    S_mu_GB = mu_GB_inv / mu_tot_inv
# BM
    S_tot = S_mu_NI + S_mu_NII_deg + S_mu_NII_ndeg + S_mu_opt + S_mu_GB
    if print_level >= 1:
        print("S_mu_NI      =", S_mu_NI)
        print("S_mu_NII_deg =", S_mu_NII_deg)
        print("S_mu_NII_ndeg=", S_mu_NII_ndeg)
        print("S_mu_opt     =", S_mu_opt)
        print("S_mu_GB      =", S_mu_GB)
        print("S_tot        =", S_tot)

# D27: a_bohr  D28: C  D29: D1  D30: D2  D31: k1  D32: k2  D33: k3  D34: a1  D35 a2
# D21: delta_EF

    return mu_tot, mu_GB, mu_ingrain, mu_opt, mu_NI, mu_II_ndeg, mu_II_deg, \
           S_tot, S_mu_NI, S_mu_NII_deg, S_mu_NII_ndeg, S_mu_opt, S_mu_GB

#    mu_Nii = cal_mobility_ii(T, N)
#    mu_Nni = cal_mobility_ni(T, N)
#    mu_op = cal_mobility_op(T, N)
#    mu = cal_mobility(T, N, mu_Nii, mu_Nni, mu_op)
    
#    print(f'mu_Nii: {mu_Nii:.4f}, mu_Nni: {mu_Nni:.4f}, mu_op: {mu_op:.4f}, mu: {mu:.4f}')
#    return 0.0


def read_data(infile, x_key = 'IGO+X', print_level = 1):
    wb = openpyxl.load_workbook(infile)
    sheet = wb.active
    ncolumns = sheet.max_column
    nrows = sheet.max_row
    data_list = [[None for i in range(nrows)] for j in range(ncolumns)]    

    labels = []
    for irow, row in enumerate(sheet.iter_rows()):
        if irow == 0:
            for c in row:
                labels.append(c.value)
            continue

        data_list.append([])
        for icol, c in enumerate(row):
            data_list[icol][irow-1] = c.value

    if print_level:
        print("labels=", labels)
#    print("data_list=", data_list[0])
#    print("data_list=", data_list[1])
    sample_list = []
    data_dict = {}
    prev_sample = ''
    for irow, x in enumerate(data_list[1]):
        if x != x_key:
            continue

        sample = data_list[0][irow]

        if sample != prev_sample:
            sample_list.append(sample)
            if prev_sample != '':
                data_dict[prev_sample] = {'T': T_list, 'N': N_list, 'mu': mu_list}
            T_list = []
            N_list = []
            mu_list = []
            prev_sample = sample
        
        T = data_list[6][irow]
        N = data_list[7][irow]
        mu = data_list[8][irow]
 
        if print_level:
            print("T,N,mu=", T, N, mu)
        T_list.append(T)
        N_list.append(N)
        mu_list.append(mu)
    data_dict[sample] = {'T': T_list, 'N': N_list, 'mu': mu_list}

    return data_dict

def extract_mu_list(data_dict, Ts):
        mu_list_sample = []
        N_list_sample = []

        for sample in data_dict.keys():
            T_list = data_dict[sample]['T']
            mu_list = data_dict[sample]['mu']
            N_list = data_dict[sample]['N']

#            print(f"sample={sample}, T_list={T_list}, mu_list={mu_list}")
            iTmin = np.argmin(T_list)
            iTmax = np.argmax(T_list)
            Tmin = T_list[iTmin]
            Tmax = T_list[iTmax]
#            print(f"sample: {sample}, Tmin: {Tmin}, Tmax: {Tmax}")

            if Ts <= Tmin:
                mu = mu_list[iTmin]
                N = N_list[iTmin]
            elif Ts >= Tmax:
                mu = mu_list[iTmax]
                N = N_list[iTmax]
            else:
                for i in range(len(T_list)-1):
                    if (T_list[i] - Ts) * (T_list[i+1] - Ts) <= 0.0:
                        mu = mu_list[i] + (mu_list[i+1] - mu_list[i]) / (T_list[i+1] - T_list[i]) * (Ts - T_list[i])
                        N = N_list[i] + (N_list[i+1] - N_list[i]) / (T_list[i+1] - T_list[i]) * (Ts - T_list[i])
                        break
            print(f"sample={sample}, T={Ts}, mu={mu}, N={N}")
            mu_list_sample.append(mu)
            N_list_sample.append(N)
        
        return mu_list_sample, N_list_sample

def plot_T():
    print()
    mu_tot_list = []
    for iN in range(len(Ns_T)):
        mu_tot_list.append([])

    print(f"{'T':10}\t{'N':12}\t{'mu_tot':12}\n")
    for iN, N in enumerate(Ns_T):
        for iT, T in enumerate(Ts_T):
            mu_tot, mu_GB, mu_ingrain, mu_opt, mu_NI, mu_II_ndeg, mu_II_deg, \
                S_tot, S_mu_NI, S_mu_NII_deg, S_mu_NII_ndeg, S_mu_opt, S_mu_GB \
                    = cal_mu(T, N)

            mu_tot_list[iN].append(mu_tot)
            print(f"{T:.4f}\t{N:12.4g}\t{mu_tot:12.4g}")

#matplotlibとsubplots(1,1)を使って、Nsをx軸、mu_tot_listの要素mu_tot_list[i]をy軸にした散布図をプロットする。散布図はデータ点を異なる色の点で描き、線でつなげる
    figsize = (10, 8)
    fontsize = 16

#    colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'w']
    fig, ax = plt.subplots(1, 1, figsize = figsize)
    #目盛り、X/Y軸ラベルのフォントサイズをfontsizeに変更
    ax.tick_params(labelsize = fontsize)
    for iN, N in enumerate(Ns_T):
        ax.plot(Ts_T, mu_tot_list[iN], label = f'{N:10.3g} cm-3')
#        ax.plot(Ts_T, mu_tot_list[iN], label = f'{N:10.3g} cm-3', color = colors[iT])

#    ax.set_xscale('log')   
#    ax.set_yscale('log')   

    ax.set_xlabel('T (K)', fontsize = fontsize)
    ax.set_ylabel('$\mu_{tot}$ (m$^2$/Vs)', fontsize = fontsize)

    ax.legend(fontsize=fontsize)
    plt.pause(0.0001)
#    plt.show()
    input("\nPress ENTER to terminate>>")

def save_data(path, labels, data_list):
    df = pd.DataFrame(np.array(data_list).T, columns = labels)
    df.to_excel(path, index = False, header = True) 

def cal_mu_list(T_list, _N_list, _N_II_deg0, _kN_II_deg, _N_NI0, _kN_NI, _Nii_ndeg0, _kNii_ndeg, _NT0, _kNT, _er, _Eop, _Nc, _a2, _delta_EF0, _kdelta_EF, print_level = 1):
    global N_List
    global N_II_deg0, kN_II_deg, N_NI0, kN_NI, Nii_ndeg0, kNii_ndeg, NT0, kNT, er, Eop, Nc, a2, delta_EF0, kdelta_EF
    N_list   = _N_list
    N_II_deg0 = _N_II_deg0
    kN_II_deg = _kN_II_deg
    N_NI0     = _N_NI0
    kN_NI     = _kN_NI
    Nii_ndeg0 = _Nii_ndeg0
    kNii_ndeg = _kNii_ndeg
    NT0       = _NT0
    kNT       = _kNT
    er        = _er
    Eop       = _Eop
    Nc        = _Nc
    a2        = _a2
    delta_EF0 = _delta_EF0
    kdelta_EF = _kdelta_EF

    if print_level:
        print()
        print("#======================================================")
        print("# cal_mu.py")
        print("#======================================================")
        print(f"{datafile=}")
        print(f"N_II_deg = {N_II_deg0:12.6g} + {kN_II_deg} * log(N)")
        print(f"N_NI     = {N_NI0:12.6g} + {kN_NI:12.6g} * log(N)")
        print(f"Nii_ndeg = {Nii_ndeg0:12.6g} + {kNii_ndeg:12.6g} * log(N)")
        print(f"NT       = {NT0:12.6g} + {kNT:12.6g} * log(N)")
        print(f"er       = {er:.6f}")
        print(f"Eop      = {Eop:.6g}")
        print(f"Nc       = {Nc:.6g}")
        print(f"a2       = {a2:.6f}")
        print(f"delta_EF = {delta_EF0:12.6g} - kBT/2 * log(N/Nc0)")

    mu_tot_list = []
    for iT in range(len(Ts_N)):
        mu_tot_list.append([])

    mu_cal_all = []
#    print(f"{'T':10}\t{'N':12}\t{'mu_tot':12}\n")
    for iT, T in enumerate(T_list):
        N = N_list[iT]
        mu_tot, mu_GB, mu_ingrain, mu_opt, mu_NI, mu_II_ndeg, mu_II_deg, \
                S_tot, S_mu_NI, S_mu_NII_deg, S_mu_NII_ndeg, S_mu_opt, S_mu_GB \
                    = cal_mu(T, N)
        mu_cal_all.append(mu_tot)
#            print(f"{T:.4f}\t{N:12.4g}\t{mu_tot:12.4g}")

    return mu_cal_all


def plot_N():
    sample = 'IGO+X'
    data_dict = read_data(infile, sample)
#    print("samples=", sample_list)
#    print("samples=", data_dict.keys())

    print()
    mu300, N300 = extract_mu_list(data_dict, 300.0)
    mu250, N250 = extract_mu_list(data_dict, 250.0)
    mu200, N200 = extract_mu_list(data_dict, 200.0)
    print("300k:", mu300)  
    print("     ", N300)
    print("250k:", mu250)
    print("     ", N250)
    print("200k:", mu200)
    print("     ", N200)
 
    mu_tot_list = []
    for iT in range(len(Ts_N)):
        mu_tot_list.append([])

    print(f"{'T':10}\t{'N':12}\t{'mu_tot':12}\n")
    for iT, T in enumerate(Ts_N):
        for N in Ns_N:
            mu_tot, mu_GB, mu_ingrain, mu_opt, mu_NI, mu_II_ndeg, mu_II_deg, \
                S_tot, S_mu_NI, S_mu_NII_deg, S_mu_NII_ndeg, S_mu_opt, S_mu_GB \
                    = cal_mu(T, N)

            mu_tot_list[iT].append(mu_tot)
            print(f"{T:.4f}\t{N:12.4g}\t{mu_tot:12.4g}")

    N_all = []
    T_all = []
    mu_obs_all = []
    mu_cal_all = []
    for key, d in data_dict.items():
        for i in range(len(d['T'])):
            T = d['T'][i]
            N = d['N'][i]
            mu_obs = d['mu'][i]

            mu_tot, mu_GB, mu_ingrain, mu_opt, mu_NI, mu_II_ndeg, mu_II_deg, \
                    S_tot, S_mu_NI, S_mu_NII_deg, S_mu_NII_ndeg, S_mu_opt, S_mu_GB \
                        = cal_mu(T, N)

            T_all.append(T)
            N_all.append(N)
            mu_obs_all.append(mu_obs)
            mu_cal_all.append(mu_tot)

    print(f"Save data to [{datafile}]")
    save_data(datafile, ['T (K)', 'N (cm-3)', 'obs', 'cal'], [T_all,  N_all, mu_obs_all, mu_cal_all])


#=====================================
# Plot
#=====================================
    print("fplot=", fplot)
    if not fplot:
        return

    figsize = (10, 8)
    fontsize = 16

    colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'w']
    fig, ax = plt.subplots(1, 1, figsize = figsize)
    #目盛り、X/Y軸ラベルのフォントサイズをfontsizeに変更
    ax.tick_params(labelsize = fontsize)
    for iT, T in enumerate(Ts_N):
        ax.plot(Ns_N, mu_tot_list[iT], label = f'{T} K', color = colors[iT])
        muT, NT = extract_mu_list(data_dict, T)
        ax.plot(NT, muT, label = f'{T} K', linestyle = '', marker = 'o', markerfacecolor = colors[iT], markeredgecolor = colors[iT])

    ax.set_xscale('log')   
#    ax.set_yscale('log')   

    ax.set_xlabel('N (m$^{-3}$)', fontsize = fontsize)
    ax.set_ylabel('$\mu_{tot}$ (m$^2$/Vs)', fontsize = fontsize)

    ax.legend(fontsize=fontsize)
    plt.pause(0.0001)
    plt.show()
    input("\nPress ENTER to terminate>>")

def plot_3d():
    Ts_3D = [50.0, 75.0, 100.0, 125.0, 150.0, 175.0, 200.0, 225.0, 250.0, 275.0, 300.0, 325.0, 350.0]
    Ns_3D = [1.0e15, 2.0e15, 4.0e15, 8.0e15, 2.0e16, 4.0e16, 8.0e16, 2.0e17, 4.0e17, 8.0e17, 2.0e18, 4.0e18, 8.0e18, 
             2.0e19, 4.0e19, 8.0e19]
    mu_list = []
    logN_list = log10(Ns_3D)
    for T in Ts_3D:
        for N in Ns_3D:
            mu_tot, mu_GB, mu_ingrain, mu_opt, mu_NI, mu_II_ndeg, mu_II_deg, \
                S_tot, S_mu_NI, S_mu_NII_deg, S_mu_NII_ndeg, S_mu_opt, S_mu_GB \
                    = cal_mu(T, N)
            mu_list.append(mu_tot)  

    N, T = np.meshgrid(logN_list, Ts_3D)
    mu = np.array(mu_list).reshape(len(Ts_3D), len(Ns_3D))

    if plot_mode.lower() == '3ds':
        fig, ax = plt.subplots(1, 1, tight_layout=True)
        #N,T,muで等高線を描く。各領域は色で塗りつぶす
        cm = 'PuOr' #'magma' # 'coolwarm'
        plot_data = ax.contour(N, T, np.array(mu), 100, cmap = cm, linestyles = '-', linewidths = 1.0)
#        plot_data = ax.contourf(N, T, log10(np.array(mu)), 100, cmap = cm)
        #等高線のmuのatainobarを表示
        fig.colorbar(plot_data, ax=ax) 
        ax.set_xlabel('log(N / cm-3)')
        ax.set_ylabel('T (K)')
    else:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ax.plot_surface(N, T, mu, cmap='coolwarm')
        ax.set_xlabel('log(N / cm-3)')
        ax.set_ylabel('T (K)')
        ax.set_zlabel('mu_tot (m^2/Vs)')
#    ax.set_xscale('log')
    plt.pause(1.0e-5)

    input("\nPress ENTER to terminate>>\n")

def plot_TNs():
    sample = 'IGO+X'
    data_dict = read_data(infile, sample, print_level = 0)
    print("samples:")
    N300_list = []
    N_data_list = []
    T_data_list = []
    mu_data_list = []
    for sample, d in data_dict.items():
        T_list  = d['T']
        N_list  = d['N']
        mu_list = d['mu']
        N300 = None
        Tmax = -100
        for N, T in zip(N_list, T_list):
            if Tmax < T:
                N300 = N
                Tmax = T
        N300_list.append(N300)
        T_data_list.append(T_list)
        N_data_list.append(N_list)
        mu_data_list.append(mu_list)
        print(f"  {sample}: N300={N300}")
    
    print()
    '''
#    mu300, N300 = extract_mu_list(data_dict, 300.0)
    print("300k:", mu300)  
    print("     ", N300)
    '''

    ndata = len(mu_data_list)
    nx = int(sqrt(ndata) + 0.99999)
    ny = int(ndata / nx + 0.9999)
    if nx * ny < ndata:
        nx += 1
    fig, axes = plt.subplots(nx, ny, figsize = (12, 12), sharex = 'all')
    axes = axes.flatten()

    for iN in range(len(mu_data_list)):
        ax = axes[iN]
        ax.tick_params(labelsize = 16)
        N300 = N300_list[iN]
        mu_list = mu_data_list[iN]
        N_list = N_data_list[iN]
        mu_T_N_list = []
        mu_T_N300_list = []
        print("iN, N300=", iN, N300)
        print("N_list=", N_list)
        for iT, T in enumerate(T_data_list[iN]):
            mu_tot, mu_GB, mu_ingrain, mu_opt, mu_NI, mu_II_ndeg, mu_II_deg, \
                    S_tot, S_mu_NI, S_mu_NII_deg, S_mu_NII_ndeg, S_mu_opt, S_mu_GB \
                        = cal_mu(T, N_list[iT])
            mu_tot_N300, mu_GB, mu_ingrain, mu_opt, mu_NI, mu_II_ndeg, mu_II_deg, \
                    S_tot, S_mu_NI, S_mu_NII_deg, S_mu_NII_ndeg, S_mu_opt, S_mu_GB \
                        = cal_mu(T, N300)
            mu_T_N_list.append(mu_tot)
            mu_T_N300_list.append(mu_tot_N300)
            print("iN=", iN, N300, T, mu_tot, mu_tot_N300)

        ax.plot(T_data_list[iN], mu_list,    label = f'obs {N300:10.4g} cm-3', marker = 'o', markersize = 5.0)
        ax.plot(T_data_list[iN], mu_T_N_list,    label = f'N', marker = 's', markersize = 3.0)
        ax.plot(T_data_list[iN], mu_T_N300_list, label = f'N@300', marker = '^', markersize = 3.0)

#        ax.set_xlabel('T (K)', fontsize = 16)   
        ax.set_ylabel('$\mu_{tot}$ (m$^2$/Vs)', fontsize = 16)
        ax.legend(fontsize = 12)
    axes[len(axes)-1].set_xlabel('T (K)', fontsize = 16)   
    
    plt.tight_layout()
    plt.pause(0.0001)
    input("\nPress ENTER to terminate>>")





def plot_TN():
    Ts_3D = [50.0, 75.0, 100.0, 125.0, 150.0, 175.0, 200.0, 225.0, 250.0, 275.0, 300.0, 325.0, 350.0]
    Ns_3D = [1.0e15, 2.0e15, 4.0e15, 8.0e15, 2.0e16, 4.0e16, 8.0e16, 2.0e17, 4.0e17, 8.0e17, 2.0e18, 4.0e18, 8.0e18, 
             2.0e19, 4.0e19, 8.0e19]


    fig, ax = plt.subplots()

    lines = []
    labels = []
    if plot_mode == 'TN':
        mu_list = []
        logN_list = log10(Ns_3D)
        for N in Ns_3D:
            mu_T = []
            for T in Ts_3D:
                mu_tot, mu_GB, mu_ingrain, mu_opt, mu_NI, mu_II_ndeg, mu_II_deg, \
                    S_tot, S_mu_NI, S_mu_NII_deg, S_mu_NII_ndeg, S_mu_opt, S_mu_GB \
                        = cal_mu(T, N)
                mu_T.append(mu_tot)
            mu_list.append(mu_T)

        for i, N in enumerate(Ns_3D):
            label = f"N={N:10.4g}"
            line, = ax.plot(Ts_3D, mu_list[i], label = label)
            lines.append(line)
            labels.append(label)
#    ax.set_xscale('log')
        ax.set_xlabel('T (K)')
        ax.set_ylabel('mu (cm2/Vs)')
    elif plot_mode == 'NT':
        mu_list = []
        logN_list = log10(Ns_3D)
        for T in Ts_3D:
            mu_N = []
            for N in Ns_3D:
                mu_tot, mu_GB, mu_ingrain, mu_opt, mu_NI, mu_II_ndeg, mu_II_deg, \
                    S_tot, S_mu_NI, S_mu_NII_deg, S_mu_NII_ndeg, S_mu_opt, S_mu_GB \
                        = cal_mu(T, N)
                mu_N.append(mu_tot)
            mu_list.append(mu_N)

        for i, T in enumerate(Ts_3D):
            label = f"T={T:6.2f}"
            line, = ax.plot(Ns_3D, mu_list[i], label = label)
            lines.append(line)
            labels.append(label)
        ax.set_xlabel('N (cm-3)')
        ax.set_ylabel('mu (cm2/Vs)')
        ax.set_xscale('log')

#ラベルを表示するためのカーソルを追加
    cursor = mplcursors.cursor(lines, hover = True)
    current_sel = None

# ホバー時に表示されるラベルのフォーマットを設定
    @cursor.connect("add")
    def on_add(sel):
        nonlocal current_sel
        print("on")
        current_sel = sel
        line_index = lines.index(sel.artist)
        idx = int(sel.index + 0.4999)
        sel.annotation.arrow_patch.set_alpha(0)
        sel.annotation.set(text = f"{labels[line_index]} T={Ts_3D[idx]}", position = (sel.target[0], sel.target[1]))
#遠くにannotationをおいて、矢印を描く場合
#        sel.annotation.set(text=f'x: {sel.target[0]:.2f}\ny: {sel.target[1]:.2f}\n{labels[line_index]}',
#                       position=(20, 20))
#        sel.annotation.get_bbox_patch().set(fc="white", alpha=0.6)

    '''
    @cursor.connect("remove")
#annotationを消す
    def on_remove(sel):
        nonlocal current_sel
        r2 = (current_sel.target[0] - sel.target[0])**2 + (current_sel.target[1] - sel.target[1])**2
#        print("remove: r2=", r2)
        if current_sel and r2 > 1.0:
#        sel.annotation.set_visible(False)
            current_sel.annotation.set(text = "", position = (sel.target[0], sel.target[1]))
            fig.canvas.draw_idle()
            current_sel = None
    '''

# マウスが動いた時に応答する関数を定義
    def on_mouse_move(event):
        nonlocal current_sel
#        print(f"move: x={event.xdata}, y={event.ydata}")
        if current_sel:
            r2 = (current_sel.target[0] - event.xdata)**2 + (current_sel.target[1] - event.ydata)**2
            if r2 > 1.0:
                current_sel.annotation.set_visible(False)
#                current_sel.annotation.set(text = "", position = (0, 0))
                fig.canvas.draw_idle()
            current_sel = None


# motion_notify_eventイベントに対するコールバック関数を登録
    fig.canvas.mpl_connect('motion_notify_event', on_mouse_move)

    ax.legend()
    
    plt.pause(1.0e-5)

    input("\nPress ENTER to terminate>>\n")

def plot():
        print()
        print(f"mode: {mode}")
        print(f"plot mode: {plot_mode}")

        plot_mode.lower()
        if '3d' in plot_mode:
            plot_3d()
        elif plot_mode == 'TN' or plot_mode == 'NT':
            plot_TN()
        elif plot_mode == 'TNs':
            plot_TNs()

def main():
    """
    Main function to be called when the script is run directly.
    Depending on the mode, it calls either the sim() or fit() function.
    """

    global mode

    update_vars()
    print()
    print(f"mode: {mode}")
    print(f"plot mode: {plot_mode}")

    T = 300.0
    N = 1.0e18
    print()
    print(f"cal_mu check at T={T}, N={N}")
    mu_tot, mu_GB, mu_ingrain, mu_opt, mu_NI, mu_II_ndeg, mu_II_deg, \
                    S_tot, S_mu_NI, S_mu_NII_deg, S_mu_NII_ndeg, S_mu_opt, S_mu_GB \
                        = cal_mu(T, N, print_level = 1)
    print("mu_tot=", mu_tot)
#    exit()

    mode = mode.lower()    
    if mode == 'N':
        plot_N()
    elif mode == 'T':
        plot_T()
    elif mode == 'fit':
        fit()
    elif mode == 'plot':
        plot()

#    print()
#    print("erf(x) check:")
#    for x in np.arange(0.0, 3.0, 0.2):
#        print(f"x={x:.4f}\terf(x)={math.erf(x):10.4g}")

if __name__ == "__main__":
    main()
