import numpy as np
from numpy import sin, cos, tan, pi, exp, log, sqrt


from tklib.tksci.tksci import log10, e, kB
from tklib.tkutils import getarg, getintarg, getfloatarg, pconv_by_type, print_line, print_data


# 与えられた変数 params と、座標 x におけるプロファイル関数の値
def ycal(xs, xk):
    x = xs[0]
    ret = 0.0
    for ip in range(0, len(xk), 3):
        I0 = xk[ip]
        x0 = xk[ip+1]
        w  = xk[ip+2]
        if w < 0.001:
            w = 0.001
        a = 0.832554611 / w
        X = a * (x - x0)
        ret += I0 * exp(-X * X)
    return [ret]

def ycal_list(xd_list, xk):
    y = []
    for i in range(len(xd_list[0])):
        yc_list = ycal([xd_list[0][i]], xk)
        y.append(yc_list[0])
    return [y]

def read_data(infile):
# Data parameters: I0, x0, w
    peaks   = [1.1, 0.4, 0.4]
#peaks = [1.1, 0.4, 0.4, 0.5, 0.6, 0.3]
#    nparams = len(peaks)

    xrange = [-1.0, 2.0]
    nx     = 101

    xstep  = (xrange[1] - xrange[0]) / (nx - 1)
    xd = []
    for i in range(nx):
        xd.append(xrange[0] + i * xstep)

    xd_list = [xd]
    yd_list = ycal_list(xd_list, peaks)

    return xd_list, yd_list


# 最小化する目的関数。残差二乗和
def minimize_func(xk, x_list, y_list, w_list, fit):
    if x_list is None:
        x_list = fit.x_list
    if y_list is None:
        y_list = fit.y_list

    nx = len(x_list[0])
    S2 = 0.0
    for i in range(nx):
        yc_list = ycal([x_list[0][i]], xk)
        d  = y_list[0][i] - yc_list[0]
        S2 += d * d

    return S2

# i番目の最適化変数による一次微分。SIMPLEX法では使わない
def diff1i(i, xk, fit):
    ip  = i // 3
    idx = i % 3
    I0 = xk[ip*3]
    x0 = xk[ip*3+1]
    w  = xk[ip*3+2]

    xd = fit.x_list[0]
    yd = fit.y_list[0]

    ret = 0.0
    for id in range(len(xd)):
        if w < 0.001:
            w = 0.001

        yc_list = ycal([xd[id]], [I0, x0, w])
        dy = yd[id] - yc_list[0]

        a  = 0.832554611 / w
        dx = a * (xd[id] - x0)
        e  = exp(-dx*dx)
        if idx == 0: # I0
            dif = e
        elif idx == 1: # x0
            dif = 2.0 * a * dx * I0 * e
        elif idx == 2: # w
            dif = 2.0 * dx * dx / w * I0 * e
        else:
            print("Error in diff1i: Invalid index (ivar={}, ipeak={}, diff_idx={})".format(i, ip, idx))
            exit()
        d = -2.0 * dif * dy
        ret += d

    return ret

# 一次微分関数配列。SIMPLEX法では使わない
def diff1(xk, fit):
    df = np.empty(len(xk))
    for i in range(len(xk)):
        df[i] = diff1i(i, xk, fit)
    return df
