import os
import sys
import csv
import numpy as np
from numpy import sin, cos, tan, pi, exp, log, sqrt
from scipy.optimize import minimize
#import pandas as pd
from matplotlib import pyplot as plt


from tklib.tksci.tksci import log10, e, kB
from tklib.tkutils import getarg, getintarg, getfloatarg, pconv_by_type, print_line, print_data, joinf
from tklib.tkapplication import tkApplication
from tklib.tkparams import tkParams
from tklib.tkvariousdata import tkVariousData
#from tklib.tkfilter import tkFilter
from tklib.tkgraphic.tkplotevent import tkPlotEvent
from tklib.tksci.tkFit_mxy import tkFit_mxy

from minimize_func_V2 import ycal, ycal_list, minimize_func, diff1, diff1i, callback, plot_prepare, plot_update, plot_finalize


"""
Fitting by SIMPLEX and Gaussian Process regressions
"""


#========================================================
# 最適化アルゴリズム
#========================================================
method = "cg"
maxiter = 100
tol = 1.0e-5

#==========================================
# Source parameters to be fitted
#==========================================
# Data parameters: I0, x0, w
peaks   = [1.1, 0.4, 0.4]
#peaks = [1.1, 0.4, 0.4, 0.5, 0.6, 0.3]
nparams = len(peaks)

# 最適化変数の初期値
x0    = [1.3,  0.6, 0.1]

xrange = [-1.0, 2.0]
nx     = 101


#==========================================
# File configurations
#==========================================
initial_csv = 'initial.csv'
final_csv   = 'final.csv'
conv_csv    = 'convergence.csv'


#==========================================
# Graph parameters
#==========================================
fplot  = 1
ngdata = 51
xgmin  = -4.0
xgmax  =  4.0
ygmin  = -4.0
ygmax  =  4.0
tsleep = 0.3


def initialize():
#================================
# Global variables
#================================
    app          = tkApplication()
    argv, narg   = app.get_argv()
    app.cparams  = tkParams()
    cparams      = app.cparams
    cparams.debug       = 0
    cparams.print_level = 0

#    config.module_dir  = './filter'
#    config.module_name = 'S-sigma-T2xlsx'
    cparams.infile = 'test.xlsx'
    cparams.outfile = 'out.xlsx'

    cparams.print_interval = 10
    cparams.plot_interval = 1

    app.varname    = [   "I0",    "x0",    "w"]
    app.unit       = [     "",      "",     ""]
    app.pk_convert = [     "",      "",     ""]
    app.optid      = [      1,       1,      1]
    app.kmin       = [    0.0,    0.01,   0.01]
    app.kmax       = [1.0e100, 1.0e100, 1.0e23]
    app.kpenalty   = [    1.0,     1.0,    1.0]

    app.add_argument(opt = None, type = "str", var_name = 'infile', opt_str = "input .xlsx file",  desc = 'input file', 
                     defval = None, optional = True);

    app.add_argument(opt = "--mode", type = "str", var_name = 'mode',  opt_str = "--mode=[fit]", desc = 'task mode',
                     defval = 'fit', optional = True);

    app.add_argument(opt = "--I0", type = "float", var_name = 'I0', opt_str = "--I0=val", desc = 'Intensity parameter',
                     defval = 1.0, optional = True);
    app.add_argument(opt = "--x0", type = "float", var_name = 'x0', opt_str = "--x0=val", desc = 'Peak center parameter',
                     defval = 1.0, optional = True);
    app.add_argument(opt = "--w", type = "float", var_name = 'w', opt_str = "--w=val", desc = 'Peak width parameter',
                     defval = 0.2, optional = True);

    app.add_argument(opt = "--method", type = "str", var_name = 'method',  opt_str = "--method=[nelder-mead]", desc = 'optimization algorism',
                     defval = 'nelder-mead', optional = True);
    app.add_argument(opt = "--nmaxiter", type = "int", var_name = 'nmaxiter',  opt_str = "--nmaxiter=int(>1)", desc = 'maximum interation number for optimization',
                     defval = 1000, optional = True);
    app.add_argument(opt = "--tol", type = "float", var_name = 'tol',  opt_str = "--tol=val", desc = 'eps for optimization',
                     defval = 1.0e-5, optional = True);

    app.add_argument(opt = "--print_interval", type = "int", var_name = 'print_interval',  opt_str = "--print_interval=val", desc = 'print interval',
                     defval = 5, optional = True);
    app.add_argument(opt = "--plot_interval", type = "int", var_name = 'plot_interval',  opt_str = "--plot_interval=val", desc = 'plot interval',
                     defval = 5, optional = True);

#=============================
# Graph configuration
#=============================
    cparams.figsize             = [8, 6]
    cparams.fontsize            = 12
    cparams.legend_fontsize     = 8
    cparams.graphupdateinterval = 10

    return app, cparams


#=============================
# Treat argments
#=============================
def update_vars(app, cparams, apply_default = True):
    args_opt, args_idx, args_vars = app.read_args(vars = cparams, check_allowed_args = True, apply_default = apply_default)
    if args_opt is None:
        error_no = args_idx
        error_message = args_vars
        app.terminate("\n\n"
                   +  "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
                   + f"!  {error_message}\n"
                   +  "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!",
                   usage = app.usage)


#==========================================
# functions
#==========================================


def fit(app, cparams):
    global xd, yd, xiter, yfmin
    global figure, ax_fit, ax_conv
    global fitdata, convdata

    print("")
    print("Peak fitting")
    print("mode   : ", cparams.mode)
    print("infile : ", cparams.infile)
    print("outfile: ", cparams.outfile)
    print("parameter file : ", cparams.parameterfile)

    xstep  = (xrange[1] - xrange[0]) / (nx - 1)
    xd = []
    for i in range(nx):
        xd.append(xrange[0] + i * xstep)
#    yd = []
    yd   = ycal_list(xd, peaks)
    yini = ycal_list(xd, x0)
    xini = [cparams.I0, cparams.x0, cparams.w]
#    fit = tkFit_mxy(tol = cparams.tol, nmaxiter = cparams.nmaxiter,
#                print_interval = cparams.print_interval, plot_interval = cparams.plot_interval)
#    fit.xlabels = ['x']
#    fit.ylabels = ['y']

#グラフを描画
    fit = tkParams()
    fit.fplot = fplot
    fit.xd = xd
    fit.yd = yd
    fit.nx = len(xd)
    fit.xini = xini
    fit.yini = yini
    fit.yfin = None
    fit.ycal = None
    fit.plt  = plt
    fit.iter = 0
    fit.xiter  = []
    fit.yfmin  = []
    plot_prepare(fit)

#scipy.minimize()による最適化
    res = minimize(lambda xk: minimize_func(xk, fit), x0, jac = lambda xk: diff1(xk, fit), method=method, tol = tol, 
                callback = lambda xk: callback(xk, fit),
                options = {'maxiter':maxiter, "disp":True})
#最適化結果の表示
    print(res)

    yfin = ycal_list(xd, res.x)
    print()
    print(f"x\tyd\tyini\tyfin")
    for i, x in enumerate(xd):
        print(f"{x:.4f}\t{yd[i]:10.4g}\t{yini[i]:10.4g}\t{yfin[i]:10.4g}")

    print()
    plot_finalize(fit)
    print("Press ENTER to terminate:", end = '')
    ret = input()

#    print()
#    fit.print_variables(heading = "Fitting parameters:")
#    print(f"Fitting configuration")
#    print(f"  method  : {cparams.method}")
#    print(f"  tol     : {cparams.tol}")
#    print(f"  nmaxiter: {cparams.nmaxiter}")

#    fit.varname    = app.varname
#    fit.unit       = app.unit
#    fit.pk         = xini.copy()
#    fit.pk_convert = app.pk_convert
#    fit.optid      = app.optid
#    fit.kmin       = app.kmin
#    fit.kmax       = app.kmax
#    fit.kpenalty   = app.kpenalty
#    nvars = len(fit.pk)

#    fit.func = lambda x_list, pk: cal_S2(x_list[0], y_list[0], len(x_list[0]), *pk)
#    fit.minimize_func = lambda pk, x_list = None, y_list = None, w_list = None: minimize_func(fit, pk, x_list, y_list)
#    fit.func = lambda x_list, pk: ycal(x_list, pk)
#    prm_path = 'temp.prm'
    '''
    def cal_S2(fit, pk, x_list, y_list, w_list):
        if os.path.isfile(prm_path):
            print(f"Remove [{prm_path}]")
            os.remove(prm_path,)
        print(f"Create [{prm_path}]")
        fp = open(prm_path, 'w')
        fp.write(f"nparams={len(pk)}\n")
        for val in pk:
            fp.write(f"{val}\n");
        if x_list:
            fp.write(f"nx={len(x_list)} {len(x_list[0])}\n")
            for d in x_list:
                for val in d:
                    fp.write(f"{val}\n");
                fp.write("\n")
        else:
            fp.write(f"nx=0 0\n")
        if y_list:
            fp.write(f"ny={len(y_list)} {len(y_list[0])}\n")
            for d in y_list:
                for val in d:
                        fp.write(f"{val}\n");
                fp.write("\n")
        else:
            fp.write(f"ny=0 0\n")
        if w_list:
            fp.write(f"nw={len(w_list)}\n")
            for val in w_list:
                fp.write(f"{val}\n");
                fp.write("\n")
        else:
            fp.write(f"nw=0\n")
        fp.close()
        cmd = f"python cal_S2.py {prm_path}"
        print(f"Execute [{cmd}]")
        os.system(cmd)
        exit()
    
    fit.minimize_func = lambda pk, x_list = None, y_list = None, w_list = None: cal_S2(fit, pk, x_list, y_list, w_list)

    yini_list = fit.cal_ylist(fit.pk)
    optpk = fit.extract_parameters()

    fini = fit.minimize_func(optpk, x_list = fit.x_list, y_list = fit.y_list, w_list = fit.w_list)
    print("")
    print(f"Initial function: fmin={fini:10.4g}")
    print_data(labels = ["x", "y_osb", "y_cal"], 
            data_list = [fit.x_list[0], fit.y_list[0], yini_list[0]], 
            label_format = '{:^15}', data_format = '{:>15.4g}', header = "Initial data:", print_level = 0)

    cparams.initial_file   = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-initial.xlsx"])
    print()
    print("Initial data are saved to [{cparams.initial_file}]")
    ret = tkVariousData().to_excel(cparams.initial_file, labels = ['x', 'y(input)', 'y(ini)'], 
                            data_list = [fit.x_list[0], fit.y_list[0], yini_list[0]])
    if ret == 0:
        app.terminate(f"Error: Can not write to [{cparams.initial_file}]", pause = True)

#========================================
# plot
#========================================
    fig, axes = plt.subplots(len(fit.y_list), 3, figsize = cparams.figsize)
#    axes = [axes]
#    axes = axes.flatten()

    fit.initial_plot(data_axes = [axes[0], axes[1]], error_axis = axes[2], yini_list = yini_list, label_ini = 'initial',
                fmin = fini, plt = plt, fig = fig,
                fontsize = cparams.fontsize)
    fit.plot_event.remove('error')

#========================================
# Optimize
#========================================
    print("")
    print("Optimize:")
    fit.print_variables(heading = "Variables")
    pfin, ffin, success, res = fit.minimize(cparams.method)
    if success:
        print(f"Converged at iteration: {res.nit}")
    else:
        print(f"Function did not converge")

#========================================
# Final result
#========================================
    I0, x0, w = [*pfin]
    fit.pk = pfin.copy()

    print("Final parameters")
    for i in range(nvars):
         print(f"  {fit.varname[i]:10}: {pfin[i]:10.4g} {fit.unit[i]}")
    print(f"    f={ffin:12.6g}")

    cparams.I0, cparams.x0, cparams.w = pfin
    print("")
    print("Save parameters to [{cparams.parameterfile}]")
    cparams.save_parameters(cparams.parameterfile, section = "Preferences")

    yfin_list = fit.cal_ylist(pfin)
    print("")
    print_data(labels = [fit.xlabels[0], "y_obs", "y_cal"], 
            data_list = [fit.x_list[0], fit.y_list[0], yfin_list[0]], 
                         label_format = '{:^15}', data_format = '{:>15.4g}', header = "Final data:", print_level = 0)
#    fit.print_scores(heading = "\nScores between y(input) and y(fit)", y1 = fit.y, y2 = yfin)

    cparams.final_file = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-final.xlsx"])
    print()
    print("Final data are saved to [{cparams.final_file}]")
    ret = tkVariousData().to_excel(cparams.final_file, labels = ['x', 'y(input)', 'y(ini)', 'y(fin)'], 
                            data_list = [fit.x_list[0], fit.y_list[0], yini_list[0], yfin_list[0]])
    if ret == 0:
        app.terminate(f"Error: Can not write to [{cparams.final_file}]", pause = True)


#=============================
# グラフの表示
#=============================
    print("")
    print("Plot optimized")
    fit.finalize_plot(yfin_list, iter = res.nit, fmin = ffin)

    fit.layout()
#    plt.tight_layout()
    plt.pause(0.01)
    '''


def main():
    print()

    app, cparams = initialize()
    update_vars(app, cparams,apply_default = True)

    cparams.logfile = app.replace_path(cparams.infile)
    print(f"Open logfile [{cparams.logfile}]")
    app.redirect(targets = ["stdout", cparams.logfile], mode = 'w')
#    app.redirect(targets = ["stdout"], mode = 'a', 
#            redirect_traceback = True, output_traceback = 'stdout', display_type_traceback = 'colored')
#    app.redirect_exception()

    cparams.parameterfile   = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}.in"])
#    cparams.parameterbkfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-back.in"])

    print("")
    print("infile               : {}".format(cparams.infile))
    print("parameter file       : {}".format(cparams.parameterfile))
#    print("parameter backup file: {}".format(cparams.parameterbkfile))
    print("mode                 : {}".format(cparams.mode))

    print("")
    print(f"Read [{cparams.parameterfile}]")
    cparams.read_parameters(cparams.parameterfile, section = "Preferences",
                ignore_keys = ["logfile", "outfile", "parameterfile", "parameterbkfile"])
#    cparams.print_parameters(heading = "\ncparams:")

# check args again to use given parameters
    update_vars(app, cparams,apply_default = False)
    cparams.print_parameters(heading = "\ncparams:")

    if cparams.mode == 'fit':
        fit(app, cparams)
    else:
        app.terminate("Error in main: Invalide mode [{}]".format(cparams.mode), usage = app.usage, pause = True)

    app.terminate("", usage = app.usage, pause = True)


if __name__ == "__main__":
    main()
