import sys
import numpy as np
from numpy import sin, cos, tan, pi, exp, log, sqrt
import pandas as pd


argv = sys.argv
nargs = len(argv)
if nargs < 5:
    print()
    print("Not enough arguments: data_path, I0, x0 and w must be given")
    print()
    exit()

data_path = argv[1]
I0 = argv[2]
x0 = argv[3]
w  = argv[4]


def ycal(xs, xk):
    x = xs[0]
    ret = 0.0
    for ip in range(0, len(xk), 3):
        I0 = xk[ip]
        x0 = xk[ip+1]
        w  = xk[ip+2]
        if w < 0.001:
            w = 0.001
        a = 0.832554611 / w
        X = a * (x - x0)
        ret += I0 * exp(-X * X)
    return [ret]

def ycal_list(xd_list, xk):
    y = []
    for i in range(len(xd_list[0])):
        yc_list = ycal([xd_list[0][i]], xk)
        y.append(yc_list[0])
    return [y]

def read_data(path):
    df = pd.read_excel(path)
    labels = df.columns.values.tolist()
    x_list = [df[labels[0]].values.tolist()]

    y_list = []
    for i in range(1, len(labels)):
        y_list.append(df[labels[i]].values.tolist())

    return labels, x_list, y_list
    
def save_data(path, labels, data_list):
    df = pd.DataFrame(np.array(data_list).T, columns = labels)
    df.to_excel(path, index = False, header = True) 

def cal_profile(data_path, I0, x0, w):
    print(f"Read input data from [{data_path}]")
    labels, xd_list, yd_list = read_data(data_path)

    I0 = float(I0)
    x0 = float(x0)
    w  = float(w)
    print(f"parameters: I0 = {I0}, x0 = {x0}, w = {w}")
    xk = [I0, x0, w]
    yc_list = ycal_list(xd_list, xk)

    print(f"Save input and calculated data to [{data_path}]")
    save_data(data_path, [labels[0], labels[1], 'cal'], [xd_list[0], yd_list[0], yc_list[0]])

    return xd_list, ycal_list

def main():
    print()
    print("===================================")
    print("cal_profile.py ", *argv[1:])
    print("===================================")
    
    cal_profile(data_path, I0, x0, w)

    print("End cal_profile.py")


if __name__ == "__main__":
    main()
