import os
import numpy as np
import pandas as pd
import openpyxl


def read_data(infile):
    df = pd.read_excel(infile)
    labels = df.columns.values.tolist()
    x_list = [df[labels[0]].values.tolist()]

    y_list = []
    for i in range(1, len(labels)):
        y_list.append(df[labels[i]].values.tolist())

    return labels, x_list, y_list

def save_data(path, labels, data_list):
    df = pd.DataFrame(np.array(data_list).T, columns = labels)
    df.to_excel(path, index = False, header = True) 

def get_cmd(xk, fit):
    return f"python cal_profile.py {fit.datafile} {xk[0]} {xk[1]} {xk[2]}"
   
def ycal_list(xd_list, xk, fit = None):
    cmd = get_cmd(xk, fit)
#    print(f"launch [{cmd}]")
    os.system(cmd)
    labels, x_list, y_list = read_data(fit.datafile)
    return [y_list[1]]

def add_history_csv(historyfile, varnames, xk, fmin):
    if not os.path.isfile(historyfile):
        print(f"Create [{historyfile}]")
        with open(historyfile, 'w') as out:
            out.write(f"-iter,min:fmin")
            for i in range(len(xk)):
                out.write(f",{varnames[i]}")
            out.write("\n")

    if os.path.isfile(historyfile):
        print(f"Add fmin and xk to [{historyfile}]")
        iter = -1
        with open(historyfile, 'r') as fp:
            for l in fp:
                iter += 1
        with open(historyfile, 'a') as out:
            out.write(f"{iter},{fmin}")
            for x in xk:
                out.write(f",{x}")
            out.write("\n")

def add_history_xlsx(historyfile, varnames, xk, fmin):
    if not os.path.isfile(historyfile):
        wb = openpyxl.Workbook()
    else:
        wb = openpyxl.load_workbook(historyfile)

    ws = wb.active
    max_row = ws.max_row
    max_col = ws.max_column
    print("max_row=", max_row)

    if max_row <= 1 and ws.cell(row=1, column=1).value is None:
        ws.cell(row=1, column=1, value='-iter')
        ws.cell(row=1, column=2, value='min:fmin')
        for i in range(len(xk)):
            ws.cell(row=1, column=3+i, value=varnames[i])

    ws.cell(row=max_row+1, column=1, value=max_row)
    ws.cell(row=max_row+1, column=2, value=fmin)
    for i in range(len(xk)):
        ws.cell(row=max_row+1, column=3+i, value=xk[i]) 
    wb.save(historyfile)

def add_history(historyfile, varnames, xk, fmin):
    ext = os.path.splitext(historyfile)[1]
    if ext.lower() == ".csv":
        return add_history_csv(historyfile, varnames, xk, fmin)
    elif ext.lower() == ".xlsx":
        return add_history_xlsx(historyfile, varnames, xk, fmin)
    else:
        print(f"Warning: Unsupported extension [{ext}] for history file")
        return

def minimize_func(xk, x_list, y_list, w_list, fit):
    if x_list is None:
        x_list = fit.x_list
    if y_list is None:
        y_list = fit.y_list

    print("xk=", xk)
    nx = len(x_list[0])
    fmin = 0.0
    yc_list = ycal_list(x_list, xk, fit)
    for i in range(nx):
        d  = y_list[0][i] - yc_list[0][i]
        fmin += d * d

    print(f"fmin={fmin:12.4g}")
    
    add_history(fit.historyfile, fit.varname, xk, fmin)

    return fmin
