import os
import sys
import glob
import re
import shutil
import openpyxl

sys.path.append("d:/git/tkProg/tklib/python")

from tklib.tkapplication import tkApplication
from tklib.tkparams import tkParams
from tklib.tkutils import getarg, getintarg, getfloatarg, pconv
from tklib.tktemplate import convert_str


"""
Add data from ini file to Excel using template
"""


plugin_ver = "template:0.2"

usage_str = '''
f"  (i) usage: python {sys.argv[0]} input_path output_path template_path"
'''[1:-1]



#================================
# global parameters
#================================
default_ext = '.in'
extensions  = ['.in', '.ini', '.prm']
input_type  = "ini file .in"
output_type = "MS-Excel .xlsx"


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0

    cparams.template_path = "paramters_template.xlsx"
    cparams.input_path    = "input.ini"
    cparams.output_path   = "paramters_db.xlsx"

def update_vars(app = None, cparams = None):
    if getarg(1, None) is None:
        app.terminate(usage = usage)

    cparams.input_path    = getarg(1, cparams.input_path)
    cparams.output_path   = getarg(2, cparams.output_path)
    cparams.template_path = getarg(3, cparams.template_path)

    cparams.infiles = [cparams.input_path]

def check_file_type(infile, inf = None, app = None, cparams = None):
    if not os.path.isfile(infile):
        return f"Error in check_file_type(): file [{infile}] does not exist"

    filebody, ext = os.path.splitext(infile)
    if ext.lower() not in extensions:
        return None

    return input_type

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infiles, app = None, cparams = None, is_print = False):
    if is_print:
        print("")
        print(f"Input file:", infiles)

    if type(infiles) is str:
        infiles = [infiles]

    keyval_list = []
    ini_list = []
    print()
    for f in infiles:
        print(f"Read [{f}]")
        ini = tkParams()
        ini.read_parameters(f)
#        ini.print_parameters()
        keyval_list.append(ini.__dict__)
        ini_list.append(ini)

    inf = {}
    inf["data_list_type"] = "[dict]"
    inf["filename"]    = infiles[0]
    inf["filenames"]   = infiles

    inf["keyval_list"]   = keyval_list
    inf["ini_list"]   = ini_list

    return inf

def get_output_path(infile, inf = None, app = None, cparams = None):
    return cparams.output_path

def print_data(inf, app = None, cparams = None, print_level = 2):
    for f, d in zip(inf["filenames"], inf["keyval_list"]):
        print(f"{f}:")
        for key, val in d.items():
            print(f"  {key}: {val}")

def plot_data(inf, app = None, cparams = None):
    print()
    print("No data to plot in this plugin")

def read_template(path):
    wb = openpyxl.load_workbook(path, data_only = False)
    ws = wb.active
    labels = []
    templates = []
    for i in range(1, ws.max_column + 1):
        labels.append(ws.cell(row = 1, column = i).value)
        templates.append(ws.cell(row = 2, column = i).value)    

    return labels, templates

def open_db_file(output_path, template_path, print_level = 1):
    if not os.path.exists(output_path):
        if print_level:
            print(f"Message: Outfile [{output_path}] does not exist.")
            print(f"         Copy [{cparams.template_path}] to [{outfile}].")
        shutil.copy(cparams.template_path, outfile)

    wb_out = openpyxl.load_workbook(outfile)
    ws_out = wb_out.active

    return wb_out, ws_out

def save_data(outfiles, inf, app = None, cparams = None, 
                wb = None, ws = None, labels = None, templates = None, is_print = False):
    if outfiles is not None:
        if type(outfiles) is str:
            outfile = outfiles
        else:
            outfile = outfiles[0]

    if wb and ws and labels and templates:
        wb_out = wb
        ws_out = ws
    else:
        print("")
        print(f"Read template [{cparams.template_path}]")
        labels, templates = read_template(cparams.template_path)
        print("labels   :", labels)
        print("templates:", templates)
        wb_out, ws_out = open_db_file(outfile, cparams.template_path)

    print("")
    if outfiles is not None:
        print(f"Add data to [{outfile}]:")
    else:
        print(f"Add data:")
    for f, d in zip(inf["filenames"], inf["keyval_list"]):
        print(f"  from [{f}]:")
        replace_dict = d
        for key, val in replace_dict.items():
            if type(val) is not str:
                continue

        abspath = os.path.abspath(f)
        replace_dict["input_path"]      = abspath
        replace_dict["input_dir_path"]  = os.path.dirname(abspath)
        replace_dict["input_dir_name"]  = os.path.basename(replace_dict["input_dir_path"])
        replace_dict["input_file_name"] = os.path.basename(f)
        replace_dict["input_file_body"] = os.path.splitext(replace_dict["input_file_name"])[0]

        irow = ws_out.max_row
        for icol, t in enumerate(templates):
            s_conv = convert_str(t, replace_dict, print_level = 0)
            s_conv = pconv(s_conv)
            print(f"    {irow}: {t}: {s_conv}")  
            ws_out.cell(row = irow + 1, column = icol + 1).value = s_conv

    if outfiles is not None:
        print(f"Save to [{outfile}]")
        wb_out.save(outfile)

    return True

def convert(inf, app = None, cparams = None):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print( "==========================================================================")
    print(" Add data from ini file to excel db file using excel template")
    print( "==========================================================================")
    print(f"Input files: {cparams.infiles}")

    logfile = app.replace_path(cparams.infiles[0], template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infiles[0], cparams = cparams)

    print(f"Input init file: {cparams.infiles}")
    print(f"Template file: {cparams.template_path}")
    print(f"Output file: {cparams.outfile}")

    file_type = check_file_type(cparams.infiles[0])
    if file_type is None:
        app.terminate(f"Error: [{cparams.infiles[0]}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infiles, cparams = cparams)
    print_data(inf)
#    inf = convert(inf, cparams = cparams)
    save_data([cparams.outfile], inf, cparams = cparams)
#    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
